/*===================================================================
 * Copyright(c) AVerMedia TECHNOLOGIES, Inc. 2017
 * All rights reserved
 * =================================================================
 * cn312_hdmi.c
 *
 *  Created on: Jan 25, 2018
 *      Author: 
 *      Version:
 * =================================================================
 */
//#include <linux/kernel.h>
//#include <linux/pci.h>
//#include <linux/interrupt.h>
#include "cxt_mgr.h"
#include "framegrabber.h"
#include "cn312_hdmi.h"
#include "cn312m_reg.h"
#include "mem_model.h"
#include "i2c_model.h"
#include "pci_model.h"
#include "sys.h"
#include "trace.h"
#include "debug.h"
#include "task_model.h"

#include "debug.h"
#include "xil_assert.h"
#include "xil_types.h"
#include "xstatus.h"
#include "xvidc.h"
#include "xparameters.h"
#if defined (XPAR_XUARTLITE_NUM_INSTANCES)
#include "xuartlite_l.h"
#else
#include "xuartps.h"
#endif

#include "xvphy.h"
//#include "xhdmi_edid.h"
#ifdef XPAR_XV_HDMIRXSS_NUM_INSTANCES
#include "xv_hdmirxss.h"
#endif
#include "xil_exception.h"
//#include "xhdmi_edid.h"
#include "xintc.h"

/***************** Macros (Inline Functions) Definitions *********************/
/* These macro values need to changed whenever there is a change in version */
#define APP_MAJ_VERSION 3
#define APP_MIN_VERSION 1

/*slave address*/
#define		XHDMI_SLAVE			0x4E

/*sub address*/
#define		XHDMI_VIN_STABLE			0x24
#define		XHDMI_VIN_WIDTH				0x28
#define		XHDMI_VIN_HEIGHT0			0x2c
#define		XHDMI_VIN_HEIGHT1			0x2e
#define		XHDMI_VIN_FRAME_RATE_CNT	0x30
#define		XHDMI_AUDIO_SAMPLE_RATE_CNT	0x34
#define		XHDMI_XADC_TEMP_OUT			0xf0

#define		XHDMI_CADJ_OFFSET		0x100
#define		XHDMI_CSC_OFFSET		0x100

#if defined (XPAR_XUARTLITE_NUM_INSTANCES)
#define UART_BASEADDR XPAR_MB_SS_0_AXI_UARTLITE_BASEADDR
#else
#define UART_BASEADDR XPAR_XUARTPS_0_BASEADDR
#endif
/**************************** Type Definitions *******************************/
static const U8_T Edid[256] = {
		 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00,
		0x06, 0xD8, 0x46, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x19, 0x1D, 0x01, 0x03, 0x80, 0xA0, 0x5A, 0x78,
		0x22, 0x76, 0x90, 0xA8, 0x54, 0x4D, 0x9F, 0x25,
		0x0E, 0x50, 0x54, 0xFF, 0xFF, 0x80, 0x95, 0x19,
		0xD1, 0x00, 0x81, 0x59, 0x81, 0x99, 0xA9, 0x40,
		0x90, 0x59, 0xB3, 0x00, 0x81, 0x3C, 0x02, 0x3A,
		0x80, 0x18, 0x71, 0x38, 0x2D, 0x40, 0x58, 0x2C,
		0x45, 0x00, 0xC4, 0x8E, 0x21, 0x00, 0x00, 0x1E,
		0x00, 0x00, 0x00, 0xFC, 0x00, 0x41, 0x56, 0x54,
		0x20, 0x43, 0x4E, 0x33, 0x31, 0x32, 0x4D, 0x0A,
		0x20, 0x20, 0x00, 0x00, 0x00, 0xFA, 0x00, 0x45,
		0x7C, 0x61, 0x7C, 0x71, 0x4F, 0x45, 0x59, 0x61,
		0x59, 0x31, 0x59, 0x0A, 0x00, 0x00, 0x00, 0xFD,
		0x00, 0x18, 0x3D, 0x0F, 0x4B, 0x11, 0x00, 0x0A,
		0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x01, 0xC7,
		0x02, 0x03, 0x28, 0xF1, 0x52, 0x10, 0x04, 0x03,
		0x02, 0x20, 0x11, 0x01, 0x1F, 0x12, 0x13, 0x14,
		0x05, 0x21, 0x22, 0x29, 0x2F, 0x2A, 0x2B, 0x23,
		0x09, 0x07, 0x07, 0x83, 0x01, 0x00, 0x00, 0x65,
		0x03, 0x0C, 0x00, 0x10, 0x00, 0xE2, 0x00, 0x4F,
		0x66, 0x21, 0x56, 0xAA, 0x51, 0x00, 0x1E, 0x30,
		0x46, 0x8F, 0x33, 0x00, 0xC4, 0x8E, 0x21, 0x00,
		0x00, 0x1E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xC6
};

/************************** Function Prototypes ******************************/

#if 0 //defined (__arm__) && (!defined(ARMR5))
int OnBoardSi5324Init(void);
#endif

#ifdef XPAR_XV_HDMIRXSS_NUM_INSTANCES
void RxConnectCallback(void *CallbackRef);
void RxBrdgOverflowCallback(void *CallbackRef);
void RxStreamUpCallback(void *CallbackRef);
void RxStreamDownCallback(void *CallbackRef);
void VphyHdmiRxInitCallback(void *CallbackRef);
void VphyHdmiRxReadyCallback(void *CallbackRef);
#endif
void VphyErrorCallback(void *CallbackRef);
#if (XPAR_VPHY_0_TRANSCEIVER == XVPHY_GTXE2)
void VphyPllLayoutErrorCallback(void *CallbackRef);
#endif
void VphyProcessError(void);

static void *g_bus_handle;
static U16_T g_addr;

/************************* Variable Definitions *****************************/
/* VPHY structure */
static XVphy              Vphy;
static U8_T                 VphyErrorFlag;
static U8_T                 VphyPllLayoutErrorFlag;


#ifdef XPAR_XV_HDMIRXSS_NUM_INSTANCES
/* HDMI RX SS structure */
static XV_HdmiRxSs        HdmiRxSs;
static XV_HdmiRxSs_Config *XV_HdmiRxSs_ConfigPtr;
#endif

#ifdef VIDEO_FRAME_CRC_EN
static Video_CRC_Config   VidFrameCRC;
#endif

#ifdef USE_HDCP
static XHdcp_Repeater     HdcpRepeater;
#endif

/* Interrupt Controller Data Structure */
#if 0 //defined (ARMR5) || (__aarch64__) || (__arm__)
static XScuGic     Intc;
#else
static XIntc       Intc;
#endif

static U8_T InfoFrame[8][32];
/**< Demo mode IsPassThrough
 * (AVER_TRUE)  = Pass-through mode
 * (AVER_FALSE) = Color Bar mode
 *  */
U8_T IsPassThrough = (AVER_FALSE);
U8_T StartTxAfterRxFlag = (AVER_FALSE);

#define RGBF_to_BT709	0
#define RGBL_to_BT709	1
#define RGBF_to_BT601	2
#define RGBL_to_BT601	3

#define BT709_to_RGBF	0
#define BT709_to_RGBL	1
#define BT601_to_RGBF	2
#define BT601_to_RGBL	3

#define BT601_to_BT709	4
#define BT709_to_BT601	5

#define hdmi_YUYV 0
#define hdmi_UYVU 1
#define hdmi_RGB3 11
#define hdmi_BGR3 12


typedef struct
{
    i2c_model_bus_handle_t bus_handle;
    pci_model_handle_t pci_handle;
    U16_T addr;
    U8_T id;
    hdmi_callback_t callback_func;
    void *cb_cxt;
    task_model_handle_t taskmgr;
    task_handle_t task_handle;
    hdmi_frameinfo_t frameinfo; 
    hdmi_event_e lock;
    //int wait_hdcp_count;
    BOOL_T is_lock;
    BOOL_T is_state_change; 
    BOOL_T is_lock_state;
    int audio_sample;
    int old_audio_sample;
    hdmi_packetinfo_t packetinfo;
    int device_index;
    int input_stream;
    int print_cnt;

#if 0 // (defined (ARMR5) || (__aarch64__)) && (!defined XPS_BOARD_ZCU104)
    XIicPs_Config *XIic0Ps_ConfigPtr;
    XIicPs_Config *XIic1Ps_ConfigPtr;
#endif

    //BOOL_T VphyErrorFlag;
    //BOOL_T VphyPllLayoutErrorFlag;

    /* Start in color bar */
    //BOOL_T IsPassThrough;
}hdmi_cxt_t;
#if 0
typedef struct 
{
    unsigned rgb_out:1; // 0: YPbPr out 1: RGB out
      
}hdmi_colorformat_setting_t;
#endif
static int hdmi_wr_reg(hdmi_cxt_t *hdmi_cxt, int reg, int val);
static int hdmi_rd_reg(hdmi_cxt_t *hdmi_cxt, int reg);

static int io_mask_write(hdmi_cxt_t *hdmi_cxt,int reg,int mask,int val);
static void hdmi_callback(hdmi_cxt_t *hdmi_cxt,hdmi_event_e event);

//static U8_T hdmi_dev_count=0;
static void hdmi_update_video_timing(hdmi_cxt_t *hdmi_cxt,int ch);
//static void hdmi_get_video_format(hdmi_cxt_t *hdmi_cxt,int *colorformat);
//static void hdmi_set_audio_format(hdmi_cxt_t *hdmi_cxt,int audioformat);


#if 0
static void hdmi_set_out_color(hdmi_cxt_t *hdmi_cxt,hdmi_colorformat_setting_t *out_color_ctrl)
{
    U8_T value=0;
    
    if(out_color_ctrl->rgb_out)
        value |= hdmi_IO_BIT_RGB_OUT;
 
    io_mask_write(hdmi_cxt,hdmi_IO_REG,hdmi_IO_BIT_RGB_OUT,value);
}
#endif

#ifndef usleep

void usleep(unsigned us)
{
    sys_udelay(us);
}

#endif

static int io_mask_write(hdmi_cxt_t *hdmi_cxt,int reg,int mask,int val)
{
    int old_value,value;
    int ret=-1;
    if(hdmi_cxt)
    {
       old_value =hdmi_rd_reg(hdmi_cxt,reg);
       value=old_value & (~mask);
       value |= (val & mask);
       ret=hdmi_wr_reg(hdmi_cxt,reg,value);
    }
    return ret;
}

static int hdmi_wr_reg(hdmi_cxt_t *hdmi_cxt, int reg, int val)
{
    U8_T wrbuf[7];
    int ret = 0;

    if(hdmi_cxt)
    {
        i2c_model_msg_t msg = {
            .addr = hdmi_cxt->addr,
            .flags = 0,
            .buf = wrbuf,
            .len = 7
        };

        wrbuf[0] = (reg >> 16) & 0xff;
        wrbuf[1] = (reg >> 8) & 0xff;
        wrbuf[2] = (reg >> 0) & 0xff;
        wrbuf[3] = (val >> 24) & 0xff;
        wrbuf[4] = (val >> 16) & 0xff; 
        wrbuf[5] = (val >> 8) & 0xff; 
        wrbuf[6] = (val >> 0) & 0xff; 

        ret = i2c_model_transfer(hdmi_cxt->bus_handle, &msg, 1);
        sys_msleep(1);
        mesg("%s addr:0x%x reg:0x%x val=0x%x \n", __func__, hdmi_cxt->addr,reg, val);
    }

    if(ret != 1) {
        mesg("............%s i2c_transfer failed %d reg:%x \n", __func__, ret , reg);
    }
    else {
        return 0;
    }

    return -1;
}

int hdmi_wr_reg_xil(int reg, U8_T* pBuf)
{
    U8_T wrbuf[7];
    int ret = 0;

    //if(1)
    {
        i2c_model_msg_t msg = {
            .addr = g_addr, //hdmi_cxt->addr,
            .flags = 0,
            .buf = wrbuf,
            .len = 7
        };

        wrbuf[0] = (reg >> 16) & 0xff;
        wrbuf[1] = (reg >> 8) & 0xff;
        wrbuf[2] = (reg >> 0) & 0xff;
        wrbuf[3] = pBuf[3];//(val >> 24) & 0xff;
        wrbuf[4] = pBuf[2];//(val >> 16) & 0xff; 
        wrbuf[5] = pBuf[1];//(val >> 8) & 0xff; 
        wrbuf[6] = pBuf[0];//(val >> 0) & 0xff; 

        ret = i2c_model_transfer(g_bus_handle, &msg, 1);
        sys_msleep(1);	
    }

    //mesg("%s i2c_transfer  0x %x %x %x %x \n", __func__, wrbuf[3],wrbuf[4],wrbuf[5],wrbuf[6]);
    if(ret != 1) {
        mesg("............%s i2c_transfer failed %d reg:%x \n", __func__, ret , reg);
    }
    else {
        return 0;
    }

    return -1;
}

static int hdmi_rd_reg(hdmi_cxt_t *hdmi_cxt, int reg)
{
    i2c_model_msg_t msg[2];
    U8_T wrbuf[3];
    U8_T rdbuf[4];
    int ret = 0;
    
    if(hdmi_cxt)
    {
        wrbuf[0] = (reg >> 16) & 0xff;
        wrbuf[1] = (reg >> 8) & 0xff;
        wrbuf[2] = (reg >> 0) & 0xff;
    
        msg[0].addr = hdmi_cxt->addr;
        msg[0].flags = 0;
        msg[0].buf = wrbuf;
        msg[0].len = 3;
    
        msg[1].addr = hdmi_cxt->addr;
        msg[1].flags = I2C_MODEL_READ;
        msg[1].buf = rdbuf;
        msg[1].len = 4;

        ret = i2c_model_transfer(hdmi_cxt->bus_handle, msg, 2);
        //mesg("%s i2c_transfer %d on sub 0x%x%x%x\n", __func__, ret,wrbuf[0],wrbuf[1],wrbuf[2]);
    }

    if(ret != 2) {
        mesg("%s[id%d] i2c_transfer failed %d reg:%x \n", __func__, hdmi_cxt->device_index,ret, reg);
    }
    else {
        mesg("%s addr:0x%x reg:0x%x  val:0x%x %x %x %x \n", __func__, hdmi_cxt->addr, reg, rdbuf[0],rdbuf[1],rdbuf[2] ,rdbuf[3]);
        return (rdbuf[0] <<24 | rdbuf[1] << 16 | rdbuf[2] << 8 | rdbuf[3]);
    }

    return -1;
}

int hdmi_rd_reg_xil(int reg, U8_T* pBuf)
{
    i2c_model_msg_t msg[2];
    U8_T wrbuf[3];
    U8_T rdbuf[4];
    int ret = 0, i;
    
    //if(hdmi_cxt)
    {
        wrbuf[0] = (reg >> 16) & 0xff;
        wrbuf[1] = (reg >> 8) & 0xff;
        wrbuf[2] = (reg >> 0) & 0xff;
    
        msg[0].addr = g_addr; //hdmi_cxt->addr;
        msg[0].flags = 0;
        msg[0].buf = wrbuf;
        msg[0].len = 3;
    
        msg[1].addr = g_addr; //hdmi_cxt->addr;
        msg[1].flags = I2C_MODEL_READ;
        msg[1].buf = rdbuf;
        msg[1].len = 4;

        ret = i2c_model_transfer(g_bus_handle, msg, 2);
    
        //mesg("%s i2c_transfer 0x %x %x %x %x \n", __func__, rdbuf[0],rdbuf[1],rdbuf[2], rdbuf[3]);
    }

    if(ret != 2) {
        mesg("%s i2c_transfer failed %d reg:%x \n", __func__,ret, reg);
    }
    else {
        for(i=0; i<4; i++)
            *(pBuf+i) = rdbuf[3-i];
        return 0;
    }

    return -1;
}

/* General HDMI Application Variable Declaration (Scratch Pad) */

/************************** Function Definitions *****************************/

#if (XPAR_VPHY_0_TRANSCEIVER == XVPHY_GTXE2)
/*****************************************************************************/
/**
*
* This function checks the system PLL, whether it's bonded or non-bonded
* @return None.
*
* @note   None.
*
******************************************************************************/
U8_T PLLBondedCheck (void)
{
	// The colorbar can only be displayed when the GT is not bonded.
	if (XVphy_IsBonded(&Vphy, 0, XVPHY_CHANNEL_ID_CH1)) {
		debug_msg("\r\nThe GT TX and RX are bonded and clocked by "
				   "the RX clock.\r\n");
		debug_msg("Please select independent PLL layout to enable "
				   "TX only mode.\r\n");
	}
	return (XVphy_IsBonded(&Vphy, 0, XVPHY_CHANNEL_ID_CH1));
}
#endif

#if 1
/*****************************************************************************/
/**
* This function outputs the video timing , Audio, Link Status, HDMI RX state of
* HDMI RX core. In addition, it also prints information about HDMI TX, and
* HDMI GT cores.
*
* @param  None.
* @return None.
* @note   None.
******************************************************************************/
void Info(void)
{
	U32_T Data;
	mesg("-----\n");
	mesg("Info\n");
	mesg("-----\n");

#ifdef XPAR_XV_HDMIRXSS_NUM_INSTANCES
	XV_HdmiRxSs_ReportInfo(&HdmiRxSs);
#endif
#ifdef VIDEO_FRAME_CRC_EN
	XVidFrameCrc_Report();
#endif

	// GT
	mesg("------------\n");
	mesg("HDMI PHY\n");
	mesg("------------\n");
	Data = XVphy_GetVersion(&Vphy);
	mesg("  VPhy version : %02d.%02d (%04x)\n",
			   ((Data >> 24) & 0xFF), ((Data >> 16) & 0xFF), (Data & 0xFFFF));
	mesg("\n");
	mesg("GT status\n");
	mesg("---------\n");
#ifdef XPAR_XV_HDMIRXSS_NUM_INSTANCES
	mesg("RX reference clock frequency: %0d Hz\n",
			   XVphy_ClkDetGetRefClkFreqHz(&Vphy, XVPHY_DIR_RX));
	if(Vphy.Config.DruIsPresent == (AVER_TRUE)) {
		mesg("DRU reference clock frequency: %0d Hz\n",
				   XVphy_DruGetRefClkFreqHz(&Vphy));
	}
#endif
	XVphy_HdmiDebugInfo(&Vphy, 0, XVPHY_CHANNEL_ID_CH1);

}
#endif

#ifdef XPAR_XV_HDMIRXSS_NUM_INSTANCES
/*****************************************************************************/
/**
* This function is called when a RX connect event has occurred.
*
* @param  None.
* @return None.
* @note   None.
******************************************************************************/
void RxConnectCallback(void *CallbackRef) {
	XV_HdmiRxSs *HdmiRxSsPtr = (XV_HdmiRxSs *)CallbackRef;

	// RX cable is disconnected
	if(HdmiRxSsPtr->IsStreamConnected == (AVER_FALSE)) {
		Vphy.HdmiRxTmdsClockRatio = 0; // Clear GT RX TMDS clock ratio

#if(LOOPBACK_MODE_EN != 1)
		/* Check for Pass-through */
		/* Doesnt require to restart colorbar
		 * if the system is in colorbar mode
		 */
		if (IsPassThrough) {
			/* Clear pass-through flag
			 * as the system is in TX-Only
			 * mode
			 */
			IsPassThrough = (AVER_FALSE);
		}
#endif

		XVphy_IBufDsEnable(&Vphy, 0, XVPHY_DIR_RX, (AVER_FALSE));

#ifdef USE_HDCP
		/* Call HDCP disconnect callback */
		XHdcp_StreamDisconnectCallback(&HdcpRepeater);
#endif
	} else {
		XVphy_IBufDsEnable(&Vphy, 0, XVPHY_DIR_RX, (AVER_TRUE));

#ifdef USE_HDCP
		/* Call HDCP connect callback */
		XHdcp_StreamConnectCallback(&HdcpRepeater);
#endif
	}

}

/*****************************************************************************/
/**
* This function is called when a RX Bridge Overflow event has occurred.
* This is Error Condition
*
* @param  None.
* @return None.
* @note   None.
******************************************************************************/
void RxBrdgOverflowCallback(void *CallbackRef) {
/*	debug_msg(ANSI_COLOR_YELLOW "RX Video Bridge Overflow"
			ANSI_COLOR_RESET "\r\n");*/
/*	XV_HdmiRx_VideoEnable(HdmiRxSs.HdmiRxPtr, (AVER_FALSE));
	XV_HdmiRx_VideoEnable(HdmiRxSs.HdmiRxPtr, (AVER_TRUE));*/
}

/*****************************************************************************/
/**
* This function is called when the GT RX reference input clock has changed.
*
* @param  None.
* @return None.
* @note   None.
******************************************************************************/
void VphyHdmiRxInitCallback(void *CallbackRef) {
	XVphy *VphyPtr = (XVphy *)CallbackRef;

	XV_HdmiRxSs_RefClockChangeInit(&HdmiRxSs);
	VphyPtr->HdmiRxTmdsClockRatio = HdmiRxSs.TMDSClockRatio;
}

/*****************************************************************************/
/**
*
* This function is called when the GT RX has been initialized.
*
* @param  None.
*
* @return None.
*
* @note   None.
*
******************************************************************************/
void VphyHdmiRxReadyCallback(void *CallbackRef) {
	XVphy *VphyPtr = (XVphy *)CallbackRef;
	XVphy_PllType RxPllType;

	/* Reset the menu to main */
	//XHdmi_MenuReset(&HdmiMenu);

	RxPllType = XVphy_GetPllType(VphyPtr, 0, XVPHY_DIR_RX, XVPHY_CHANNEL_ID_CH1);
	if (!(RxPllType == XVPHY_PLL_TYPE_CPLL)) {
		XV_HdmiRxSs_SetStream(&HdmiRxSs, VphyPtr->HdmiRxRefClkHz,
			(XVphy_GetLineRateHz(&Vphy, 0, XVPHY_CHANNEL_ID_CMN0)/1000000));

	} else {
		XV_HdmiRxSs_SetStream(&HdmiRxSs, VphyPtr->HdmiRxRefClkHz,
			 (XVphy_GetLineRateHz(&Vphy, 0, XVPHY_CHANNEL_ID_CH1)/1000000));
	}
}
#endif

/*****************************************************************************/
/**
*
* This function is called whenever an error condition in VPHY occurs.
* This will fill the FIFO of VPHY error events which will be processed outside
* the ISR.
*
* @param  CallbackRef is the VPHY instance pointer
* @param  ErrIrqType is the VPHY error type
*
* @return None.
*
* @note   None.
*
******************************************************************************/
void VphyErrorCallback(void *CallbackRef) {
	VphyErrorFlag = AVER_TRUE;
}

#if (XPAR_VPHY_0_TRANSCEIVER == XVPHY_GTXE2)
/*****************************************************************************/
/**
*
* This function is called whenever a GTXE2 PLL layout error condition in VPHY
* occurs. This function can be used automatically switch the PLL layout.
* This will set the VPHY PLL Layout error flag to AVER_TRUE.
*
* @param  CallbackRef is the VPHY instance pointer
*
* @return None.
*
* @note   None.
*
******************************************************************************/
void VphyPllLayoutErrorCallback(void *CallbackRef) {
	VphyPllLayoutErrorFlag = AVER_TRUE;
}
#endif

/*****************************************************************************/
/**
*
* This function is called in the application to process the pending
* VPHY errors
*
* @param  None.
*
* @return None.
*
* @note   This function can be expanded to perform necessary actions depending
* 		  on the error type. For example, XVPHY_ERR_PLL_LAYOUT can be used to
* 		  automatically switch in and out of bonded mode for GTXE2 devices
*
******************************************************************************/
void VphyProcessError(void) {
	if (VphyErrorFlag == AVER_TRUE) {
		debug_msg(ANSI_COLOR_RED "VPHY Error: See log for details"
				   ANSI_COLOR_RESET "\r\n");
	}
	/* Clear Flag */
	VphyErrorFlag = AVER_FALSE;

#if (XPAR_VPHY_0_TRANSCEIVER == XVPHY_GTXE2)
	if (VphyPllLayoutErrorFlag == AVER_TRUE) {
		debug_msg(ANSI_COLOR_RED "VPHY Error: Try changing to "
				   "another PLL Layout" ANSI_COLOR_RESET "\r\n");
	}
	/* Clear Flag */
	VphyPllLayoutErrorFlag = AVER_FALSE;
#endif
}


#ifdef XPAR_XV_HDMIRXSS_NUM_INSTANCES
/*****************************************************************************/
/**
*
* This function is called when a RX aux irq has occurred.
*
* @param  None.
*
* @return None.
*
* @note   None.
*
******************************************************************************/
void RxAuxCallback(void *CallbackRef) {
}

/*****************************************************************************/
/**
*
* This function is called when a RX audio irq has occurred.
*
* @param  None.
*
* @return None.
*
* @note   None.
*
******************************************************************************/
void RxAudCallback(void *CallbackRef) {
}

/*****************************************************************************/
/**
*
* This function is called when a RX link status irq has occurred.
*
* @param  None.
*
* @return None.
*
* @note   None.
*
******************************************************************************/
void RxLnkStaCallback(void *CallbackRef) {
	XV_HdmiRxSs *HdmiRxSsPtr = (XV_HdmiRxSs *)CallbackRef;

	if (IsPassThrough) {
		/* Reset RX when the link error has reached its maximum */
		if ((HdmiRxSsPtr->IsLinkStatusErrMax) &&
				(Vphy.Quads[0].Plls[0].RxState == XVPHY_GT_STATE_READY)) {

			/* Pulse RX PLL reset */
			XVphy_ClkDetFreqReset(&Vphy, 0, XVPHY_DIR_RX);
		}
	}
}

/*****************************************************************************/
/**
*
* This function is called when the RX DDC irq has occurred.
*
* @param  None.
*
* @return None.
*
* @note   None.
*
******************************************************************************/
void RxDdcCallback(void *CallbackRef) {
}

/*****************************************************************************/
/**
*
* This function is called when the RX HDCP irq has occurred.
*
* @param  None.
*
* @return None.
*
* @note   None.
*
******************************************************************************/
void RxHdcpCallback(void *CallbackRef) {
}

/*****************************************************************************/
/**
*
* This function is called when the RX stream is down.
*
* @param  None.
*
* @return None.
*
* @note   None.
*
******************************************************************************/
void RxStreamDownCallback(void *CallbackRef) {

#if(LOOPBACK_MODE_EN != 1)
#endif

#ifdef USE_HDCP
	/* Call HDCP stream-down callback */
	XHdcp_StreamDownCallback(&HdcpRepeater);
#endif

}

/*****************************************************************************/
/**
*
* This function is called when the RX stream init
*
* @param  None.
*
* @return None.
*
* @note   None.
*
******************************************************************************/
void RxStreamInitCallback(void *CallbackRef) {
	XV_HdmiRxSs *HdmiRxSsPtr = (XV_HdmiRxSs *)CallbackRef;
	XVidC_VideoStream *HdmiRxSsVidStreamPtr;
	U32_T Status;
    // debug_msg("RxStreamInitCallback\r\n");
	// Calculate RX MMCM parameters
	// In the application the YUV422 colordepth is 12 bits
	// However the HDMI transports YUV422 in 8 bits.
	// Therefore force the colordepth to 8 bits when the colorspace is YUV422

	HdmiRxSsVidStreamPtr = XV_HdmiRxSs_GetVideoStream(HdmiRxSsPtr);

	if (HdmiRxSsVidStreamPtr->ColorFormatId == XVIDC_CSF_YCRCB_422) {
		Status = XVphy_HdmiCfgCalcMmcmParam(&Vphy, 0, XVPHY_CHANNEL_ID_CH1,
											XVPHY_DIR_RX,
											HdmiRxSsVidStreamPtr->PixPerClk,
											XVIDC_BPC_8);
	}

	// Other colorspaces
	else {
		Status = XVphy_HdmiCfgCalcMmcmParam(&Vphy, 0, XVPHY_CHANNEL_ID_CH1,
											XVPHY_DIR_RX,
											HdmiRxSsVidStreamPtr->PixPerClk,
											HdmiRxSsVidStreamPtr->ColorDepth);
	}

	if (Status == XST_FAILURE) {
		return;
	}

	// Enable and configure RX MMCM
	XVphy_MmcmStart(&Vphy, 0, XVPHY_DIR_RX);

	sys_msleep(10);
}

/*****************************************************************************/
/**
*
* This function is called when the RX stream is up.
*
* @param  None.
*
* @return None.
*
* @note   None.
*
******************************************************************************/
void RxStreamUpCallback(void *CallbackRef) {
	XV_HdmiRxSs *HdmiRxSsPtr = (XV_HdmiRxSs *)CallbackRef;
	debug_msg("RX stream is up\r\n");
	XVidC_ReportStreamInfo(&HdmiRxSsPtr->HdmiRxPtr->Stream.Video);


#ifdef USE_HDCP
	/* Call HDCP stream-up callback */
	XHdcp_StreamUpCallback(&HdcpRepeater);
#endif

#ifdef VIDEO_FRAME_CRC_EN
	/* Reset Video Frame CRC */
	XVidFrameCrc_Reset();
#endif
}
#endif

/*****************************************************************************/
/**
*
* This function setups the interrupt system so interrupts can occur for the
* HDMI cores. The function is application-specific since the actual system
* may or may not have an interrupt controller. The HDMI cores could be
* directly connected to a processor without an interrupt controller.
* The user should modify this function to fit the application.
*
* @param  None.
*
* @return
*   - XST_SUCCESS if interrupt setup was successful.
*   - A specific error code defined in "xstatus.h" if an error
*   occurs.
*
* @note   This function assumes a Microblaze system and no operating
*   system is used.
*
******************************************************************************/
int SetupInterruptSystem(void) {
	int Status;
#if 0  //defined (ARMR5) || (__aarch64__) || (__arm__)
	XScuGic *IntcInstPtr = &Intc;
#else
	XIntc *IntcInstPtr = &Intc;
#endif

	/*
	 * Initialize the interrupt controller driver so that it's ready to
	 * use, specify the device ID that was generated in xparameters.h
	 */
#if 0 //defined (ARMR5) || (__aarch64__) || (__arm__)
	XScuGic_Config *IntcCfgPtr;
	IntcCfgPtr = XScuGic_LookupConfig(XPAR_SCUGIC_0_DEVICE_ID);
	if(IntcCfgPtr == NULL) {
		debug_msg("ERR:: Interrupt Controller not found");
		return (XST_DEVICE_NOT_FOUND);
	}
	Status = XScuGic_CfgInitialize(IntcInstPtr,
								   IntcCfgPtr,
								   IntcCfgPtr->CpuBaseAddress);
#else
	Status = XIntc_Initialize(IntcInstPtr, XPAR_INTC_0_DEVICE_ID);
#endif
	if (Status != XST_SUCCESS) {
		debug_msg("Intc initialization failed!\r\n");
		return XST_FAILURE;
	}

	/*
	 * Start the interrupt controller such that interrupts are recognized
	 * and handled by the processor
	 */
#if 0 //defined (__MICROBLAZE__)
	Status = XIntc_Start(IntcInstPtr, XIN_REAL_MODE);
	//Status = XIntc_Start(IntcInstPtr, XIN_SIMULATION_MODE);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}
#endif

	Xil_ExceptionInit();
	/*
	 * Register the interrupt controller handler with the exception table.
	 */
#if 0 //defined (ARMR5) || (__aarch64__) || (__arm__)
	Xil_ExceptionRegisterHandler(XIL_EXCEPTION_ID_INT,
								 (Xil_ExceptionHandler)XScuGic_InterruptHandler,
								 (XScuGic *)IntcInstPtr);
#else
	Xil_ExceptionRegisterHandler(XIL_EXCEPTION_ID_INT,
								 (Xil_ExceptionHandler)XIntc_InterruptHandler,
								 (XIntc *)IntcInstPtr);
#endif

	return (XST_SUCCESS);
}


void Xil_AssertCallbackRoutine(U8_T *File, S32_T Line) 
{
    debug_msg("Assertion in File %s, on line %0d\r\n", File, Line);
}

void Xil_AssertCallbackFunction(U8_T *File, S32_T Line) 
{
    debug_msg("Assertion in File %s, on line %0d\r\n", File, Line);
}


#if 0
static void hdmi_get_video_format(hdmi_cxt_t *hdmi_cxt,int *colorformat)
{
	U32_T InputVideoWidth;
	U32_T InputVideoHeight;
	hdmi_eResolutionType_t	InputType;

	InputVideoWidth = hdmi_cxt->frameinfo.width;
	InputVideoHeight = hdmi_cxt->frameinfo.height;

	InputType = (InputVideoHeight <= 576) ? hdmi_RES_SD : hdmi_RES_HD;

	//beacuse the hdmi only has full color range.
	if(InputType == hdmi_RES_SD)	
	{
		*colorformat = IT601_FULL_RANGE;
		hdmi_cxt->frameinfo.colorspace = hdmi_OUT_YUV601;
	}
	else
	{
		*colorformat = IT709_FULL_RANGE;
		hdmi_cxt->frameinfo.colorspace = hdmi_OUT_YUV709;
	}
}
#endif

void hdmi_set_stream(handle_t hdmi_handle, U32_T input)
{
    hdmi_cxt_t *hdmi_cxt = hdmi_handle;
    
    if(hdmi_cxt)
        hdmi_cxt->input_stream = input;  
    mesg("%s.....input=%d\n",__func__,input);
}


#define HDMI_INFOFRAME_HEADER_SIZE  3
#define HDMI_AVI_INFOFRAME_SIZE    14

#define AviInfoFrameRgbLimitedRange 0
#define AviInfoFrameRgbFullRange 1
#define AviInfoFrameUnknownRgbRange 2

eAVER_COLORIMETRY GetAVerColormetryFromAviInfoFrame(unsigned char *pBuf)
{
    unsigned char buf[255];
    unsigned char *ptr = buf;
    static unsigned int LengthOffset = 2, nDataByte1Offset = 4;

    if (!pBuf)
    {
        return 22;
    }

    mem_model_memset(ptr, 0, 255);
    mem_model_memcpy(ptr, ((unsigned char *)pBuf), HDMI_INFOFRAME_HEADER_SIZE);
    mem_model_memcpy(ptr + HDMI_INFOFRAME_HEADER_SIZE, ((unsigned char *)pBuf) + nDataByte1Offset, ptr[LengthOffset]);
    ptr += HDMI_INFOFRAME_HEADER_SIZE;

    enum hdmi_colorspace colorspace = (enum hdmi_colorspace) ((ptr[0] >> 5) & 0x7);
    enum hdmi_colorimetry colorimetry = (enum hdmi_colorimetry) ((ptr[1] >> 6) & 0x3);
    enum hdmi_extended_colorimetry extended_colorimetry = (enum hdmi_extended_colorimetry) ((ptr[2] >> 4) & 0x7);
    unsigned char video_code = ptr[3] & 0xff;
    enum hdmi_quantization_range quantization_range = (enum hdmi_quantization_range) ((ptr[2] >> 2) & 0x3);
    enum hdmi_ycc_quantization_range ycc_quantization_range = (enum hdmi_ycc_quantization_range) ((ptr[4] >> 6) & 0x3);
    enum hdmi_additional_colorimetry_extension ace = (enum hdmi_additional_colorimetry_extension) ((ptr[13] >> 4) & 0xf);

    eAVER_COLORIMETRY tempcolorimetry;
    if (colorspace == HDMI_COLORSPACE_RGB)
    {
        if (colorimetry == HDMI_COLORIMETRY_NONE)
        {
            tempcolorimetry = AVER_COLORIMETRY_RGB_LIMITED;
        }
        else if (colorimetry == HDMI_COLORIMETRY_EXTENDED)
        {
            if (extended_colorimetry == HDMI_EXTENDED_COLORIMETRY_ADOBE_RGB)
                tempcolorimetry = AVER_COLORIMETRY_Adobe_RGB;
            else if (extended_colorimetry == HDMI_EXTENDED_COLORIMETRY_BT2020)
                tempcolorimetry = AVER_COLORIMETRY_BT2020_RGB;
            else if (extended_colorimetry == HDMI_EXTENDED_COLORIMETRY_RESERVED)
            {
                if (ace == HDMI_ACE_DCI_P3_RGB_D65)
                    tempcolorimetry = AVER_COLORIMETRY_DCI_P3_D65;
                else if (ace == HDMI_ACE_DCI_P3_RGB_THEATER)
                    tempcolorimetry = AVER_COLORIMETRY_DCI_P3_Theater;
                else tempcolorimetry = AVER_COLORIMETRY_UNKNOWN;
            }
            else tempcolorimetry = AVER_COLORIMETRY_UNKNOWN;
        }
        else tempcolorimetry = AVER_COLORIMETRY_UNKNOWN;

        // return RGB colorimetry
        if (tempcolorimetry != AVER_COLORIMETRY_UNKNOWN)
        {
            int range = AviInfoFrameUnknownRgbRange;
            if (colorspace == HDMI_COLORSPACE_RGB)
            {
                switch (quantization_range)
                {
                    case HDMI_QUANTIZATION_RANGE_DEFAULT:
                        if (video_code > 1) //CE format
                            range = AviInfoFrameRgbLimitedRange; //RGB_LIMITED
                        else //IT format
                            range = AviInfoFrameRgbFullRange; //RGB_FULL
                    break;
                    case HDMI_QUANTIZATION_RANGE_LIMITED:
                        range = AviInfoFrameRgbLimitedRange; //RGB_LIMITED
                    break;
                    case HDMI_QUANTIZATION_RANGE_FULL:
                        range = AviInfoFrameRgbFullRange; //RGB_FULL
                    break;
                    default:
                    break;
                }
            }

            if (range == AviInfoFrameRgbLimitedRange)
                return tempcolorimetry;
            else if (range == AviInfoFrameRgbFullRange)
                return (eAVER_COLORIMETRY)(AVER_COLORIMETRY_FULL_RANGE + tempcolorimetry);
            else return AVER_COLORIMETRY_UNKNOWN;
        }
        else return AVER_COLORIMETRY_UNKNOWN;
    }
    else
    {
        if (colorimetry == HDMI_COLORIMETRY_NONE)
        {
            if (video_code == 1 ||			//640x480p59.94
                video_code == 2 || video_code == 3 ||	//720x480p59.94
                video_code == 6 || video_code == 7 ||	//1440x480i59.94
                video_code == 8 || video_code == 9 ||	//1440x240p60
                video_code == 10 || video_code == 11 ||	//2880x480i59.94
                video_code == 12 || video_code == 13 ||	//2880x240p60
                video_code == 14 || video_code == 15 ||	//1440x480p59.94
                video_code == 17 || video_code == 18 ||	//720x576p50
                video_code == 21 || video_code == 22 ||	//1440x576i50
                video_code == 23 || video_code == 24 ||	//1440x288p50
                video_code == 25 || video_code == 26 ||	//2880x576i50
                video_code == 27 || video_code == 28 ||	//2880x288p50
                video_code == 29 || video_code == 30 ||	//1440x576p50
                video_code == 35 || video_code == 36 ||	//2880x480p59.94
                video_code == 37 || video_code == 38 ||	//2880x576p50
                video_code == 42 || video_code == 43 ||	//720x576p100
                video_code == 44 || video_code == 45 ||	//1440x576i100
                video_code == 48 || video_code == 49 ||	//720x480p119.88
                video_code == 50 || video_code == 51 ||	//1440x480i119.88
                video_code == 52 || video_code == 53 ||	//720x576p200
                video_code == 54 || video_code == 55 ||	//1440x576i200
                video_code == 56 || video_code == 57 ||	//720x480p239.76
                video_code == 58 || video_code == 59)	//1440x480i239.76
            {
                tempcolorimetry = AVER_COLORIMETRY_YUV_601;
            }
            else
            {
                tempcolorimetry = AVER_COLORIMETRY_YUV_709;
            }
        }
        else if (colorimetry == HDMI_COLORIMETRY_ITU_601)
        {
            tempcolorimetry = AVER_COLORIMETRY_YUV_601;
        }
        else if (colorimetry == HDMI_COLORIMETRY_ITU_709)
        {
            tempcolorimetry = AVER_COLORIMETRY_YUV_709;
        }
        else if (colorimetry == HDMI_COLORIMETRY_EXTENDED)
        {
            if (extended_colorimetry == HDMI_EXTENDED_COLORIMETRY_XV_YCC_601)
                tempcolorimetry = AVER_COLORIMETRY_xvYCC_601;
            else if (extended_colorimetry == HDMI_EXTENDED_COLORIMETRY_XV_YCC_709)
                tempcolorimetry = AVER_COLORIMETRY_xvYCC_709;
            else if (extended_colorimetry == HDMI_EXTENDED_COLORIMETRY_S_YCC_601)
                tempcolorimetry = AVER_COLORIMETRY_sYCC_601;
            else if (extended_colorimetry == HDMI_EXTENDED_COLORIMETRY_ADOBE_YCC_601)
                tempcolorimetry = AVER_COLORIMETRY_Adobe_YCC_601;
            else if (extended_colorimetry == HDMI_EXTENDED_COLORIMETRY_BT2020_CONST_LUM)
                tempcolorimetry = AVER_COLORIMETRY_BT2020_cYCC;
            else if (extended_colorimetry == HDMI_EXTENDED_COLORIMETRY_BT2020)
                tempcolorimetry = AVER_COLORIMETRY_BT2020_YCC;
            else tempcolorimetry = AVER_COLORIMETRY_UNKNOWN;

        }
        else tempcolorimetry = AVER_COLORIMETRY_UNKNOWN;

        // return YUV colorimetry
        if (tempcolorimetry != AVER_COLORIMETRY_UNKNOWN)
        {
            if (ycc_quantization_range == HDMI_YCC_QUANTIZATION_RANGE_LIMITED)
                return tempcolorimetry;
            else if (ycc_quantization_range == HDMI_YCC_QUANTIZATION_RANGE_FULL)
                return (eAVER_COLORIMETRY)(AVER_COLORIMETRY_FULL_RANGE + tempcolorimetry);
            else return AVER_COLORIMETRY_UNKNOWN;
        }
        else return AVER_COLORIMETRY_UNKNOWN;
    }
}


int ConvertAVerColorimetry(eAVER_COLORIMETRY aver_color)
{
    int colorspace = 0;
    switch (aver_color)
    {
    case AVER_COLORIMETRY_RGB_LIMITED:
        colorspace = 0;
        break;
    case AVER_COLORIMETRY_RGB_FULL:
        colorspace = 1;
        break;
    case AVER_COLORIMETRY_YUV_601:
        colorspace = 2;
        break;
    case AVER_COLORIMETRY_YUV_709:
        colorspace = 3;
        break;
    case AVER_COLORIMETRY_xvYCC_601:
        colorspace = 4;
        break;
    case AVER_COLORIMETRY_xvYCC_709:
        colorspace = 5;
        break;
    case AVER_COLORIMETRY_YUV_601_FULL:
        colorspace = 6;
        break;
    case AVER_COLORIMETRY_YUV_709_FULL:
        colorspace = 7;
        break;
    case AVER_COLORIMETRY_sYCC_601:
    case AVER_COLORIMETRY_sYCC_601_FULL:
        colorspace = 8;
        break;
    case AVER_COLORIMETRY_Adobe_YCC_601:
    case AVER_COLORIMETRY_Adobe_YCC_601_FULL:
        colorspace = 9;
        break;
    case AVER_COLORIMETRY_Adobe_RGB:
    case AVER_COLORIMETRY_Adobe_RGB_FULL:
    default:
        colorspace = 10;
    }

    return colorspace;
}


static void hdmi_update_video_timing(hdmi_cxt_t *hdmi_cxt,int ch)
{ 
    if(hdmi_cxt)
    {
        hdmi_cxt->frameinfo.width = hdmi_cxt->packetinfo.m_PacketWidth;
        hdmi_cxt->frameinfo.height = hdmi_cxt->packetinfo.m_PacketHeight;
        hdmi_cxt->frameinfo.is_interlace = hdmi_cxt->packetinfo.m_PacketInterlace;
        hdmi_cxt->frameinfo.framerate = (hdmi_cxt->packetinfo.m_PacketFrameRate+10)/100; //5994
        hdmi_cxt->frameinfo.denominator = hdmi_cxt->packetinfo.m_PacketFrameRate;
        //hdmi_cxt->frameinfo.denominator = (freq/(htotal*vtotal))*1000;
        if (hdmi_cxt->print_cnt ==0)
        {
            mesg("=====%s frameinfo.denominator %d\n",__func__,hdmi_cxt->frameinfo.denominator);   
            hdmi_cxt->print_cnt = 1;
        }
    }
}


static void hdmi_task(void *data)
{
    hdmi_cxt_t *hdmi_cxt=data;
   
    int data_tmp=0x00;
    int data_tmp1=0x00;
    int data_tmp2=0x00;
    int data_tmp3=0x00;
    //int ret;
    int ch, temp;
    unsigned char Index;
    
    XV_HdmiRx *InstancePtr = HdmiRxSs.HdmiRxPtr;
    XVidC_VideoStream* HdmiRxSsVidStreamPtr;


     //XV_HdmiTxSS_HdmiTxIntrHandler(&HdmiTxSs);
 
     XV_HdmiRxSS_HdmiRxIntrHandler(&HdmiRxSs);
 
    //XVphy_InterruptHandler(&Vphy);
    XIntc_InterruptHandler(&Intc);

    /* Read header word and update AUX header field */
    InstancePtr->Aux.Header.Data = XV_HdmiRx_ReadReg(InstancePtr->Config.BaseAddress, (XV_HDMIRX_AUX_DAT_OFFSET));

    for ( Index = 0x0; Index < 8; Index++) 
    {
        /* Read data word and update AUX data field */
        InstancePtr->Aux.Data.Data[Index] = XV_HdmiRx_ReadReg(InstancePtr->Config.BaseAddress, (XV_HDMIRX_AUX_DAT_OFFSET));
    }    

#ifdef USE_HDCP
    if (XV_HdmiRxSs_HdcpIsReady(&HdmiRxSs)) 
    {
        /* Poll HDCP */
        XHdcp_Poll(&HdcpRepeater);
    }
#endif

    /* VPHY error */
    VphyProcessError();

    HdmiRxSsVidStreamPtr = XV_HdmiRxSs_GetVideoStream(&HdmiRxSs);
    if (HdmiRxSsVidStreamPtr && HdmiRxSs.HdmiRxPtr)
    {
        // do not call this function under Dispatch level
        unsigned int packet_frame_rate = 0, packet_interlace = 0, packet_audio_sampling_bits=0, packet_audio_sampling_Freq=0, packet_colorspace = 0, packet_vic = 0;

        if (HdmiRxSs.IsStreamUp)
            hdmi_cxt->is_lock = AVER_TRUE;
        else
            hdmi_cxt->is_lock = AVER_FALSE;    
        hdmi_cxt->packetinfo.m_PacketWidth = HdmiRxSsVidStreamPtr->Timing.HActive; 
        hdmi_cxt->packetinfo.m_PacketHeight = HdmiRxSsVidStreamPtr->Timing.VActive; 
        //mesg("xxDevice: lock: %d (XHDMI-vfmt) width:%d height:%d",  hdmi_cxt->is_lock, hdmi_cxt->packetinfo.m_PacketWidth, hdmi_cxt->packetinfo.m_PacketHeight);
        hdmi_cxt->packetinfo.m_PacketInterlace = HdmiRxSsVidStreamPtr->IsInterlaced;
        //mesg("xxDevice: Interlace: %s (XHDMI-vfmt)",  HdmiRxSsVidStreamPtr->IsInterlaced ? "packet_interlace" : "progressive");

        unsigned int dwordData = XV_HdmiRx_In32(XHDMI_VIN_FRAME_RATE_CNT);
        packet_frame_rate = dwordData ? (2700000000 / dwordData) + 1 : 0;
        if (packet_interlace)
            packet_frame_rate = packet_frame_rate / 2;

        if (packet_frame_rate >= 11900 && packet_frame_rate <= 12100)
            packet_frame_rate = 12000; //119~121
        else if (packet_frame_rate >= 9900 && packet_frame_rate <= 10100)
            packet_frame_rate = 10000; //99~100
        else if (packet_frame_rate >= 8400 && packet_frame_rate <= 8600)
            packet_frame_rate = 8500; //84~86
        else if (packet_frame_rate >= 7400 && packet_frame_rate <= 7600)
            packet_frame_rate = 7500; //74~76
        else if (packet_frame_rate >= 7100 && packet_frame_rate <= 7300)
            packet_frame_rate = 7200; //71~73
        else if (packet_frame_rate >= 6900 && packet_frame_rate <= 7099)
            packet_frame_rate = 7000; //69~70.99
        else if (packet_frame_rate >= 5997 && packet_frame_rate <= 6100)
            packet_frame_rate = 6000; //59.97~61
        else if (packet_frame_rate >= 5900 && packet_frame_rate <= 5996)
            packet_frame_rate = 5994; //59~59.96
        else if (packet_frame_rate >= 5500 && packet_frame_rate <= 5700)
            packet_frame_rate = 5600; //55~57
        else if (packet_frame_rate >= 4900 && packet_frame_rate <= 5100)
            packet_frame_rate = 5000; //49~51
        else if (packet_frame_rate >= 2999 && packet_frame_rate <= 3100)
            packet_frame_rate = 3000; //29.99~31
        else if (packet_frame_rate >= 2900 && packet_frame_rate <= 2998)
            packet_frame_rate = 2997; //29~29.98
        else if (packet_frame_rate >= 2450 && packet_frame_rate <= 2600)
            packet_frame_rate = 2500; //24.5~26
        else if (packet_frame_rate >= 2399 && packet_frame_rate <= 2449)
            packet_frame_rate = 2400; //23.99~24.49
        else if (packet_frame_rate >= 2300 && packet_frame_rate <= 2398)
            packet_frame_rate = 2398; //23~23.98
        hdmi_cxt->packetinfo.m_PacketFrameRate = packet_frame_rate;
        //mesg("xxDevice:  frame rate: %d (XHDMI-vfmt)", hdmi_cxt->packetinfo.m_PacketFrameRate);

        //copy InfoFrame
        U8_T tempAviInfoFrame[32];
        U8_T index = HdmiRxSs.HdmiRxPtr->Aux.Header.Byte[0] & 0x0f;
        if ((HdmiRxSs.HdmiRxPtr->Aux.Header.Byte[0] > 0x80)
                && (HdmiRxSs.HdmiRxPtr->Aux.Header.Byte[0] < 0x90) && (index < 8))
        {
            mem_model_memset(tempAviInfoFrame, 0, 32);
            mem_model_memcpy(tempAviInfoFrame, HdmiRxSs.HdmiRxPtr->Aux.Header.Byte, 3);//31
            mem_model_memcpy(tempAviInfoFrame + 3, &HdmiRxSs.HdmiRxPtr->Aux.Data.Byte[0], 7);//28
            mem_model_memcpy(tempAviInfoFrame + 3 + 7, &HdmiRxSs.HdmiRxPtr->Aux.Data.Byte[8], 7);//21
            mem_model_memcpy(tempAviInfoFrame + 3 + 7 + 7, &HdmiRxSs.HdmiRxPtr->Aux.Data.Byte[16], 7);//14
            mem_model_memcpy(tempAviInfoFrame + 3 + 7 + 7 + 7, &HdmiRxSs.HdmiRxPtr->Aux.Data.Byte[24], 7);//7
            mem_model_memcpy(&InfoFrame[index][0], tempAviInfoFrame, 32);
        }
        if (!HdmiRxSs.IsStreamUp)
            mem_model_memset(&InfoFrame[2][0], 0, 32);
        // Read AVI InfoFrame
        U8_T *pAviInfoFrame = &InfoFrame[2][0];

        eAVER_COLORIMETRY aver_color = GetAVerColormetryFromAviInfoFrame(pAviInfoFrame);
        if (AVER_COLORIMETRY_UNKNOWN != aver_color)
        {
        	if(AVER_COLORIMETRY_RGB_LIMITED == aver_color)
        	{
        		packet_colorspace = 0;
        	}
        	else if(AVER_COLORIMETRY_RGB_FULL == aver_color)
        	{
        		packet_colorspace = 1;
        	}
        	else if(AVER_COLORIMETRY_YUV_601 == aver_color)
        	{
        		packet_colorspace = 2;
        	}
        	else if(AVER_COLORIMETRY_YUV_709 == aver_color)
        	{
        		packet_colorspace = 3;
        	}
        	else if(AVER_COLORIMETRY_xvYCC_601 == aver_color)
        	{
        		packet_colorspace = 4;
        	}
        	else if(AVER_COLORIMETRY_xvYCC_709 == aver_color)
        	{
        		packet_colorspace = 5;
        	}
        	else if(AVER_COLORIMETRY_YUV_601_FULL == aver_color)
        	{
        		packet_colorspace = 6;
        	}
        	else if(AVER_COLORIMETRY_YUV_709_FULL == aver_color)
        	{
        		packet_colorspace = 7;
        	}
        	else if((AVER_COLORIMETRY_sYCC_601 == aver_color) || (AVER_COLORIMETRY_sYCC_601_FULL == aver_color))
        	{
        		packet_colorspace = 8;
        	}
        	else if((AVER_COLORIMETRY_Adobe_YCC_601 == aver_color) || (AVER_COLORIMETRY_Adobe_YCC_601_FULL == aver_color))
        	{
        		packet_colorspace = 9;
        	}
        	else if((AVER_COLORIMETRY_Adobe_RGB == aver_color) || (AVER_COLORIMETRY_Adobe_RGB_FULL == aver_color))
        	{
        		packet_colorspace = 10;
        	}
        	else
        	{
        		packet_colorspace = 10;
        	}
        }
        else
        {
            packet_colorspace = 1; //default RGB Full
        }
#if 0
        mesg("xxDevice ColorSpace %s (XHDMI-vfmt)", 
            packet_colorspace == 0 ? "RGB_LIMITED" :
            packet_colorspace == 1 ? "RGB_FULL" :
            packet_colorspace == 2 ? "YUV_601" :
            packet_colorspace == 3 ? "YUV_709" :
            packet_colorspace == 4 ? "XVYCC_601" :
            packet_colorspace == 5 ? "XVYCC_709" :
            packet_colorspace == 6 ? "YUV_601_FULL" :
            packet_colorspace == 7 ? "YUV_709_FULL" :
            packet_colorspace == 8 ? "sYCC 601" :
            packet_colorspace == 9 ? "Adobe YCC 601" : "Adobe RGB");
#endif

        packet_vic = (pAviInfoFrame[7] & 0x7f);
        hdmi_cxt->packetinfo.m_SamplingMode = (packet_colorspace < 2 || packet_colorspace == 10) ? 0
            : (HdmiRxSs.HdmiRxPtr->Stream.Video.ColorFormatId == XVIDC_CSF_RGB) ? 0
            : (HdmiRxSs.HdmiRxPtr->Stream.Video.ColorFormatId == XVIDC_CSF_YCRCB_422) ? 1
            : (HdmiRxSs.HdmiRxPtr->Stream.Video.ColorFormatId == XVIDC_CSF_YCRCB_444) ? 2
            : (HdmiRxSs.HdmiRxPtr->Stream.Video.ColorFormatId == XVIDC_CSF_YCRCB_420) ? 3
            : 0;
#if 0
        mesg("xxDevice VIC: %d (XHDMI-vfmt)", packet_vic);
        mesg("xxDevice Sampling_mode: %s (XHDMI-vfmt)", hdmi_cxt->packetinfo.m_SamplingMode == VIDEO_RGB_MODE ? "RGB" :
            hdmi_cxt->packetinfo.m_SamplingMode == VIDEO_444_MODE ? "YUV444" : hdmi_cxt->packetinfo.m_SamplingMode == VIDEO_422_MODE ? "YUV422" :
            hdmi_cxt->packetinfo.m_SamplingMode == VIDEO_420_MODE ? "YUV420" : "Unknown sample mode!");
#endif

            if ((hdmi_cxt->packetinfo.m_PacketWidth >= 1920 && hdmi_cxt->packetinfo.m_PacketHeight >= 2160))  // 4k mode could be 2k2k or TMDS clock over 170MHz  
                hdmi_cxt->packetinfo.m_WorkingMode = DEVICE_4k2k_MODE;
            else if ((hdmi_cxt->packetinfo.m_PacketWidth <= 720) && (hdmi_cxt->packetinfo.m_PacketHeight <= 576))
                hdmi_cxt->packetinfo.m_WorkingMode = DEVICE_480_MODE;
            else
                hdmi_cxt->packetinfo.m_WorkingMode = DEVICE_720_MODE;

            if (packet_colorspace == 0)
            {
                hdmi_cxt->packetinfo.m_ColorSpaceMode = VIDEO_RGB_LIMIT_FMT;
            }
            else if((packet_colorspace == 1) || (packet_colorspace == 10))
            {
                hdmi_cxt->packetinfo.m_ColorSpaceMode = VIDEO_RGB_FULL_FMT;
            }else{
                hdmi_cxt->packetinfo.m_ColorSpaceMode = VIDEO_YUV_FMT;
            }

        //no used
        unsigned int CsData[6];
        unsigned int addr = 0x38, i;
        for ( i = 0; i < 6; i++)
        {
            CsData[i] = (unsigned int) Xil_In32(addr);
            addr += 4;
        }

        hdmi_cxt->audio_sample = hdmi_AUDIO_RATE_48000;
        unsigned char AudioFreq = (CsData[0] >> 24) & 0xf;
        if (AudioFreq == 0)
        {
            packet_audio_sampling_Freq = 44100;
            hdmi_cxt->audio_sample = hdmi_AUDIO_RATE_44100;
        }
        else if (AudioFreq == 2)
        {
            packet_audio_sampling_Freq = 48000;
            hdmi_cxt->audio_sample = hdmi_AUDIO_RATE_48000;
        }
        else if (AudioFreq == 3)
        {
            packet_audio_sampling_Freq = 32000;
            hdmi_cxt->audio_sample = hdmi_AUDIO_RATE_32000;
        }

        if(hdmi_cxt->old_audio_sample != hdmi_cxt->audio_sample)
        {
            hdmi_cxt->old_audio_sample = hdmi_cxt->audio_sample;
            hdmi_callback(hdmi_cxt,HDMI_AUDIO_CHG);
        }


#if 1
        unsigned char FieldSize = (CsData[1] & 0x1);
        unsigned char WordLength = ((CsData[1] >> 1) & 0x7);
        if (WordLength == 5)
        {
            packet_audio_sampling_bits = FieldSize ? 24 : 20;
        }
        else if (WordLength == 4)
        {
            packet_audio_sampling_bits = FieldSize ? 23 : 19;
        }
        else if (WordLength == 2)
        {
            packet_audio_sampling_bits = FieldSize ? 22 : 18;
        }
        else if (WordLength == 6)
        {
            packet_audio_sampling_bits = FieldSize ? 21 : 17;
        }
        else if (WordLength == 1)
        {
            packet_audio_sampling_bits = FieldSize ? 20 : 16;
        }
#endif
     

#if 1
        if (hdmi_cxt->packetinfo.m_PacketVideoLock != hdmi_cxt->packetinfo.m_Old_PacketVideoLock
                || hdmi_cxt->packetinfo.m_PacketWidth != hdmi_cxt->packetinfo.m_Old_PacketWidth
                || hdmi_cxt->packetinfo.m_PacketHeight != hdmi_cxt->packetinfo.m_Old_PacketHeight
                || hdmi_cxt->packetinfo.m_PacketInterlace != hdmi_cxt->packetinfo.m_Old_PacketInterlace
                || hdmi_cxt->packetinfo.m_PacketFrameRate != hdmi_cxt->packetinfo.m_Old_PacketFrameRate
                || hdmi_cxt->packetinfo.m_ColorSpaceMode != hdmi_cxt->packetinfo.m_Old_ColorSpaceMode
                || hdmi_cxt->packetinfo.m_SamplingMode != hdmi_cxt->packetinfo.m_Old_SamplingMode)
        {
            if (hdmi_cxt->is_state_change == AVER_FALSE)
            {
                mesg("\n");
                mesg("hdmi state change to UNLOCK\n");
            }

            hdmi_cxt->is_state_change = AVER_TRUE;

            #if 1
            if (hdmi_cxt->is_lock_state == AVER_TRUE)
            {
                hdmi_callback(hdmi_cxt,HDMI_UNLOCK);
                hdmi_cxt->is_lock_state = AVER_FALSE;
                hdmi_cxt->print_cnt = 0;
            }
            #endif

            #if 1
            hdmi_cxt->is_lock = AVER_FALSE;
            #endif

            hdmi_cxt->packetinfo.m_Old_PacketVideoLock = hdmi_cxt->packetinfo.m_PacketVideoLock;
            hdmi_cxt->packetinfo.m_Old_PacketWidth = hdmi_cxt->packetinfo.m_PacketWidth;
            hdmi_cxt->packetinfo.m_Old_PacketHeight = hdmi_cxt->packetinfo.m_PacketHeight;
            hdmi_cxt->packetinfo.m_Old_PacketInterlace = hdmi_cxt->packetinfo.m_PacketInterlace;
            hdmi_cxt->packetinfo.m_Old_PacketFrameRate = hdmi_cxt->packetinfo.m_PacketFrameRate;
            hdmi_cxt->packetinfo.m_Old_ColorSpaceMode = hdmi_cxt->packetinfo.m_ColorSpaceMode;
            hdmi_cxt->packetinfo.m_Old_SamplingMode = hdmi_cxt->packetinfo.m_SamplingMode;
        }
#endif
			
#if 1			
        if (hdmi_cxt->is_lock == AVER_TRUE)
        {
            hdmi_update_video_timing(hdmi_cxt,ch);
			
            if (hdmi_cxt->is_state_change == AVER_TRUE)
            {
                //hdmi_set_audio_format(hdmi_cxt,audioformat);
                mesg("\n");
                mesg("hdmi  state change to LOCK\n");
                mesg("hdmi  m_PacketWidth %d\n", hdmi_cxt->packetinfo.m_PacketWidth);
                mesg("hdmi  m_PacketHeight %d\n", hdmi_cxt->packetinfo.m_PacketHeight); 
                mesg("hdmi  m_PacketInterlace %s", hdmi_cxt->packetinfo.m_PacketInterlace ? "Interlace" : "Progressive");    
                mesg(" xxDevice Audio sample rate: %d Hz(XHDMI-vfmt)\n",  packet_audio_sampling_Freq);
                mesg(" xxDevice Audio Bit count: %d (XHDMI-vfmt)\n", packet_audio_sampling_bits);
                #if 1   
                mesg("hdmi  m_ColorSpaceMode:(%d): %s", hdmi_cxt->packetinfo.m_ColorSpaceMode,
                        hdmi_cxt->packetinfo.m_ColorSpaceMode == VIDEO_RGB_FULL_FMT ? "VIDEO_RGB_FULL_FMT" :
                        hdmi_cxt->packetinfo.m_ColorSpaceMode == VIDEO_YUV_FMT ? "VIDEO_YUV_FMT" : "UNKNOWN");
			
                mesg("hdmi  m_SamplingMode(%d): %s", hdmi_cxt->packetinfo.m_SamplingMode,
                        hdmi_cxt->packetinfo.m_SamplingMode == VIDEO_RGB_MODE ? "VIDEO_RGB_MODE" :
                        hdmi_cxt->packetinfo.m_SamplingMode == VIDEO_422_MODE ? "VIDEO_422_MODE" :
                        hdmi_cxt->packetinfo.m_SamplingMode == VIDEO_444_MODE ? "VIDEO_444_MODE" : "UNKNOWN");
                #endif
			    
                hdmi_cxt->is_state_change = AVER_FALSE;
                hdmi_callback(hdmi_cxt,HDMI_LOCK);
                hdmi_cxt->is_lock_state = AVER_TRUE;
            }
        }
        else
        {
            if(hdmi_cxt->is_state_change == AVER_FALSE)
            {
                mesg("hdmi unlock status...\n");
                hdmi_cxt->is_state_change = AVER_TRUE;
                hdmi_callback(hdmi_cxt,HDMI_UNLOCK);
                hdmi_cxt->is_lock_state = AVER_FALSE;
                hdmi_cxt->print_cnt = 0;
            }
        }
#endif  	
    }
}

static void hdmi_yuv2rgb_coeff(hdmi_cxt_t *hdmi_cxt,int ch, int CoeffMode)
{
    if(hdmi_cxt)
    { 
	U32_T data;
        mesg("hdmi_yuv2rgb_coeff mode:%d\n", CoeffMode);
	data = YUVtoRGB_TABLE[g_y_coef][CoeffMode] + (YUVtoRGB_TABLE[g_y_coef_s][CoeffMode] << 16);
	Xil_Out32(V_Y2R_G_Y_COEF(XHDMI_CSC_OFFSET), data);

	data = YUVtoRGB_TABLE[g_cb_coef][CoeffMode] + (YUVtoRGB_TABLE[g_cb_coef_s][CoeffMode] << 16);
	Xil_Out32(V_Y2R_G_CB_COEF(XHDMI_CSC_OFFSET), data);

	data = YUVtoRGB_TABLE[g_cr_coef][CoeffMode] + (YUVtoRGB_TABLE[g_cr_coef_s][CoeffMode] << 16);
	Xil_Out32(V_Y2R_G_CR_COEF(XHDMI_CSC_OFFSET), data);

	data = YUVtoRGB_TABLE[y2_offset][CoeffMode] + (YUVtoRGB_TABLE[y2_offset_s][CoeffMode] << 16);
	Xil_Out32(V_Y2R_YY_OFFSET(XHDMI_CSC_OFFSET), data);

	data = YUVtoRGB_TABLE[b_y_coef][CoeffMode] + (YUVtoRGB_TABLE[b_y_coef_s][CoeffMode] << 16);
	Xil_Out32(V_Y2R_B_Y_COEF(XHDMI_CSC_OFFSET), data);

	data = YUVtoRGB_TABLE[b_cb_coef][CoeffMode] + (YUVtoRGB_TABLE[b_cb_coef_s][CoeffMode] << 16);
	Xil_Out32(V_Y2R_B_CB_COEF(XHDMI_CSC_OFFSET), data);

	data = YUVtoRGB_TABLE[b_cr_coef][CoeffMode] + (YUVtoRGB_TABLE[b_cr_coef_s][CoeffMode] << 16);
	Xil_Out32(V_Y2R_B_CR_COEF(XHDMI_CSC_OFFSET), data);

	data = YUVtoRGB_TABLE[cb2_offset][CoeffMode] + (YUVtoRGB_TABLE[cb2_offset_s][CoeffMode] << 16);
	Xil_Out32(V_Y2R_CCB_OFFSET(XHDMI_CSC_OFFSET), data);

	data = YUVtoRGB_TABLE[r_y_coef][CoeffMode] + (YUVtoRGB_TABLE[r_y_coef_s][CoeffMode] << 16);
	Xil_Out32(V_Y2R_R_Y_COEF(XHDMI_CSC_OFFSET), data);

	data = YUVtoRGB_TABLE[r_cb_coef][CoeffMode] + (YUVtoRGB_TABLE[r_cb_coef_s][CoeffMode] << 16);
	Xil_Out32(V_Y2R_R_CB_COEF(XHDMI_CSC_OFFSET), data);

	data = YUVtoRGB_TABLE[r_cr_coef][CoeffMode] + (YUVtoRGB_TABLE[r_cr_coef_s][CoeffMode] << 16);
	Xil_Out32(V_Y2R_R_CR_COEF(XHDMI_CSC_OFFSET), data);

	data = YUVtoRGB_TABLE[cr2_offset][CoeffMode] + (YUVtoRGB_TABLE[cr2_offset_s][CoeffMode] << 16);
	Xil_Out32(V_Y2R_CCR_OFFSET(XHDMI_CSC_OFFSET), data);
    }
}

static void hdmi_rgb2yuv_coeff(hdmi_cxt_t *hdmi_cxt,int ch, int CoeffMode)
{
    if(hdmi_cxt)
    {
	U32_T data;
        mesg("hdmi_rgb2yuv_coeff mode:%d \n", CoeffMode);

	data = RGBtoYUV_TABLE[y_r_coef][CoeffMode] + (RGBtoYUV_TABLE[y_r_coef_s][CoeffMode] << 16);
	Xil_Out32(V_R2Y_Y_R_COEF(XHDMI_CSC_OFFSET), data);

	data = RGBtoYUV_TABLE[y_g_coef][CoeffMode] + (RGBtoYUV_TABLE[y_g_coef_s][CoeffMode] << 16);
	Xil_Out32(V_R2Y_Y_G_COEF(XHDMI_CSC_OFFSET), data);

	data = RGBtoYUV_TABLE[y_b_coef][CoeffMode] + (RGBtoYUV_TABLE[y_b_coef_s][CoeffMode] << 16);
	Xil_Out32(V_R2Y_Y_B_COEF(XHDMI_CSC_OFFSET), data);

	data = RGBtoYUV_TABLE[y_offset][CoeffMode] + (RGBtoYUV_TABLE[y_offset_s][CoeffMode] << 16);
	Xil_Out32(V_R2Y_Y_OFFSET(XHDMI_CSC_OFFSET), data);

	data = RGBtoYUV_TABLE[cb_r_coef][CoeffMode] + (RGBtoYUV_TABLE[cb_r_coef_s][CoeffMode] << 16);
	Xil_Out32(V_R2Y_CB_R_COEF(XHDMI_CSC_OFFSET), data);

	data = RGBtoYUV_TABLE[cb_g_coef][CoeffMode] + (RGBtoYUV_TABLE[cb_g_coef_s][CoeffMode] << 16);
	Xil_Out32(V_R2Y_CB_G_COEF(XHDMI_CSC_OFFSET), data);

	data = RGBtoYUV_TABLE[cb_b_coef][CoeffMode] + (RGBtoYUV_TABLE[cb_b_coef_s][CoeffMode] << 16);
	Xil_Out32(V_R2Y_CB_B_COEF(XHDMI_CSC_OFFSET), data);

	data = RGBtoYUV_TABLE[cb_offset][CoeffMode] + (RGBtoYUV_TABLE[cb_offset_s][CoeffMode] << 16);
	Xil_Out32(V_R2Y_CB_OFFSET(XHDMI_CSC_OFFSET), data);

	data = RGBtoYUV_TABLE[cr_r_coef][CoeffMode] + (RGBtoYUV_TABLE[cr_r_coef_s][CoeffMode] << 16);
	Xil_Out32(V_R2Y_CR_R_COEF(XHDMI_CSC_OFFSET), data);

	data = RGBtoYUV_TABLE[cr_g_coef][CoeffMode] + (RGBtoYUV_TABLE[cr_g_coef_s][CoeffMode] << 16);
	Xil_Out32(V_R2Y_CR_G_COEF(XHDMI_CSC_OFFSET), data);

	data = RGBtoYUV_TABLE[cr_b_coef][CoeffMode] + (RGBtoYUV_TABLE[cr_b_coef_s][CoeffMode] << 16);
	Xil_Out32(V_R2Y_CR_B_COEF(XHDMI_CSC_OFFSET), data);

	data = RGBtoYUV_TABLE[cr_offset][CoeffMode] + (RGBtoYUV_TABLE[cr_offset_s][CoeffMode] << 16);
	Xil_Out32(V_R2Y_CR_OFFSET(XHDMI_CSC_OFFSET), data);
    }
}

static void hdmi_yuv2rgb_enable(hdmi_cxt_t *hdmi_cxt,int ch, BOOL_T Enable)
{
    //hdmi_cxt_t *hdmi_cxt=hdmi_handle;
 
    if(hdmi_cxt)
    {
	U8_T data;
	data = Xil_In8(V_YUV_TO_RGB_EN(XHDMI_CSC_OFFSET));
        mesg("yuv to rgb enable:%x data:%x \n", V_YUV_TO_RGB_EN(XHDMI_CSC_OFFSET) , data);
	if (Enable)
		data |= Bit(V_YUV_TO_RGB_EN_ENABLE);
	else
		data &= (~Bit(V_YUV_TO_RGB_EN_ENABLE));
	Xil_Out8(V_YUV_TO_RGB_EN(XHDMI_CSC_OFFSET), data);
    }
}

static void hdmi_rgb2yuv_enable(hdmi_cxt_t *hdmi_cxt,int ch, BOOL_T Enable)
{
    //hdmi_cxt_t *hdmi_cxt=hdmi_handle;

    if(hdmi_cxt)
    {
	U8_T data;
	data = Xil_In8(V_RGB_TO_YUV_EN(XHDMI_CSC_OFFSET));
        mesg("rgb to yuv enable:%x data:%x \n", V_RGB_TO_YUV_EN(XHDMI_CSC_OFFSET) , data);
	if (Enable)
		data |= Bit(V_RGB_TO_YUV_EN_ENABLE);
	else
		data &= (~Bit(V_RGB_TO_YUV_EN_ENABLE));
	Xil_Out8(V_RGB_TO_YUV_EN(XHDMI_CSC_OFFSET), data); 
    }
}

void hdmi_get_frameinfo(handle_t hdmi_handle,hdmi_frameinfo_t *frameinfo) 
{
    hdmi_cxt_t *hdmi_cxt=hdmi_handle;
    
    if(hdmi_cxt)
    {
        if (hdmi_cxt->frameinfo.is_interlace)
            mesg("%s locked rx %dx%d%c%d\n",__func__,hdmi_cxt->frameinfo.width,hdmi_cxt->frameinfo.height,(hdmi_cxt->frameinfo.is_interlace)?'i':'p',2*hdmi_cxt->frameinfo.framerate);
        else
            mesg("%s locked rx %dx%d%c%d\n",__func__,hdmi_cxt->frameinfo.width,hdmi_cxt->frameinfo.height,(hdmi_cxt->frameinfo.is_interlace)?'i':'p',hdmi_cxt->frameinfo.framerate);
     
        if(frameinfo)
            *frameinfo=hdmi_cxt->frameinfo;
    }
}

#if 1
void hdmi_set_color_space(handle_t hdmi_handle,hdmi_packetinfo_t *packetinfo) 
{
    hdmi_cxt_t *hdmi_cxt=hdmi_handle;
    int ch;
    int rgb_color_adjust=0xff;
    int yuv_color_adjust=0xff;
    int data;
    
    if(hdmi_cxt)
    {
        mesg("%s...\n",__func__);
        ch = hdmi_cxt->device_index;
        #if 1
        mesg("(ch%d) m_ColorSpaceMode %s\n", ch,
                packetinfo->m_ColorSpaceMode == VIDEO_RGB_FULL_FMT ? "VIDEO_RGB_FULL_FMT" :
                packetinfo->m_ColorSpaceMode == VIDEO_YUV_FMT ? "VIDEO_YUV_FMT" : "UNKNOWN");
	
        mesg("(ch%d) m_SamplingMode %s\n", ch,
                hdmi_cxt->packetinfo.m_SamplingMode == VIDEO_RGB_MODE ? "VIDEO_RGB_MODE" :
                hdmi_cxt->packetinfo.m_SamplingMode == VIDEO_422_MODE ? "VIDEO_422_MODE" :
                hdmi_cxt->packetinfo.m_SamplingMode == VIDEO_444_MODE ? "VIDEO_444_MODE" : "UNKNOWN");
    
        mesg("(ch%d) m_WorkingMode_in %s\n", ch,
                packetinfo->m_WorkingMode_in == DEVICE_4k2k_MODE ? "DEVICE_4k2k_MODE" :
                packetinfo->m_WorkingMode_in == DEVICE_720_MODE ? "DEVICE_720_MODE" :
                packetinfo->m_WorkingMode_in == DEVICE_480_MODE ? "DEVICE_480_MODE" : "UNKNOWN"); 
		        	
        mesg("(ch%d) m_WorkingMode_out %s\n", ch,
                packetinfo->m_WorkingMode_out == DEVICE_4k2k_MODE ? "DEVICE_4k2k_MODE" :
                packetinfo->m_WorkingMode_out == DEVICE_720_MODE ? "DEVICE_720_MODE" :
                packetinfo->m_WorkingMode_out == DEVICE_480_MODE ? "DEVICE_480_MODE" : "UNKNOWN"); 	        	
	#endif
		   
        packetinfo->m_PixelMode = SINGLE_PIXEL_MODE;
        packetinfo->m_PacketRepetition = 1;
        
        mesg("(ch%d) m_PixelFmt_out = %d\n", ch, packetinfo->m_PixelFmt_out);
        
        switch (packetinfo->m_ColorSpaceMode)		// i/p color space
        {
            case VIDEO_YUV_FMT:					//i/p YUV
                //FPGA should support RGB3, but not ready now
                if ((packetinfo->m_PixelFmt_out == hdmi_RGB3) || (packetinfo->m_PixelFmt_out == hdmi_BGR3))	//o/p RGB	
                {
                    rgb_color_adjust = 0xff;
                    switch (packetinfo->m_WorkingMode_in)
                    {
                        case DEVICE_480_MODE:			// ip:640, o/p: 640/720/4k2k
                            yuv_color_adjust = BT601_to_RGBF;	// 0x2;	// na, bt601->rgb_F, assume use FLL as output
                            break;
                        case DEVICE_720_MODE:			// ip:720, o/p: 640/720/4k2k
                            yuv_color_adjust = BT709_to_RGBF;	// 0x0;// 0x2;	// na, bt709->rgb_F, assume use FLL as output
                            break;
                        case DEVICE_4k2k_MODE:			// ip:4k2k, o/p: 640/720/4k2k
                            yuv_color_adjust = BT709_to_RGBF;	// 0x0;// 0x2;	// na, bt2020->rgb_F, assume use FLL as output,	assume bt2020 as bt709
                            break;
                    }
                }else{
                    rgb_color_adjust = 0xff;//i/p yuv, o/p yuv
                    switch (packetinfo->m_WorkingMode_in)		
                    {
                        case DEVICE_480_MODE:
                            switch (packetinfo->m_WorkingMode_out)
                            {
                                case DEVICE_480_MODE:	// i/p:480, o/p: 480
                                    yuv_color_adjust = 0xff;		// na, bt601->bt601
                                    break;
                                case DEVICE_720_MODE:	// i/p:480, o/p: 720
                                    yuv_color_adjust = BT601_to_BT709;// 0x4;	// na, bt601->bt707
                                    break;
                                case DEVICE_4k2k_MODE:	// i/p:480, o/p: 4k2k
                                    yuv_color_adjust = BT601_to_BT709;// 0x4;	// na, bt601->bt 2020,	assume bt2020 as bt709
                                    break;
                            }
                            break;

                        case DEVICE_720_MODE:
                            switch (packetinfo->m_WorkingMode_out)
                            {
                                case DEVICE_480_MODE:	// i/p:720, o/p: 480
                                    yuv_color_adjust = BT709_to_BT601;// 0x5;	// na, bt709->bt601
                                    break;
                                case DEVICE_720_MODE:	// i/p:720, o/p: 720
                                    yuv_color_adjust = 0xff;		// na, bt709->bt709
                                    break;
                                case DEVICE_4k2k_MODE:	// i/p:720, o/p: 4k2k
                                    yuv_color_adjust = 0xff;		// na, bt709->bt 2020,	assume bt2020 as bt709
                                    break;
                            }
                            break;

                        case DEVICE_4k2k_MODE:  
                            switch (packetinfo->m_WorkingMode_out)
                            {
                                case DEVICE_480_MODE:	// i/p:4k2k, o/p: 480
                                    yuv_color_adjust = BT709_to_BT601;// 0x5;	// na, bt2020->bt601,	assume bt2020 as bt709
                                    break;
                                case DEVICE_720_MODE:	// i/p:4k2k, o/p: 720
                                    yuv_color_adjust = 0xff;		// na, bt2020->bt709,	assume bt2020 as bt709
                                    break;
                                case DEVICE_4k2k_MODE:	// i/p:4k2k, o/p: 4k2k
                                    yuv_color_adjust = 0xff;		// na, bt2020->bt2020,	assume bt2020 as bt709
                                    break;
                            }
                            break;
                        }
                    }
                    break;
		
            case VIDEO_RGB_FULL_FMT:
                if ((packetinfo->m_PixelFmt_out == hdmi_RGB3) || (packetinfo->m_PixelFmt_out == hdmi_BGR3))
                {	
                    // i/p RGB_F   o/p RGB
                    switch (packetinfo->m_WorkingMode_in)
                    {
                        case DEVICE_480_MODE:			// i/p:RGB_F 480, o/p: rgb_F 480/720/4k2k
                            rgb_color_adjust = 0xff;	// 0x2;		// rgb_F->bt601
                            yuv_color_adjust = 0xff;	// 0x2;		// bt601->rgb_F
                            break;
                        case DEVICE_720_MODE:			// i/p:RGB_F 720, o/p: rgb_F 480/720/4k2k
                            rgb_color_adjust = 0xff;	// 0x0;		// rgb_F->bt709
                            yuv_color_adjust = 0xff;	// 0x0;		// bt709->rgb_F
                            break;
                        case DEVICE_4k2k_MODE:			// i/p:RGB_F 4k2k, o/p: rgb_F 480/720/4k2k						
                            rgb_color_adjust = 0xff;	// 0x0;		// rgb_F->bt2020,	assume bt2020 as bt709	
                            yuv_color_adjust = 0xff;	// 0x0;		// bt2020->rgb_F		
                            break;
                    }
                }else{
                    // i/p RGB_F   o/p YUV
                    switch (packetinfo->m_WorkingMode_in)
                    {
                        case DEVICE_480_MODE:		
                            rgb_color_adjust = RGBF_to_BT601;		
                            switch (packetinfo->m_WorkingMode_out)
                            {
                                case DEVICE_480_MODE:			
                                    rgb_color_adjust = RGBF_to_BT601;		
                                    yuv_color_adjust = 0xff;				 
                                    break;
                                case DEVICE_720_MODE:			
                                    rgb_color_adjust = RGBF_to_BT709;	
                                    yuv_color_adjust = 0xff;				
                                    break;
                                case DEVICE_4k2k_MODE:			
                                    rgb_color_adjust = RGBF_to_BT709;	
                                    yuv_color_adjust = 0xff;				
                                    break;
                            }
                            break;
                        case DEVICE_720_MODE:
                            rgb_color_adjust = RGBF_to_BT709;		
                            switch (packetinfo->m_WorkingMode_out)
                            {
                                case DEVICE_480_MODE:			
                                    rgb_color_adjust = RGBF_to_BT601;		
                                    yuv_color_adjust = 0xff;					
                                    break;
                                case DEVICE_720_MODE:			
                                    rgb_color_adjust = RGBF_to_BT709;		
                                    yuv_color_adjust = 0xff;				
                                    break;
                                case DEVICE_4k2k_MODE:			
                                    rgb_color_adjust = RGBF_to_BT709;		
                                    yuv_color_adjust = 0xff;				
                                    break;
                            }
                            break;
                        case DEVICE_4k2k_MODE:						
                            switch (packetinfo->m_WorkingMode_out)
                            {
                                case DEVICE_480_MODE:			
                                    rgb_color_adjust = RGBF_to_BT601;	
                                    yuv_color_adjust = 0xff;				
                                    break;
                                case DEVICE_720_MODE:			
                                    rgb_color_adjust = RGBF_to_BT709;	
                                    yuv_color_adjust = 0xff;				
                                    break;
                                case DEVICE_4k2k_MODE:			
                                    rgb_color_adjust = RGBF_to_BT709;		
                                    yuv_color_adjust = 0xff;				
                                    break;
                            }
                            break;
                    }
                }
                break;
		
            case VIDEO_RGB_LIMIT_FMT:
                if ((packetinfo->m_PixelFmt_out == hdmi_RGB3) || (packetinfo->m_PixelFmt_out == hdmi_BGR3))
                {
                    // i/p RGB_L   o/p RGB
                    switch (packetinfo->m_WorkingMode_in)
                    {
                        case DEVICE_480_MODE:			//i/p: rgb_L 480, o/p: RGB_L 480/720/4k2k
                            rgb_color_adjust = RGBL_to_BT601;	// 0x3;		// rgb_L->bt601
                            yuv_color_adjust = BT601_to_RGBF;	// 0x2;		// bt601->rgb_L
                            break;
                        case DEVICE_720_MODE:			//i/p: rgb_L 720, o/p: RGB_L 480/720/4k2k
                            rgb_color_adjust = RGBL_to_BT709;	// 0x1;		// rgb_L->bt709
                            yuv_color_adjust = BT709_to_RGBF;	// 0x2;		// bt709->rgb_L
                            break;
                        case DEVICE_4k2k_MODE:			//i/p: rgb_L 4k2k, o/p: RGB_L 480/720/4k2k					
                            rgb_color_adjust = RGBL_to_BT709;	// 0x1;		// rgb_L->bt2020,	assume bt2020 as bt709	
                            yuv_color_adjust = BT709_to_RGBF;	// 0x2;		// bt2020->rgb_L
                            break;
                    }
                }else{
                    // i/p RGB_L   o/p YUV
                    switch (packetinfo->m_WorkingMode_in)
                    {
                        case DEVICE_480_MODE:		
                            switch (packetinfo->m_WorkingMode_out)
                            {
                                case DEVICE_480_MODE:		
                                    rgb_color_adjust = RGBL_to_BT601;	
                                    yuv_color_adjust = 0xff;			
                                    break;
                                case DEVICE_720_MODE:		
                                    rgb_color_adjust = RGBL_to_BT709;	
                                    yuv_color_adjust = 0xff;	
                                    break;
                                case DEVICE_4k2k_MODE:		
                                    rgb_color_adjust = RGBL_to_BT709;	
                                    yuv_color_adjust = 0xff;	
                                    break;
                            }
                            break;
                        case DEVICE_720_MODE:
                            switch (packetinfo->m_WorkingMode_out)
                            {
                                case DEVICE_480_MODE:		
                                    rgb_color_adjust = RGBL_to_BT601;	
                                    yuv_color_adjust = 0xff;			
                                    break;
                                case DEVICE_720_MODE:		
                                    rgb_color_adjust = RGBL_to_BT709;	
                                    yuv_color_adjust = 0xff;			
                                    break;
                                case DEVICE_4k2k_MODE:		
                                    rgb_color_adjust = RGBL_to_BT709;	
                                    yuv_color_adjust = 0xff;			
                                    break;
                            }
                            break;
                        case DEVICE_4k2k_MODE:						
                            switch (packetinfo->m_WorkingMode_out)
                            {
                                case DEVICE_480_MODE:		
                                    rgb_color_adjust = RGBL_to_BT601;		
                                    yuv_color_adjust = 0xff;	
                                    break;
                                case DEVICE_720_MODE:		
                                    rgb_color_adjust = RGBL_to_BT709;		
                                    yuv_color_adjust = 0xff;			
                                    break;
                                case DEVICE_4k2k_MODE:		
                                    rgb_color_adjust = RGBL_to_BT709;	
                                    yuv_color_adjust = 0xff;			
                                    break;
                            }
                            break;
                    }
                }
                break;
	}	

        mesg("%s rgb_color_adjust=%d;yuv_color_adjust=%d\n",__func__,rgb_color_adjust,yuv_color_adjust);	

        mesg("Video CFG: [rgb_color_adjust(%d)]: %s , [yuv_color_adjust(%d)]: %s",
                rgb_color_adjust,
                rgb_color_adjust == 0 ? "rgb_F to yuv(bt709)" :
                rgb_color_adjust == 1 ? "rgb_L to yuv(bt709)" :
                rgb_color_adjust == 2 ? "rgb_F to yuv(bt601)" :
                rgb_color_adjust == 3 ? "rgb_L to yuv(bt601)" : "Not Set",
                yuv_color_adjust,
                yuv_color_adjust == 0 ? "yuv(bt709) to rgb_F" :
                yuv_color_adjust == 1 ? "yuv(bt709) to rgb_L" :
                yuv_color_adjust == 2 ? "yuv(bt601) to rgb_F" :
                yuv_color_adjust == 3 ? "yuv(bt601) to rgb_L" :
                yuv_color_adjust == 4 ? "yuv(bt601) to yuv(bt709)" :
                yuv_color_adjust == 5 ? "yuv(bt709) to yuv(bt601)" : "Not Set");
				
        //hdmi_rgb2yuv_enable(hdmi_cxt,ch,AVER_TRUE);
        //mesg(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>ch%d rgb2yuv enable\n",ch);

#if 1				
        if (rgb_color_adjust != 0xff)
        {
            hdmi_rgb2yuv_coeff(hdmi_cxt,ch,rgb_color_adjust);
            hdmi_rgb2yuv_enable(hdmi_cxt,ch,AVER_TRUE);// i/p is yuv, o/p is rgb
            mesg("....ch%d rgb color adjust enable\n",ch);
        }
        else
        {
            hdmi_rgb2yuv_enable(hdmi_cxt,ch,AVER_FALSE);// i/p is yuv, o/p is rgb
            mesg("....ch%d rgb color adjust disable\n",ch);
        }

        if (yuv_color_adjust != 0xff)
        {
            hdmi_yuv2rgb_coeff(hdmi_cxt,ch,yuv_color_adjust);
            hdmi_yuv2rgb_enable(hdmi_cxt,ch,AVER_TRUE);
            mesg("...ch%d yuv color adjust enable\n",ch);
        }
        else
        {
            hdmi_yuv2rgb_enable(hdmi_cxt,ch,AVER_FALSE);
            mesg("...ch%d yuv color adjust disable\n",ch);
        }
#endif

        //data = hdmi_rd_reg(hdmi_cxt,V_RGB_TO_YUV_EN(ch)); 
        //mesg("(ch%d)V_RGB_TO_YUV_EN reg 0x%x=%x\n",ch,V_RGB_TO_YUV_EN(ch),data);
	
        //data = hdmi_rd_reg(hdmi_cxt,V_YUV_TO_RGB_EN(ch)); 
        //mesg("(ch%d)V_YUV_TO_RGB_EN reg 0x%x=%x\n",ch,V_YUV_TO_RGB_EN(ch),data);

    }			
}
#endif

int hdmi_get_audio_sample(handle_t hdmi_handle)
{
    hdmi_cxt_t *hdmi_cxt=hdmi_handle;
    if(hdmi_cxt)
        return hdmi_cxt->audio_sample;
    else
        return hdmi_AUDIO_RATE_48000;
}

void hdmi_get_packetinfo(handle_t hdmi_handle,hdmi_packetinfo_t *packetinfo) 
{
    hdmi_cxt_t *hdmi_cxt=hdmi_handle;
    int ch;
    
    mesg("%s...\n",__func__);
    if(hdmi_cxt)
    {
        ch = hdmi_cxt->device_index;
        mesg("(ch%d) m_ColorSpaceMode %s", ch,
		        	hdmi_cxt->packetinfo.m_ColorSpaceMode == VIDEO_RGB_FULL_FMT ? "VIDEO_RGB_FULL_FMT" :
		        	hdmi_cxt->packetinfo.m_ColorSpaceMode == VIDEO_YUV_FMT ? "VIDEO_YUV_FMT" : "UNKNOWN");
			
        mesg("(ch%d) m_SamplingMode %s", ch,
		        	hdmi_cxt->packetinfo.m_SamplingMode == VIDEO_RGB_MODE ? "VIDEO_RGB_MODE" :
		        	hdmi_cxt->packetinfo.m_SamplingMode == VIDEO_422_MODE ? "VIDEO_422_MODE" :
		        	hdmi_cxt->packetinfo.m_SamplingMode == VIDEO_444_MODE ? "VIDEO_444_MODE" : "UNKNOWN");
    
        mesg("(ch%d) m_WorkingMode %s", ch,
		        	hdmi_cxt->packetinfo.m_WorkingMode == DEVICE_4k2k_MODE ? "DEVICE_4k2k_MODE" :
		        	hdmi_cxt->packetinfo.m_WorkingMode == DEVICE_720_MODE ? "DEVICE_720_MODE" :
		        	hdmi_cxt->packetinfo.m_WorkingMode == DEVICE_480_MODE ? "DEVICE_480_MODE" : "UNKNOWN"); 
		   
        hdmi_cxt->packetinfo.m_PixelMode = SINGLE_PIXEL_MODE;
        hdmi_cxt->packetinfo.m_PacketRepetition = 1;
       
        if(packetinfo)
            *packetinfo=hdmi_cxt->packetinfo;
    }
}

void hdmi_set_out_format(handle_t hdmi_handle,hdmi_out_format_e format)
{
    //hdmi_cxt_t *hdmi_cxt = hdmi_handle;
    
    //hdmi_wr_reg(hdmi_cxt, hdmi_IO_REG_IO_REG_03, format);
}

void hdmi_set_audio_out_enable(handle_t hdmi_handle,BOOL_T Enable,int ch)
{
    hdmi_cxt_t *hdmi_cxt = hdmi_handle;
    int ret;
    
    if(hdmi_cxt)
    {
        if (Enable == AVER_FALSE)
            ret = io_mask_write(hdmi_cxt, V_hdmi_I2S(ch), V_hdmi_I2S_ENABLE, AVER_FALSE); //AVER_FALSE
        else
            ret = io_mask_write(hdmi_cxt, V_hdmi_I2S(ch), V_hdmi_I2S_ENABLE, AVER_TRUE); //AVER_FALSE
    }
}


static void hdmi_callback(hdmi_cxt_t *hdmi_cxt,hdmi_event_e event)
{
    if(hdmi_cxt)
    {
        if(hdmi_cxt->callback_func)
        {
            hdmi_cxt->callback_func(hdmi_cxt->cb_cxt,event);
        }
    }
}

void hdmi_register_callback(handle_t hdmi_handle, hdmi_callback_t callback_func, void *cxt)
{
    hdmi_cxt_t *hdmi_cxt = hdmi_handle;

    if (hdmi_cxt)
    {
        if (hdmi_cxt->callback_func == NULL)
        {
            hdmi_cxt->lock = HDMI_UNLOCK;

            if (callback_func)
            {
                hdmi_cxt->callback_func = callback_func;
                hdmi_cxt->cb_cxt = cxt;
            }

            if (hdmi_cxt->taskmgr)
            {
                if (hdmi_cxt->task_handle == NULL)
                {
                    hdmi_cxt->task_handle = task_model_create_task(hdmi_cxt->taskmgr, hdmi_task, hdmi_cxt, "hdmi_task");
                    mesg("%s create hdmi_task\n",__func__);
                }
                task_model_run_task_period(hdmi_cxt->taskmgr, hdmi_cxt->task_handle, /*500*/300, -1, NULL); 
            }
        }
    }
    mesg("%s\n",__func__);	
}

void hdmi_unregister_callback(handle_t hdmi_handle)
{
    hdmi_cxt_t *hdmi_cxt = hdmi_handle;

    if (hdmi_cxt)
    {
        if (hdmi_cxt->callback_func != NULL)
        {
            if (hdmi_cxt->taskmgr)
            {
                 if (hdmi_cxt->task_handle) 
                 {
                     task_model_release_task(hdmi_cxt->taskmgr, hdmi_cxt->task_handle);
                     hdmi_cxt->task_handle = NULL;
                 }
             }

             hdmi_cxt->callback_func = NULL;
             hdmi_cxt->cb_cxt = NULL;
        }
    }
    mesg("%s\n",__func__);	
}


void StartInterrupt(void *handle)
{
    mesg("jeff_hwInit: enable MER+");
    hdmi_cxt_t *hdmi_cxt = handle;
    XIntc *IntcInstPtr = &Intc;
    //StartISRTimer();
    XIntc_Start(IntcInstPtr, XIN_REAL_MODE);
    pci_model_mmio_write(hdmi_cxt->pci_handle, 0, 0x10, Bit(10));
    mesg("jeff_hwInit: enable MER-");
}

void StopInterrupt(void *handle)
{
    mesg("jeff_hwInit: disable MER+");
    hdmi_cxt_t *hdmi_cxt = handle;
    XIntc *IntcInstPtr = &Intc;
    XIntc_Stop(IntcInstPtr);
    pci_model_mmio_write(hdmi_cxt->pci_handle, 0, 0x10, Bit(10));
    //StopISRTimer();
    //m_EnableISRCnt = 0;
    mesg("jeff_hwInit: disable MER-");
}

void IdentifyMsiInterrupt(void *hdmi_cxt, unsigned int MessageID)
{
    hdmi_wr_reg(hdmi_cxt,0x10, Bit(10)); //Shawn_20190613 [Bit 31 -> 10]
}


unsigned int HardwareInit(void* handle)
{
    mesg("+");
    hdmi_cxt_t *hdmi_cxt = handle;
    int chip_id;
    int fpga_id;
#if 1
    StopInterrupt(hdmi_cxt);

    // Reset RX+
    mesg(" jeff_hwInit: reset+");
    unsigned int GpioValue = pci_model_mmio_read(hdmi_cxt->pci_handle,0, 0x40);
    mesg("read gpio:%x \n", GpioValue);
    GpioValue |= Bit(0);
    mesg("write gpio:%x \n", GpioValue);
    pci_model_mmio_write(hdmi_cxt->pci_handle, 0, 0x40, GpioValue);

    sys_msleep(15);

    GpioValue &= ~Bit(0);
    mesg("write gpio:%x \n", GpioValue);
    pci_model_mmio_write(hdmi_cxt->pci_handle, 0, 0x40, GpioValue);

    sys_msleep(15);

    //Shawn_20190510+s
    GpioValue |= Bit(0);
    mesg("write gpio:%x \n", GpioValue);
    pci_model_mmio_write(hdmi_cxt->pci_handle, 0, 0x40, GpioValue);

    sys_msleep(15);
    //Shawn_20190510+e

    mesg(" jeff_hwInit: reset-");

    #if 1 
    chip_id=hdmi_rd_reg(hdmi_cxt,0x00);
    mesg("======================HDMI RX version reg 0x00=%x\n",chip_id);
    fpga_id=hdmi_rd_reg(hdmi_cxt,0x04);
    mesg("======================HDMI FPGA id reg 0x04=%x\n",fpga_id);
    #endif

    // Reset RX-
#endif
    U32_T Status = XST_FAILURE;

    XVphy_Config *XVphyCfgPtr;
    VphyErrorFlag = AVER_FALSE;
    VphyPllLayoutErrorFlag = AVER_FALSE;

    /* Start in color bar */
    IsPassThrough = (AVER_FALSE);

    /* Initialize IIC */

    /* Initialize external clock generator */

    /* Disable TMDS181 HPD passthrough for ZCU106 Rev B and below */

    /* Load HDCP keys from EEPROM */

    /* Initialize IRQ */
    Status = SetupInterruptSystem();
    if (Status == XST_FAILURE) 
    {
        debug_msg("IRQ init failed.\r\n\r\n");
        return XST_FAILURE;
    }

#ifdef XPAR_XV_HDMIRXSS_NUM_INSTANCES
    // Initialize HDMI RX Subsystem
    /* Get User Edid Info */
    mesg("Get User Edid Info!\n");
    XV_HdmiRxSs_SetEdidParam(&HdmiRxSs, (U8_T*)&Edid, sizeof(Edid));
    mesg("Edid:%x Length:%x set Edid:%x Length:%x \n", &Edid, sizeof(Edid), HdmiRxSs.EdidPtr,HdmiRxSs.EdidLength);
    XV_HdmiRxSs_ConfigPtr = XV_HdmiRxSs_LookupConfig(XPAR_XV_HDMIRX_0_DEVICE_ID);

    if(XV_HdmiRxSs_ConfigPtr == NULL) 
    {
        HdmiRxSs.IsReady = 0;
        return (XST_DEVICE_NOT_FOUND);
    }

    //Initialize top level and all included sub-cores
    mesg("Initialize top level and all included sub-cores! \n");
    Status = XV_HdmiRxSs_CfgInitialize(&HdmiRxSs, XV_HdmiRxSs_ConfigPtr, XV_HdmiRxSs_ConfigPtr->BaseAddress);
    if(Status != XST_SUCCESS) 
    {
        debug_msg("ERR:: HDMI RX Subsystem Initialization failed %d\r\n", Status);
        return(XST_FAILURE);
    }

    /* Set the Application version in RXSs driver structure */
    mesg("Set the Application version in RXSs driver structure!\n");
    XV_HdmiRxSS_SetAppVersion(&HdmiRxSs, APP_MAJ_VERSION, APP_MIN_VERSION);

    //Register HDMI RX SS Interrupt Handler with Interrupt Controller
    mesg("jeff_hwInit: Register HDMI RX SS Interrupt Handler with Interrupt Controller");
    mesg("jeff_hwInit: XIntc_Connect \n");
    Status |= XIntc_Connect(&Intc,
            XPAR_INTC_0_V_HDMIRXSS_0_VEC_ID,
            (XInterruptHandler)XV_HdmiRxSS_HdmiRxIntrHandler, (void *)&HdmiRxSs);

    if (Status == XST_SUCCESS) 
    {
        mesg("jeff_hwInit: XIntc_Enable \n");
        XIntc_Enable(&Intc, XPAR_INTC_0_V_HDMIRXSS_0_VEC_ID );
    }else {
        mesg("ERR:: Unable to register HDMI RX interrupt handler \n");
        mesg("HDMI RX SS initialization error \n");
        return XST_FAILURE;
    }

    /* RX callback setup */
    mesg("jeff_hwInit: RX callback setup");
    XV_HdmiRxSs_SetCallback(&HdmiRxSs, XV_HDMIRXSS_HANDLER_CONNECT, (void *)RxConnectCallback, (void *)&HdmiRxSs);
    XV_HdmiRxSs_SetCallback(&HdmiRxSs, XV_HDMIRXSS_HANDLER_BRDGOVERFLOW, (void *)RxBrdgOverflowCallback, (void *)&HdmiRxSs);
    XV_HdmiRxSs_SetCallback(&HdmiRxSs, XV_HDMIRXSS_HANDLER_AUX, (void *)RxAuxCallback, (void *)&HdmiRxSs);
    XV_HdmiRxSs_SetCallback(&HdmiRxSs, XV_HDMIRXSS_HANDLER_AUD, (void *)RxAudCallback, (void *)&HdmiRxSs);
    XV_HdmiRxSs_SetCallback(&HdmiRxSs, XV_HDMIRXSS_HANDLER_LNKSTA, (void *)RxLnkStaCallback, (void *)&HdmiRxSs);
    //XV_HdmiRxSs_SetCallback(&HdmiRxSs, XV_HDMIRXSS_HANDLER_DDC, RxDdcCallback, (void *)&HdmiRxSs);
    XV_HdmiRxSs_SetCallback(&HdmiRxSs, XV_HDMIRXSS_HANDLER_STREAM_DOWN, (void *)RxStreamDownCallback, (void *)&HdmiRxSs);
    XV_HdmiRxSs_SetCallback(&HdmiRxSs, XV_HDMIRXSS_HANDLER_STREAM_INIT, (void *)RxStreamInitCallback, (void *)&HdmiRxSs);
    XV_HdmiRxSs_SetCallback(&HdmiRxSs, XV_HDMIRXSS_HANDLER_STREAM_UP, (void *)RxStreamUpCallback, (void *)&HdmiRxSs);

#endif

    // Initialize Video PHY
    // The GT needs to be initialized after the HDMI RX and TX.
    // The reason for this is the GtRxInitStartCallback
    // calls the RX stream down callback.
    XVphyCfgPtr = XVphy_LookupConfig(XPAR_VPHY_0_DEVICE_ID);
    if (XVphyCfgPtr == NULL) 
    {
        debug_msg("Video PHY device not found\r\n\r\n");
        return XST_FAILURE;
    }

    /* Register VPHY Interrupt Handler */
    Status = XIntc_Connect(&Intc, XPAR_INTC_0_VPHY_0_VEC_ID, (XInterruptHandler)XVphy_InterruptHandler, (void *)&Vphy);
    if (Status != XST_SUCCESS) 
    {
        debug_msg("HDMI VPHY Interrupt Vec ID not found!\r\n");
        return XST_FAILURE;
    }

    /* Initialize HDMI VPHY */
    Status = XVphy_Hdmi_CfgInitialize(&Vphy, 0, XVphyCfgPtr);
    if (Status != XST_SUCCESS) 
    {
        debug_msg("HDMI VPHY initialization error\r\n");
        return XST_FAILURE;
    }

    /* Enable VPHY Interrupt */
    XIntc_Enable(&Intc, XPAR_INTC_0_VPHY_0_VEC_ID);


#ifdef XPAR_XV_HDMIRXSS_NUM_INSTANCES
    XVphy_SetHdmiCallback(&Vphy, XVPHY_HDMI_HANDLER_RXINIT, (void *)VphyHdmiRxInitCallback, (void *)&Vphy);
    XVphy_SetHdmiCallback(&Vphy, XVPHY_HDMI_HANDLER_RXREADY, (void *)VphyHdmiRxReadyCallback, (void *)&Vphy);
#endif

    XVphy_SetErrorCallback(&Vphy, (void *)VphyErrorCallback, (void *)&Vphy);

#if (XPAR_VPHY_0_TRANSCEIVER == XVPHY_GTXE2)
    XVphy_SetPllLayoutErrorCallback(&Vphy, (void *)VphyPllLayoutErrorCallback, (void *)&Vphy);
#endif

    mesg("---------------------------------");
    mesg("jeff_hwInit: Enable exceptions");
    /* Enable exceptions. */
    Xil_AssertSetCallback((Xil_AssertCallback)Xil_AssertCallbackFunction);
    //Xil_ExceptionEnable();
    StartInterrupt(hdmi_cxt);

    mesg("jeff_hwInit: RX Hardware Init done--------------------------------");

    //m_WorkingMode = m_Old_WorkingMode;
    //m_Old_WorkingMode = 0xff;
    //m_ColorSpaceMode = m_Old_ColorSpaceMode;
    //m_Old_ColorSpaceMode = 0xff;
    //m_SamplingMode = m_Old_SamplingMode;
    //m_Old_SamplingMode = 0xff;

    sys_msleep(500);
    StartInterrupt(hdmi_cxt);
    return XST_SUCCESS;
}


void PowerChipSel(void *hdmi_cxt, unsigned int enable, unsigned int* pdwWaitTime)
{
    mesg("+");

    if (enable)
    {
        //StopTimer();

        //Power on chip
        //StartTimer(1000);
        StartInterrupt(hdmi_cxt);

        //wait 0ms
        *pdwWaitTime = 0;
    }else{
        StopInterrupt(hdmi_cxt);
        //StopTimer();
        //StopISRTimer();
        //StopAVIinfoTimer();
        //StopCheckReloadTimer();

        //wait 200ms
        *pdwWaitTime = 200;
    }

    mesg("-");
}

void PowerSet(void *hdmi_cxt, unsigned int enable, unsigned int* pdwWaitTime)
{
    mesg("+");

    if (enable)
    {
        HardwareInit(hdmi_cxt);
    }
    else
    {
        //PowerChipSel(AVER_FALSE, pdwWaitTime);
    }

    mesg("-");
}

static int hdmi_irq_func(void *cxt)
{
    mesg("%s.............hdmi_irq_func...........\n",__func__);
    hdmi_cxt_t *hdmi_cxt = cxt;
    pci_model_mmio_write(hdmi_cxt->pci_handle, 0, 0x10, Bit(10));
    XIntc_InterruptHandler(&Intc);
    return 0;
}

void hdmi_hardware_init(handle_t hdmi_handle)
{
	hdmi_cxt_t *hdmi_cxt = hdmi_handle;
	int chip_id;
	 
	hdmi_cxt->is_lock_state=AVER_FALSE;
	#if 0
    chip_id=hdmi_rd_reg(hdmi_cxt,0x00);
    mesg("%s HDMI RX chip version reg 0x00=%x\n",__func__,chip_id);
    #endif
    
    HardwareInit(hdmi_cxt);
	
}

static char* hdmi_name = "hdmi";

static void *cn312_hdmi_attach(i2c_model_bus_handle_t bus_handle,unsigned short i2c_addr,void *drv_set_info)
{
    hdmi_cxt_t *hdmi_cxt=NULL;
    hdmi_out_ctrl_t *cfg=drv_set_info;
  
    i2c_model_handle_t i2c_mgr=NULL;
    cxt_mgr_handle_t cxt_mgr=NULL;
    task_model_handle_t task_model_handle=NULL;
    task_handle_t task_handle=NULL;
    pci_model_handle_t pci_handle = NULL;
    //int ret;
    int ch;

    //int data_tmp=0xff;
    enum
    {
        NO_ERROR,
        ERROR_ALLOC_CXT,
        NO_MATCH_CHIPID,
        NO_CFG,
        ERROR_GET_TASKMGR,
        ERROR_ALLOC_TASK,
    }err=NO_ERROR;
    
    mesg("%s......device_index=%d\n",__func__,cfg->device_index);

    do
    {
        //U16_T chip_id=0;
        //BOOL_T match_chipid=AVER_FALSE;
        //U8_T value;
        
        if(!cfg)
        {
            err=NO_CFG;
            break;
        }
        
        hdmi_cxt=mem_model_alloc_buffer(sizeof(hdmi_cxt_t));
        if(!hdmi_cxt)
        {
            err=ERROR_ALLOC_CXT;
            break;
        }

        hdmi_cxt->bus_handle=bus_handle;
        g_bus_handle = bus_handle;
        hdmi_cxt->addr=i2c_addr;
        g_addr = i2c_addr;
        mesg("i2c addr:%x g_addr:%x bus_handle:%x g_bus_handle:%x \n", hdmi_cxt->addr,g_addr, bus_handle, g_bus_handle);
        
        //hdmi_cxt->id=hdmi_dev_count++;
        
        hdmi_cxt->device_index = cfg->device_index;
       
        i2c_mgr=i2c_model_bus_get_mgr(bus_handle);
        cxt_mgr=get_cxt_manager_from_context(i2c_mgr);
        task_model_handle=cxt_manager_get_context(cxt_mgr,TASK_MODEL_CXT_ID,0);
        if(!task_model_handle)
        {
            err=ERROR_GET_TASKMGR;
            break;
        }
        
        pci_handle = cxt_manager_get_context(cxt_mgr,PCI_CXT_ID,0);
        mesg("cn312m pci irq:%x name:%s \n", pci_get_irq(pci_handle), hdmi_name);
        hdmi_cxt->pci_handle = pci_handle;
        //if(request_irq(pci_get_irq(pci_handle), aver_hdmi_irq,IRQF_SHARED, hdmi_name, hdmi_cxt)<0)
        {
             //mesg_err("Failed to request hdmi IRQ ! \n");
        }
        hdmi_cxt->device_index = cfg->device_index;
        
        //ch = cfg->device_index;
        
        mesg("%s...hdmi_ch=%d\n",__func__,hdmi_cxt->device_index);
        
        
        //break;
        
        hdmi_cxt->is_lock_state=AVER_FALSE;
        ch = cfg->device_index;
        //for (ch = 0; ch < ACTIVE_PATH_NUM; ch++)
        {
            hdmi_cxt->is_state_change = AVER_TRUE; //rr
            hdmi_cxt->packetinfo.m_WorkingMode = DEVICE_720_MODE;
            hdmi_cxt->packetinfo.m_ColorSpaceMode = VIDEO_YUV_FMT;
            hdmi_cxt->packetinfo.m_SamplingMode = VIDEO_422_MODE;
            hdmi_cxt->packetinfo.m_PixelMode = SINGLE_PIXEL_MODE;
            hdmi_cxt->packetinfo.m_PacketRepetition = 1;
            hdmi_cxt->packetinfo.m_PacketVideoLock = 0;
            hdmi_cxt->packetinfo.m_PacketWidth = 0;
            hdmi_cxt->packetinfo.m_PacketHeight = 0;
            hdmi_cxt->packetinfo.m_PacketInterlace = 0;
            hdmi_cxt->packetinfo.m_PacketFrameRate = 0;
            hdmi_cxt->packetinfo.m_RxFrameRate = 0;

            hdmi_cxt->packetinfo.m_Old_WorkingMode = DEVICE_720_MODE;
            hdmi_cxt->packetinfo.m_Old_ColorSpaceMode = VIDEO_YUV_FMT;
            hdmi_cxt->packetinfo.m_Old_SamplingMode = VIDEO_422_MODE;
            hdmi_cxt->packetinfo.m_Old_PixelMode = SINGLE_PIXEL_MODE;
            hdmi_cxt->packetinfo.m_Old_PacketRepetition = 1;
            hdmi_cxt->packetinfo.m_Old_PacketVideoLock = 0;
            hdmi_cxt->packetinfo.m_Old_PacketWidth = 0;
            hdmi_cxt->packetinfo.m_Old_PacketHeight = 0;
            hdmi_cxt->packetinfo.m_Old_PacketInterlace = 0;
            hdmi_cxt->packetinfo.m_Old_PacketFrameRate = 0;
            hdmi_cxt->packetinfo.m_Old_RxFrameRate = 0;
            hdmi_cxt->audio_sample = hdmi_AUDIO_RATE_48000;
            hdmi_cxt->old_audio_sample = hdmi_AUDIO_RATE_48000;
        }
	    
        HardwareInit(hdmi_cxt);
        // initial for image attached control //0824
        hdmi_cxt->is_state_change = AVER_FALSE; 
        hdmi_cxt->is_lock_state = AVER_TRUE;

        hdmi_cxt->taskmgr = task_model_handle;
        cxt_manager_ref_context(hdmi_cxt->taskmgr);
        #if 1
        task_handle = task_model_create_task(task_model_handle, hdmi_task, hdmi_cxt, "hdmi_task");
        if (!task_handle) {
            err = ERROR_ALLOC_TASK;
            break;
        }
        #endif

        //mesg("hdmi chip_version:%08x\n",hdmi_rd_reg(hdmi_cxt,FPGA_VERSION));
        //hdmi_wr_reg(hdmi_cxt, V_RGB_TO_YUV_EN(0), 0x01);
        //mesg("...\n");
        #if 0
        //hdmi_BT1120_enable
        for (ch = 0; ch < ACTIVE_PATH_NUM; ch++)
        {
            ret = io_mask_write(hdmi_cxt, V_hdmi_BT1120(ch), V_hdmi_BT1120_ENABLE, AVER_TRUE);
            ret = io_mask_write(hdmi_cxt, V_hdmi_I2S(ch), V_hdmi_I2S_ENABLE, AVER_TRUE); //AVER_FALSE
            data_tmp = hdmi_rd_reg(hdmi_cxt,V_hdmi_BT1120(ch)); 
            mesg("(ch%d)BT1120_enable reg 0x%x=%x\n",ch,V_hdmi_BT1120(ch),data_tmp);
        }
        #endif
        
    }while(0);
    if(err!=NO_ERROR)
    {
        debug_msg("%s err %d\n",__func__,err);
        switch(err)
        {
            mem_model_free_buffer(hdmi_cxt);
        case ERROR_ALLOC_TASK:
        case ERROR_GET_TASKMGR:
        case ERROR_ALLOC_CXT:
        case NO_CFG:
             break;   
        default:
             break;
        }
    }
    
    return hdmi_cxt;
}

void hdmi_power_on(handle_t hdmi_handle)
{
    hdmi_cxt_t *hdmi_cxt=hdmi_handle;
    task_model_handle_t task_model_handle = NULL;
    task_handle_t task_handle = NULL;
    int i;
    
    mesg("%s\n",__func__);

    if(!hdmi_cxt)
    {
        debug_msg("%s hdmi_cxt is NULL", __func__);
        return;
    }

    // initialize variables
      
    mesg("%s %p\n",__func__,hdmi_cxt);
    hdmi_hardware_init(hdmi_cxt);

    sys_msleep(10);
#if 1
    task_model_handle = hdmi_cxt->taskmgr;
    mesg("%s taskmgr %p\n",__func__,task_model_handle);
    task_handle=task_model_create_task(task_model_handle,hdmi_task,hdmi_cxt,"hdmi_task");
    if(!task_handle)
    {
        mesg("%s task_handle is NULL", __func__);
        return;
    }
    #if 1
    
    mesg("=====%s hdmi_callback for resume\n",__func__);
    hdmi_callback(hdmi_cxt,HDMI_RESUME);
  
    #endif
    mesg("%s run_task_period\n",__func__);
    hdmi_cxt->task_handle=task_handle;
    task_model_run_task_period(task_model_handle,task_handle,1000,-1,NULL);
#endif
}

void hdmi_power_off(handle_t hdmi_handle)
{
    hdmi_cxt_t *hdmi_cxt=hdmi_handle;

    mesg("%s\n",__func__);

    if(!hdmi_cxt)
    {
        mesg("%s hdmi_cxt is NULL", __func__);
        return;
    }

#if 1
    if(hdmi_cxt->task_handle)
    {
        task_model_release_task(hdmi_cxt->taskmgr,hdmi_cxt->task_handle);
        hdmi_cxt->task_handle = NULL;
    }
#endif    
}

static void cn312_hdmi_detach(void *cxt)
{
    hdmi_cxt_t *hdmi_cxt=cxt;
    if(hdmi_cxt)
    {
        //int i;

        if (hdmi_cxt->taskmgr) {
            if (hdmi_cxt->task_handle) {
                task_model_release_task(hdmi_cxt->taskmgr, hdmi_cxt->task_handle);
            }
            cxt_manager_unref_context(hdmi_cxt->taskmgr);
        }
        mem_model_free_buffer(hdmi_cxt);
    }
}
I2C_MODEL_DRV_DECLARE(hdmi_ID);
