/******************************************************************************
*
* Copyright (C) 2014 - 2016 Xilinx, Inc. All rights reserved.
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* Use of the Software is limited solely to applications:
* (a) running on a Xilinx device, or
* (b) that interact with a Xilinx device through a bus or interconnect.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
* XILINX  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* Except as contained in this notice, the name of the Xilinx shall not be used
* in advertising or otherwise to promote the sale, use or other dealings in
* this Software without prior written authorization from Xilinx.
*
******************************************************************************/
/*****************************************************************************/
/**
*
* @file xil_io.c
*
* Contains I/O functions for memory-mapped or non-memory-mapped I/O
* architectures.
*
* @note
*
* This file contains architecture-dependent code.
*
* <pre>
* MODIFICATION HISTORY:
*
* Ver   Who      Date     Changes
* ----- -------- -------- -----------------------------------------------
* 5.00 	pkp  	 05/29/14 First release
* </pre>
******************************************************************************/


/***************************** Include Files *********************************/
#include "xil_io.h"
#include "xil_types.h"
#include "xil_assert.h"
#include "debug.h"


extern int hdmi_rd_reg_xil(int reg, U8_T* pBuf);

static void I2CRead(U32_T Addr, U8_T* pBuf)
{
#if 0
    U8_T slaveAddr = 0x4E;
    int i = 0;

    // write slave-address and generate start
    slaveAddr &= 0xfe;
    I2CWriteData(slaveAddr, 0x90);

    // write sub-address
    for (i = 2; i > 0; i--)
    {
        I2CWriteData((Addr >> (i * 8)) & 0xff, 0x10);
    }
    I2CWriteData(Addr & 0xff, 0x50);

    // write slave-address(R) and generate restart
    slaveAddr |= 0x01;
    I2CWriteData(slaveAddr, 0x90);

    for (i = 3; i > 0; i--)
    {
        I2CReadData(pBuf + i, 0x20);
    }

    // read last data and generate stop
    I2CReadData(pBuf, 0x68);
#endif

    hdmi_rd_reg_xil(Addr, pBuf);
}

extern int hdmi_wr_reg_xil(int reg, U8_T* pBuf);

static void I2CWrite(U32_T Addr, U8_T* pBuf)
{
#if 0
    U8_T slaveAddr = 0x4E;
    int i = 0;

    // write slave-address and generate start
    slaveAddr &= 0xfe;
    I2CWriteData(slaveAddr, 0x90);

    // write sub-address
    for ( i = 2; i >= 0; i--)
    {
        I2CWriteData((Addr >> (i * 8)) & 0xff, 0x10);
    }

    // write data
    for ( i = 3; i > 0; i--)
    {
        I2CWriteData(*(pBuf + i), 0x10);
    }

    // write last data and generate stop
    I2CWriteData(*(pBuf), 0x50);
#endif

    hdmi_wr_reg_xil(Addr, pBuf);
}
/*****************************************************************************/
/**
*
* @brief    Performs an input operation for a memory location by reading
*           from the specified address and returning the 8 Bit Value read from
*            that address.
*
* @param	Addr: contains the address to perform the input operation
*
* @return	The 8 Bit Value read from the specified input address.

*
******************************************************************************/
 U8_T Xil_In8(UINTPTR Addr)
{
    //CHECK_ADDR_READ;

	//return READ_REGISTER_UCHAR((PUCHAR)(g_pMemBase + Addr));//*(volatile U8_T *) Addr; //AVerMedia_Jeff_20190313
    U8_T pBuf[4] = { 0 };
    I2CRead(Addr, pBuf);
    return pBuf[0];
}

/*****************************************************************************/
/**
*
* @brief    Performs an input operation for a memory location by reading from
*           the specified address and returning the 16 Bit Value read from that
*           address.
*
* @param	Addr: contains the address to perform the input operation
*
* @return	The 16 Bit Value read from the specified input address.
*
******************************************************************************/
 U16_T Xil_In16(UINTPTR Addr)
{
    //CHECK_ADDR_READ;

	//return READ_REGISTER_USHORT((PUSHORT)(g_pMemBase + Addr));//*(volatile U16_T *) Addr; //AVerMedia_Jeff_20190313
    U8_T pBuf[4] = { 0 };
    I2CRead(Addr, pBuf);
    return (pBuf[1] << 8) | pBuf[0];
}

/*****************************************************************************/
/**
*
* @brief    Performs an input operation for a memory location by
*           reading from the specified address and returning the 32 Bit Value
*           read  from that address.
*
* @param	Addr: contains the address to perform the input operation
*
* @return	The 32 Bit Value read from the specified input address.
*
******************************************************************************/
 U32_T Xil_In32(UINTPTR Addr)
{
    //CHECK_ADDR_READ;

    //return READ_REGISTER_ULONG((PULONG)(g_pMemBase + Addr));//*(volatile U32_T *) Addr; //AVerMedia_Jeff_20190313
    U8_T pBuf[4] = { 0 };
    I2CRead(Addr, pBuf);
    //mesg("jeff_i2c: ***I2C Read Addr(0x%08x) Data(0x%02X%02X%02X%02X) => %x \n", Addr, pBuf[3], pBuf[2], pBuf[1], pBuf[0], (pBuf[3] << 24) | (pBuf[2] << 16) | (pBuf[1] << 8) | pBuf[0]);
    return (pBuf[3] << 24) | (pBuf[2] << 16) | (pBuf[1] << 8) | pBuf[0];
}

/*****************************************************************************/
/**
*
* @brief     Performs an input operation for a memory location by reading the
*            64 Bit Value read  from that address.
*
*
* @param	Addr: contains the address to perform the input operation
*
* @return	The 64 Bit Value read from the specified input address.
*
******************************************************************************/
static INLINE U64_T Xil_In64(UINTPTR Addr)
{
    //CHECK_ADDR_READ;

	//return READ_REGISTER_ULONG64((PULONG64)(g_pMemBase + Addr));//*(volatile U64_T *) Addr; //AVerMedia_Jeff_20190313
    U8_T pBuf[4] = { 0 };
    I2CRead(Addr, pBuf);
    return (pBuf[3] << 24) | (pBuf[2] << 16) | (pBuf[1] << 8) | pBuf[0];
}

/*****************************************************************************/
/**
*
* @brief    Performs an output operation for an memory location by
*           writing the 8 Bit Value to the the specified address.
*
* @param	Addr: contains the address to perform the output operation
* @param	Value: contains the 8 Bit Value to be written at the specified
*           address.
*
* @return	None.
*
******************************************************************************/
void Xil_Out8(UINTPTR Addr, U8_T Value)
{
    //CHECK_ADDR_WRITE;

	//volatile U8_T *LocalAddr = (volatile U8_T *)Addr;
	//WRITE_REGISTER_UCHAR((PUCHAR)(g_pMemBase + Addr), Value);//*LocalAddr = Value; //AVerMedia_Jeff_20190313
    U8_T pBuf[4] = { 0 };
    pBuf[0] = Value;
    I2CWrite(Addr, pBuf);
}

/*****************************************************************************/
/**
*
* @brief    Performs an output operation for a memory location by writing the
*            16 Bit Value to the the specified address.
*
* @param	Addr contains the address to perform the output operation
* @param	Value contains the Value to be written at the specified address.
*
* @return	None.
*
******************************************************************************/
void Xil_Out16(UINTPTR Addr, U16_T Value)
{
    CHECK_ADDR_WRITE;

	//volatile U16_T *LocalAddr = (volatile U16_T *)Addr; //AVerMedia_Jeff_20190313
	//WRITE_REGISTER_USHORT((PUSHORT)(g_pMemBase + Addr), Value);//*LocalAddr = Value; //AVerMedia_Jeff_20190313
    U8_T pBuf[4] = { 0 };
    pBuf[0] = Value & 0xff;
    pBuf[1] = (Value >> 8) & 0xff;
    I2CWrite(Addr, pBuf);
}

/*****************************************************************************/
/**
*
* @brief    Performs an output operation for a memory location by writing the
*           32 Bit Value to the the specified address.
*
* @param	Addr contains the address to perform the output operation
* @param	Value contains the 32 Bit Value to be written at the specified
*           address.
*
* @return	None.
*
******************************************************************************/
void Xil_Out32(UINTPTR Addr, U32_T Value)
{
    //CHECK_ADDR_WRITE;

#ifndef ENABLE_SAFETY
	//volatile U32_T *LocalAddr = (volatile U32_T *)Addr; //AVerMedia_Jeff_20190313
	//WRITE_REGISTER_ULONG((PULONG)(g_pMemBase + Addr), Value);//*LocalAddr = Value; //AVerMedia_Jeff_20190313
    U8_T pBuf[4] = { 0 };
    pBuf[0] = Value & 0xff;
    pBuf[1] = (Value >> 8) & 0xff;
    pBuf[2] = (Value >> 16) & 0xff;
    pBuf[3] = (Value >> 24) & 0xff;
    //mesg("jeff_i2c: ***I2C Write Addr(0x%08x) Data(0x%02X%02X%02X%02X) => %x \n", Addr, pBuf[3], pBuf[2], pBuf[1], pBuf[0], *((U32_T*)pBuf));
    I2CWrite(Addr, pBuf);
#else
	XStl_RegUpdate(Addr, Value);
#endif
}

/*****************************************************************************/
/**
*
* @brief    Performs an output operation for a memory location by writing the
*           64 Bit Value to the the specified address.
*
* @param	Addr contains the address to perform the output operation
* @param	Value contains 64 Bit Value to be written at the specified address.
*
* @return	None.
*
******************************************************************************/
static INLINE void Xil_Out64(UINTPTR Addr, U64_T Value)
{
    CHECK_ADDR_WRITE;

	//volatile U64_T *LocalAddr = (volatile U64_T *)Addr; //AVerMedia_Jeff_20190313
	//WRITE_REGISTER_ULONG64((PULONG64)(g_pMemBase + Addr), Value);//*LocalAddr = Value; //AVerMedia_Jeff_20190313
    U8_T pBuf[4] = { 0 };
    pBuf[0] = Value & 0xff;
    pBuf[1] = (Value >> 8) & 0xff;
    pBuf[2] = (Value >> 16) & 0xff;
    pBuf[3] = (Value >> 24) & 0xff;
    I2CWrite(Addr, pBuf);
}


#if defined (__MICROBLAZE__)
#ifdef __LITTLE_ENDIAN__
static INLINE U16_T Xil_In16BE(UINTPTR Addr)
#else
static INLINE U16_T Xil_In16LE(UINTPTR Addr)
#endif
#else
static INLINE U16_T Xil_In16BE(UINTPTR Addr)
#endif
{
	U16_T value = Xil_In16(Addr);
	return Xil_EndianSwap16(value);
}

#if defined (__MICROBLAZE__)
#ifdef __LITTLE_ENDIAN__
static INLINE U32_T Xil_In32BE(UINTPTR Addr)
#else
static INLINE U32_T Xil_In32LE(UINTPTR Addr)
#endif
#else
static INLINE U32_T Xil_In32BE(UINTPTR Addr)
#endif
{
	U32_T value = Xil_In32(Addr);
	return Xil_EndianSwap32(value);
}

#if defined (__MICROBLAZE__)
#ifdef __LITTLE_ENDIAN__
static INLINE void Xil_Out16BE(UINTPTR Addr, U16_T Value)
#else
static INLINE void Xil_Out16LE(UINTPTR Addr, U16_T Value)
#endif
#else
static INLINE void Xil_Out16BE(UINTPTR Addr, U16_T Value)
#endif
{
	Value = Xil_EndianSwap16(Value);
	Xil_Out16(Addr, Value);
}

#if defined (__MICROBLAZE__)
#ifdef __LITTLE_ENDIAN__
static INLINE void Xil_Out32BE(UINTPTR Addr, U32_T Value)
#else
static INLINE void Xil_Out32LE(UINTPTR Addr, U32_T Value)
#endif
#else
static INLINE void Xil_Out32BE(UINTPTR Addr, U32_T Value)
#endif
{
	Value = Xil_EndianSwap32(Value);
	Xil_Out32(Addr, Value);
}

/*****************************************************************************/
/**
*
* @brief    Perform a 16-Bit endian converion.
*
* @param	Data: 16 Bit value to be converted
*
* @return	16 Bit Data with converted endianess
*
******************************************************************************/
U16_T Xil_EndianSwap16(U16_T Data)
{
	return (U16_T) (((Data & 0xFF00U) >> 8U) | ((Data & 0x00FFU) << 8U));
}

/*****************************************************************************/
/**
*
* @brief    Perform a 32-Bit endian converion.
*
* @param	Data: 32 Bit value to be converted
*
* @return	32 Bit data with converted endianess
*
******************************************************************************/
U32_T Xil_EndianSwap32(U32_T Data)
{
	U16_T LoWord;
	U16_T HiWord;

	/* get each of the half words from the 32 Bit word */

	LoWord = (U16_T) (Data & 0x0000FFFFU);
	HiWord = (U16_T) ((Data & 0xFFFF0000U) >> 16U);

	/* byte swap each of the 16 Bit half words */

	LoWord = (((LoWord & 0xFF00U) >> 8U) | ((LoWord & 0x00FFU) << 8U));
	HiWord = (((HiWord & 0xFF00U) >> 8U) | ((HiWord & 0x00FFU) << 8U));

	/* swap the half words before returning the value */

	return ((((U32_T)LoWord) << (U32_T)16U) | (U32_T)HiWord);
}
