/*******************************************************************************
 *
 * Copyright (C) 2015 - 2016 Xilinx, Inc.  All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * Use of the Software is limited solely to applications:
 * (a) running on a Xilinx device, or
 * (b) that interact with a Xilinx device through a bus or interconnect.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * XILINX  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF
 * OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * Except as contained in this notice, the name of the Xilinx shall not be used
 * in advertising or otherwise to promote the sale, use or other dealings in
 * this Software without prior written authorization from Xilinx.
 *
*******************************************************************************/
/******************************************************************************/
/**
 *
 * @file xvphy_i.c
 *
 * Contains generic APIs that are locally called or used within the
 * VPHY driver.
 *
 * @note	None.
 *
 * <pre>
 * MODIFICATION HISTORY:
 *
 * Ver   Who  Date     Changes
 * ----- ---- -------- -----------------------------------------------
 * 1.0   gm,  11/09/16 Initial release.
 * 1.4   gm   29/11/16 Fixed c++ compiler warnings
 *                     Added xcvr adaptor functions for C++ compilations
 * 1.6   gm   06/08/17 Added XVphy_MmcmLocked, XVphy_ErrorHandler and
 *                              XVphy_PllLayoutErrorHandler APIs
 * 1.7   gm   13/09/17 Added GTYE4 support
 * </pre>
 *
*******************************************************************************/

/******************************* Include Files ********************************/

#include "xstatus.h"
#include "xvphy.h"
#include "xvphy_i.h"
#include "xvphy_hdmi.h"
#include "sleep.h"
#include "xvphy_gt.h"
#include "debug.h"
#include "xparameters.h"

/* DRP register space. */
#define XVPHY_DRP_RXCDR_CFG(n)		(0xA8 + n)

#define XVPHY_DRP_RXDATAWIDTH		0x11
#define XVPHY_DRP_OUT_DIV_PROG		0x88
#define XVPHY_DRP_TXCLK25		    0x6A
#define XVPHY_DRP_TXDATAWIDTH	    0x6B
#define XVPHY_DRP_RXCDR_CFG_WORD0	0xA8
#define XVPHY_DRP_RXCDR_CFG_WORD1	0xA9
#define XVPHY_DRP_RXCDR_CFG_WORD2	0xAA
#define XVPHY_DRP_RXCDR_CFG_WORD3	0xAB
#define XVPHY_DRP_RXCDR_CFG_WORD4	0xAC

/* PLL operating ranges. */
#define XVPHY_CPLL_MIN		1600000000LL
#define XVPHY_CPLL_MAX		3300000000LL

const U8_T Gtpe2CpllDivsM[]	= {1, 2, 0};
const U8_T Gtpe2CpllDivsN1[]	= {4, 5, 0};
const U8_T Gtpe2CpllDivsN2[]	= {1, 2, 3, 4, 5, 0};
const U8_T Gtpe2CpllDivsD[]	= {1, 2, 4, 8, 0};
#if 0
const XVphy_GtConfig Gtpe2Config = {
	.CfgSetCdr = XVphy_Gtpe2CfgSetCdr,
	.CheckPllOpRange = XVphy_Gtpe2CheckPllOpRange,
	.OutDivChReconfig = XVphy_Gtpe2OutDivChReconfig,
	.ClkChReconfig = XVphy_Gtpe2ClkChReconfig,
	.ClkCmnReconfig = XVphy_Gtpe2ClkCmnReconfig,
	.RxChReconfig = XVphy_Gtpe2RxChReconfig,
	.TxChReconfig = XVphy_Gtpe2TxChReconfig,

	.CpllDivs = {
		.M = Gtpe2CpllDivsM,
		.N1 = Gtpe2CpllDivsN1,
		.N2 = Gtpe2CpllDivsN2,
		.D = Gtpe2CpllDivsD,
	},
	.QpllDivs = {
		.M = Gtpe2CpllDivsM,
		.N1 = Gtpe2CpllDivsN1,
		.N2 = Gtpe2CpllDivsN2,
		.D = Gtpe2CpllDivsD,
	},
};
#endif
/**************************** Function Prototypes *****************************/


/**************************** Function Definitions ****************************/
static U8_T XVphy_MToDrpEncoding(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId);
static U16_T XVphy_NToDrpEncoding(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId, U8_T NId);
static U8_T XVphy_DToDrpEncoding(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId, XVphy_DirectionType Dir);
static U8_T XVphy_DrpEncodePllMN2(U8_T AttrEncode);
static U8_T XVphy_DrpEncodePllN1(U8_T AttrEncode);
static U8_T XVphy_DrpEncodeCpllTxRxD(U8_T AttrEncode);
static U16_T XVphy_DrpEncodeClk25(U32_T RefClkFreqHz);

U32_T XVphy_Gtpe2CfgSetCdr(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId);
U32_T XVphy_Gtpe2CheckPllOpRange(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId, U64_T PllClkOutFreqHz);
U32_T XVphy_Gtpe2OutDivChReconfig(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId, XVphy_DirectionType Dir);
U32_T XVphy_Gtpe2ClkChReconfig(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId);
U32_T XVphy_Gtpe2ClkCmnReconfig(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId CmnId);
U32_T XVphy_Gtpe2TxChReconfig(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId);
U32_T XVphy_Gtpe2RxChReconfig(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId);
U32_T XVphy_Gtpe2TxPllRefClkDiv1Reconfig(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId);
U32_T XVphy_Gtpe2RxPllRefClkDiv1Reconfig(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId);

U32_T XVphy_CfgSetCdr(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId);
U32_T XVphy_CheckPllOpRange(XVphy *InstancePtr, U8_T QuadId,
							XVphy_ChannelId ChId, U64_T PllClkOutFreqHz);
U32_T XVphy_OutDivChReconfig(XVphy *InstancePtr, U8_T QuadId,
							XVphy_ChannelId ChId, XVphy_DirectionType Dir);
U32_T XVphy_ClkChReconfig(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId);
U32_T XVphy_ClkCmnReconfig(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId);
U32_T XVphy_RxChReconfig(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId);
U32_T XVphy_TxChReconfig(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId);

/*****************************************************************************/
/**
* This function will set the clock and data recovery (CDR) values for a given
* channel.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
*
* @return
*		- XST_SUCCESS if the configuration was successful.
*		- XST_FAILURE otherwise.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_Gtpe2CfgSetCdr(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId)
{
	XVphy_Channel *ChPtr, *CmnPtr;
	XVphy_ChannelId CmnId;
	XVphy_PllType PllType;
	U32_T PllClkInFreqHz;

	/* Set CDR values only for CPLLs. */
	if ((ChId < XVPHY_CHANNEL_ID_CH1) || (ChId > XVPHY_CHANNEL_ID_CH4)) {
		return XST_FAILURE;
	}

	ChPtr = &InstancePtr->Quads[QuadId].Plls[XVPHY_CH2IDX(ChId)];

	ChPtr->PllParams.Cdr[0] = 0x1010;
	ChPtr->PllParams.Cdr[1] = 0x2104;
	ChPtr->PllParams.Cdr[3] = 0x07FE;

	if (InstancePtr->Config.RxProtocol == XVPHY_PROTOCOL_DP) {
		PllType = XVphy_GetPllType(InstancePtr, QuadId, XVPHY_DIR_RX, ChId);
		CmnId = XVphy_GetRcfgChId(InstancePtr, QuadId, XVPHY_DIR_RX, PllType);
		CmnPtr = &InstancePtr->Quads[QuadId].Plls[XVPHY_CH2IDX(CmnId)];
		PllClkInFreqHz = XVphy_GetQuadRefClkFreq(InstancePtr, QuadId,
					CmnPtr->CpllRefClkSel);

		if (PllClkInFreqHz == 270000000) {
		    ChPtr->PllParams.Cdr[2] = 0x4060;
			ChPtr->PllParams.Cdr[4] = 0x0001;
		}
		else if (PllClkInFreqHz == 135000000) {
		    ChPtr->PllParams.Cdr[2] = 0x2060;
			ChPtr->PllParams.Cdr[4] = 0x0011;
		}
		/* RBR does not use DP159 forwarded clock and expects 162MHz. */
		else {
		    ChPtr->PllParams.Cdr[2] = 0x2060;
			ChPtr->PllParams.Cdr[4] = 0x0001;
		}
	}

	else if (InstancePtr->Config.RxProtocol == XVPHY_PROTOCOL_HDMI) {

		ChPtr->PllParams.Cdr[4] = 0x0001;

		/* Update the RXCDR_CFG2 settings. */
		switch (ChPtr->RxOutDiv) {
			case 1:
				ChPtr->PllParams.Cdr[2] = 0x4060;
				break;
			case 2:
				ChPtr->PllParams.Cdr[2] = 0x2060;
				break;
			case 4:
				ChPtr->PllParams.Cdr[2] = 0x1060;
				break;
			case 8:
				ChPtr->PllParams.Cdr[2] = 0x0860;
				break;
			default:
				ChPtr->PllParams.Cdr[2] = 0x1060;
				break;
		}
	}

	return XST_SUCCESS;
}

/*****************************************************************************/
/**
* This function will check if a given PLL output frequency is within the
* operating range of the PLL for the GT type.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
* @param	PllClkOutFreqHz is the frequency to check.
*
* @return
*		- XST_SUCCESS if the frequency resides within the PLL's range.
*		- XST_FAILURE otherwise.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_Gtpe2CheckPllOpRange(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId, U64_T PllClkOutFreqHz)
{
	U32_T Status = XST_FAILURE;

	/* Suppress Warning Messages */
	InstancePtr = InstancePtr;
	QuadId = QuadId;
	ChId = ChId;

	if ((XVPHY_CPLL_MIN <= PllClkOutFreqHz) &&
			(PllClkOutFreqHz <= XVPHY_CPLL_MAX)) {
		Status = XST_SUCCESS;
	}

	return Status;
}

/*****************************************************************************/
/**
* This function will set the output divider logic for a given channel.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
* @param	Dir is an indicator for RX or TX.
*
* @return
*		- XST_SUCCESS if the configuration was successful.
*		- XST_FAILURE otherwise.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_Gtpe2OutDivChReconfig(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId, XVphy_DirectionType Dir)
{
	U16_T DrpVal;
	U16_T WriteVal;
    U32_T Status = XST_SUCCESS;

	/* Obtain current DRP register value for TX/RX dividers. */
	Status |= XVphy_DrpRd(InstancePtr, QuadId, ChId, 0x88, &DrpVal);

	if (Dir == XVPHY_DIR_RX) {
		/* Mask out RX_OUT_DIV. */
		DrpVal &= ~0x07;
		/* Set RX_OUT_DIV. */
		WriteVal = (XVphy_DToDrpEncoding(InstancePtr, QuadId, ChId,
				XVPHY_DIR_RX) & 0x7);
		DrpVal |= WriteVal;
	}
	else {
		/* Mask out TX_OUT_DIV. */
		DrpVal &= ~0x70;
		/* Set TX_OUT_DIV. */
		WriteVal = (XVphy_DToDrpEncoding(InstancePtr, QuadId, ChId,
				XVPHY_DIR_TX) & 0x7);
		DrpVal |= (WriteVal << 4);
	}
	/* Write new DRP register value for TX/RX dividers. */
	Status |= XVphy_DrpWr(InstancePtr, QuadId, ChId, 0x88, DrpVal);

	return Status;
}

/*****************************************************************************/
/**
* This function will configure the channel clock settings.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
*
* @return
*		- XST_FAILURE otherwise.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_Gtpe2ClkChReconfig(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId)
{
	/* Suppress Warning Messages */
	InstancePtr = InstancePtr;
	QuadId = QuadId;
	ChId = ChId;

	/* GTPE2 doesn't have channel PLL. */
	return XST_FAILURE;
}

/*****************************************************************************/
/**
* This function will configure the common channel clock settings.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	CmnId is the common channel ID to operate on.
*
* @return
*		- XST_SUCCESS if the configuration was successful.
*		- XST_FAILURE otherwise.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_Gtpe2ClkCmnReconfig(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId CmnId)
{
	U16_T DrpVal;
	U16_T WriteVal;
    U32_T Status = XST_SUCCESS;

	/* Obtain current DRP register value for PLL dividers. */
	Status |= XVphy_DrpRd(InstancePtr, QuadId, XVPHY_CHANNEL_ID_CMN,
			(CmnId == XVPHY_CHANNEL_ID_CMN0) ? 0x04 : 0x2B, &DrpVal);
	/* Mask out clock divider bits. */
	DrpVal &= ~(0x1FFF);
	/* Set CPLL_FBDIV. */
	WriteVal = (XVphy_NToDrpEncoding(InstancePtr, QuadId, CmnId, 2) & 0x3F);
	DrpVal |= WriteVal;
	/* Set CPLL_FBDIV_45. */
	WriteVal = (XVphy_NToDrpEncoding(InstancePtr, QuadId, CmnId, 1) & 0x1);
	DrpVal |= (WriteVal << 7);
	/* Set CPLL_REFCLKDIV. */
	WriteVal = (XVphy_MToDrpEncoding(InstancePtr, QuadId, CmnId) & 0x1F);
	DrpVal |= (WriteVal << 9);
	/* Write new DRP register value for PLL dividers. */
	Status |= XVphy_DrpWr(InstancePtr, QuadId, XVPHY_CHANNEL_ID_CMN,
			(CmnId == XVPHY_CHANNEL_ID_CMN0) ? 0x04 : 0x2B, DrpVal);

	return Status;
}

/*****************************************************************************/
/**
* This function will configure the channel's RX settings.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
*
* @return
*		- XST_SUCCESS if the configuration was successful.
*		- XST_FAILURE otherwise.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_Gtpe2RxChReconfig(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId)
{
	XVphy_Channel *ChPtr;
	U16_T DrpVal;
	U8_T CfgIndex;
    U32_T Status = XST_SUCCESS;

	ChPtr = &InstancePtr->Quads[QuadId].Plls[XVPHY_CH2IDX(ChId)];

	/* RXCDR_CFG(CfgIndex) */
	for (CfgIndex = 0; CfgIndex < 5; CfgIndex++) {
		DrpVal = ChPtr->PllParams.Cdr[CfgIndex];
		if (!DrpVal) {
			/* Don't modify RX_CDR configuration. */
			continue;
		}
		Status |= XVphy_DrpWr(InstancePtr, QuadId, ChId,
				XVPHY_DRP_RXCDR_CFG(CfgIndex), DrpVal);
	}

	if ((InstancePtr->Config.RxProtocol == XVPHY_PROTOCOL_HDMI) &&
	    (InstancePtr->HdmiRxDruIsEnabled)) {
		Status |= XVphy_DrpWr(InstancePtr, QuadId, ChId, 0x2A,0x0);
		Status |= XVphy_DrpWr(InstancePtr, QuadId, ChId, 0x2B,0x0);

		/* Set RX_DATA_WIDTH to 20 bits */
		Status |= XVphy_DrpRd(InstancePtr, QuadId, ChId,
				XVPHY_DRP_RXDATAWIDTH, &DrpVal);
		DrpVal &= ~0x3800;
		DrpVal |= (3 << 11);
		Status |= XVphy_DrpWr(InstancePtr, QuadId, ChId,
                XVPHY_DRP_RXDATAWIDTH, DrpVal);
	}

	Status |= XVphy_Gtpe2RxPllRefClkDiv1Reconfig(InstancePtr, QuadId, ChId);

	return Status;
}

/*****************************************************************************/
/**
* This function will configure the channel's TX settings.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
*
* @return
*		- XST_SUCCESS if the configuration was successful.
*		- XST_FAILURE otherwise.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_Gtpe2TxChReconfig(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId)
{
	U16_T DrpVal;
    U32_T Status = XST_SUCCESS;

	if (InstancePtr->Config.TxProtocol == XVPHY_PROTOCOL_HDMI) {
        Status |= XVphy_DrpRd(InstancePtr, QuadId, ChId,
                XVPHY_DRP_TXDATAWIDTH, &DrpVal);
        DrpVal &= ~(0x07);

		if (InstancePtr->Config.TransceiverWidth == 2) {
			/* Set TX Data width of the TX GT to 20 bits */
			DrpVal |= 3;
		}
		else {
			/* Set TX Data width of the TX GT to 40 bits */
			DrpVal |= 5;
		}
		Status |= XVphy_DrpWr(InstancePtr, QuadId, ChId,
                    XVPHY_DRP_TXDATAWIDTH, DrpVal);
	}
    Status |= XVphy_Gtpe2TxPllRefClkDiv1Reconfig(InstancePtr, QuadId, ChId);

	return Status;
}

/*****************************************************************************/
/**
* This function will configure the channel's TX CLKDIV1 settings.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
*
* @return
*		- XST_SUCCESS if the configuration was successful.
*		- XST_FAILURE otherwise.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_Gtpe2TxPllRefClkDiv1Reconfig(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId)
{
	U16_T DrpVal;
	U32_T TxRefClkHz;
	XVphy_Channel *CmnPtr;
	XVphy_ChannelId CmnId;
	XVphy_PllType PllType;
    U32_T Status = XST_SUCCESS;

	if (InstancePtr->Config.TxProtocol == XVPHY_PROTOCOL_HDMI) {
		TxRefClkHz = InstancePtr->HdmiTxRefClkHz;
	}
	else {
		PllType = XVphy_GetPllType(InstancePtr, QuadId, XVPHY_DIR_TX, ChId);
		CmnId = XVphy_GetRcfgChId(InstancePtr, QuadId, XVPHY_DIR_TX, PllType);
		CmnPtr = &InstancePtr->Quads[QuadId].Plls[XVPHY_CH2IDX(CmnId)];
		TxRefClkHz = XVphy_GetQuadRefClkFreq(InstancePtr, QuadId,
								CmnPtr->PllRefClkSel);
	}

	Status |= XVphy_DrpRd(InstancePtr, QuadId, ChId, XVPHY_DRP_TXCLK25, &DrpVal);
	DrpVal &= ~(0x1F);
	DrpVal |= XVphy_DrpEncodeClk25(TxRefClkHz);
	Status |= XVphy_DrpWr(InstancePtr, QuadId, ChId, XVPHY_DRP_TXCLK25,
                    DrpVal);
	return Status;
}

/*****************************************************************************/
/**
* This function will configure the channel's RX CLKDIV1 settings.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
*
* @return
*		- XST_SUCCESS if the configuration was successful.
*		- XST_FAILURE otherwise.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_Gtpe2RxPllRefClkDiv1Reconfig(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId)
{
	U16_T DrpVal;
	U32_T RxRefClkHz;
	XVphy_Channel *CmnPtr;
	XVphy_ChannelId CmnId;
	XVphy_PllType PllType;
    U32_T Status = XST_SUCCESS;

	if (InstancePtr->Config.RxProtocol == XVPHY_PROTOCOL_HDMI) {
		RxRefClkHz = InstancePtr->HdmiRxRefClkHz;
	}
	else {
		PllType = XVphy_GetPllType(InstancePtr, QuadId, XVPHY_DIR_RX, ChId);
		CmnId = XVphy_GetRcfgChId(InstancePtr, QuadId, XVPHY_DIR_RX, PllType);
		CmnPtr = &InstancePtr->Quads[QuadId].Plls[XVPHY_CH2IDX(CmnId)];
		RxRefClkHz = XVphy_GetQuadRefClkFreq(InstancePtr, QuadId,
								CmnPtr->PllRefClkSel);
	}

	Status |= XVphy_DrpRd(InstancePtr, QuadId, ChId, XVPHY_DRP_RXDATAWIDTH,
                &DrpVal);
	DrpVal &= ~(0x07C0);
	DrpVal |= XVphy_DrpEncodeClk25(RxRefClkHz) << 6;
	Status |= XVphy_DrpWr(InstancePtr, QuadId, ChId, XVPHY_DRP_RXDATAWIDTH,
                DrpVal);
	return Status;
}

/*****************************************************************************/
/**
* This function will translate the configured M value to DRP encoding.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
*
* @return	The DRP encoding for M.
*
* @note		None.
*
******************************************************************************/
static U8_T XVphy_MToDrpEncoding(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId)
{
	U8_T MRefClkDiv;
	U8_T DrpEncode;

	if ((ChId == XVPHY_CHANNEL_ID_CMN0) ||
			(ChId == XVPHY_CHANNEL_ID_CMN1)) {
		MRefClkDiv = InstancePtr->Quads[QuadId].Plls[XVPHY_CH2IDX(ChId)]
				.PllParams.MRefClkDiv;
	}
	else {
		MRefClkDiv = 0;
	}

	DrpEncode = XVphy_DrpEncodePllMN2(MRefClkDiv);

	return DrpEncode;
}

/*****************************************************************************/
/**
* This function will translate the configured D value to DRP encoding.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
* @param	Dir is an indicator for RX or TX.
*
* @return	The DRP encoding for D.
*
* @note		None.
*
******************************************************************************/
static U8_T XVphy_DToDrpEncoding(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId, XVphy_DirectionType Dir)
{
	U8_T OutDiv;
	U8_T DrpEncode;

	OutDiv = InstancePtr->Quads[QuadId].Plls[XVPHY_CH2IDX(ChId)].
			OutDiv[Dir];

	DrpEncode = XVphy_DrpEncodeCpllTxRxD(OutDiv);

	return DrpEncode;
}

/*****************************************************************************/
/**
* This function will translate the configured N1/N2 value to DRP encoding.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
* @param	NId specified to operate on N1 (if == 1) or N2 (if == 2).
*
* @return	The DRP encoding for N1/N2.
*
* @note		None.
*
******************************************************************************/
static U16_T XVphy_NToDrpEncoding(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId, U8_T NId)
{
	U8_T NFbDiv;
	U16_T DrpEncode;

	if (NId == 1) {
		NFbDiv = InstancePtr->Quads[QuadId].Plls[XVPHY_CH2IDX(ChId)].
				PllParams.N1FbDiv;
		DrpEncode = XVphy_DrpEncodePllN1(NFbDiv);
	}
	else {
		NFbDiv = InstancePtr->Quads[QuadId].Plls[XVPHY_CH2IDX(ChId)].
			PllParams.N2FbDiv;
		DrpEncode = XVphy_DrpEncodePllMN2(NFbDiv);
	}

	return DrpEncode;
}

/*****************************************************************************/
/**
* This function will translate the configured QPLL's M or CPLL's M or N2
* values to DRP encoding.
*
* @param	AttrEncode is the attribute to encode.
*
* @return	The DRP encoding for the QPLL's M or CPLL's M or N2 values.
*
* @note		None.
*
******************************************************************************/
static U8_T XVphy_DrpEncodePllMN2(U8_T AttrEncode)
{
	U8_T DrpEncode;

	switch (AttrEncode) {
	case 1:
		DrpEncode = 16;
		break;
	case 6:
		DrpEncode = 5;
		break;
	case 10:
		DrpEncode = 7;
		break;
	case 12:
		DrpEncode = 13;
		break;
	case 20:
		DrpEncode = 15;
		break;
	case 2:
	case 3:
	case 4:
	case 5:
	case 8:
	case 16:
		DrpEncode = (AttrEncode - 2);
		break;
	default:
		DrpEncode = 0xF;
		break;
	}

	return DrpEncode;
}

/*****************************************************************************/
/**
* This function will translate the configured PLL's N1 value to DRP encoding.
*
* @param	AttrEncode is the attribute to encode.
*
* @return	The DRP encoding for the PLL's N1 value.
*
* @note		None.
*
******************************************************************************/
static U8_T XVphy_DrpEncodePllN1(U8_T AttrEncode)
{
	U8_T DrpEncode;

	DrpEncode = (AttrEncode - 4) & 0x1;

	return DrpEncode;
}

/*****************************************************************************/
/**
* This function will translate the configured CPLL's D values to DRP encoding.
*
* @param	AttrEncode is the attribute to encode.
*
* @return	The DRP encoding for the CPLL's D value.
*
* @note		None.
*
******************************************************************************/
static U8_T XVphy_DrpEncodeCpllTxRxD(U8_T AttrEncode)
{
	U8_T DrpEncode;

	switch (AttrEncode) {
	case 1:
		DrpEncode = 0;
		break;
	case 2:
		DrpEncode = 1;
		break;
	case 4:
		DrpEncode = 2;
		break;
	case 8:
		DrpEncode = 3;
		break;
	case 16:
		DrpEncode = 4;
		break;
	default:
		DrpEncode = 0x4;
		break;
	}

	return DrpEncode;
}

/*****************************************************************************/
/**
* This function will translate the configured CLK25 to DRP encoding.
*
* @param	AttrEncode is the attribute to encode.
*
* @return	The DRP encoding for the CLK25 value.
*
* @note		None.
*
******************************************************************************/
static U16_T XVphy_DrpEncodeClk25(U32_T RefClkFreqHz)
{
	U16_T DrpEncode;
	U32_T RefClkFreqMHz = RefClkFreqHz / 1000000;

	DrpEncode = ((RefClkFreqMHz / 25) +
			(((RefClkFreqMHz % 25) > 0) ? 1 : 0)) - 1;

	return (DrpEncode & 0x1F);
}
/*****************************************************************************/
/**
* This function will enable or disable the LPM logic in the Video PHY core.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
* @param	Dir is an indicator for TX or RX.
* @param	Enable will enable (if 1) or disable (if 0) the LPM logic.
*
* @return	None.
*
* @note		None.
*
******************************************************************************/
void XVphy_SetRxLpm(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId,
		XVphy_DirectionType Dir, U8_T Enable)
{
	U32_T RegVal;
	U32_T MaskVal;

	/* Suppress Warning Messages */
	QuadId = QuadId;
	Dir = Dir;

	RegVal = XVphy_ReadReg(InstancePtr->Config.BaseAddr,
							XVPHY_RX_EQ_CDR_REG);

	if (ChId == XVPHY_CHANNEL_ID_CHA) {
		MaskVal = XVPHY_RX_CONTROL_RXLPMEN_ALL_MASK;
	}
	else {
		MaskVal = XVPHY_RX_CONTROL_RXLPMEN_MASK(ChId);
	}

	if (Enable) {
		RegVal |= MaskVal;
	}
	else {
		RegVal &= ~MaskVal;
	}
	XVphy_WriteReg(InstancePtr->Config.BaseAddr, XVPHY_RX_EQ_CDR_REG,
									RegVal);
}

/*****************************************************************************/
/**
* This function will set the TX voltage swing value for a given channel.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
* @param	Vs is the voltage swing value to write.
*
* @return	None.
*
* @note		None.
*
******************************************************************************/
void XVphy_SetTxVoltageSwing(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId, U8_T Vs)
{
	U32_T RegVal;
	U32_T MaskVal;
	U32_T RegOffset;

	/* Suppress Warning Messages */
	QuadId = QuadId;

	if ((ChId == XVPHY_CHANNEL_ID_CH1) || (ChId == XVPHY_CHANNEL_ID_CH2)) {
		RegOffset = XVPHY_TX_DRIVER_CH12_REG;
	}
	else {
		RegOffset = XVPHY_TX_DRIVER_CH34_REG;
	}
	RegVal = XVphy_ReadReg(InstancePtr->Config.BaseAddr, RegOffset);

	MaskVal = XVPHY_TX_DRIVER_TXDIFFCTRL_MASK(ChId);
	RegVal &= ~MaskVal;
	RegVal |= (Vs << XVPHY_TX_DRIVER_TXDIFFCTRL_SHIFT(ChId));
	XVphy_WriteReg(InstancePtr->Config.BaseAddr, RegOffset, RegVal);
}

/*****************************************************************************/
/**
* This function will set the TX pre-emphasis value for a given channel.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
* @param	Pe is the pre-emphasis value to write.
*
* @return	None.
*
* @note		None.
*
******************************************************************************/
void XVphy_SetTxPreEmphasis(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId,
		U8_T Pe)
{
	U32_T RegVal;
	U32_T MaskVal;
	U32_T RegOffset;

	/* Suppress Warning Messages */
	QuadId = QuadId;

	if ((ChId == XVPHY_CHANNEL_ID_CH1) || (ChId == XVPHY_CHANNEL_ID_CH2)) {
		RegOffset = XVPHY_TX_DRIVER_CH12_REG;
	}
	else {
		RegOffset = XVPHY_TX_DRIVER_CH34_REG;
	}
	RegVal = XVphy_ReadReg(InstancePtr->Config.BaseAddr, RegOffset);

	MaskVal = XVPHY_TX_DRIVER_TXPRECURSOR_MASK(ChId);
	RegVal &= ~MaskVal;
	RegVal |= (Pe << XVPHY_TX_DRIVER_TXPRECURSOR_SHIFT(ChId));
	XVphy_WriteReg(InstancePtr->Config.BaseAddr, RegOffset, RegVal);
}

/*****************************************************************************/
/**
* This function writes the current software configuration for the reference
* clock selections to hardware for the specified quad on all channels.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
*
* @return
*		- XST_SUCCESS.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_WriteCfgRefClkSelReg(XVphy *InstancePtr, U8_T QuadId)
{
	U32_T RegVal = 0;
	XVphy_Channel *ChPtr;
	XVphy_GtType GtType = InstancePtr->Config.XcvrType;

	/* Point to the first channel since settings apply to all channels. */
	ChPtr = &InstancePtr->Quads[QuadId].Ch1;

	/* PllRefClkSel. */
	/* - QPLL0. */
	RegVal &= ~XVPHY_REF_CLK_SEL_QPLL0_MASK;

	RegVal = InstancePtr->Quads[QuadId].Cmn0.PllRefClkSel;
	/* - CPLL. */
	RegVal &= ~XVPHY_REF_CLK_SEL_CPLL_MASK;
	RegVal |= (ChPtr->CpllRefClkSel << XVPHY_REF_CLK_SEL_CPLL_SHIFT);

	if ((GtType == XVPHY_GT_TYPE_GTHE3) ||
            (GtType == XVPHY_GT_TYPE_GTHE4) ||
            (GtType == XVPHY_GT_TYPE_GTYE4) ||
            (GtType == XVPHY_GT_TYPE_GTPE2)) {
		/* - QPLL1. */
		RegVal &= ~XVPHY_REF_CLK_SEL_QPLL1_MASK;
		RegVal |= (InstancePtr->Quads[QuadId].Cmn1.PllRefClkSel <<
				XVPHY_REF_CLK_SEL_QPLL1_SHIFT);
	}

	/* SysClkDataSel. PLLCLKSEL */
	RegVal &= ~XVPHY_REF_CLK_SEL_SYSCLKSEL_MASK;
	/* - TXSYSCLKSEL[0]. TXPLLCLKSEL*/
	RegVal |= (ChPtr->TxDataRefClkSel <<
		XVPHY_REF_CLK_SEL_TXSYSCLKSEL_DATA_SHIFT(GtType)) &
		XVPHY_REF_CLK_SEL_TXSYSCLKSEL_DATA_MASK(GtType);

	/* - RXSYSCLKSEL[0]. RXPLLCLKSEL*/
	RegVal |= (ChPtr->RxDataRefClkSel <<
		XVPHY_REF_CLK_SEL_RXSYSCLKSEL_DATA_SHIFT(GtType)) &
		XVPHY_REF_CLK_SEL_RXSYSCLKSEL_DATA_MASK(GtType);

	/* SysClkOutSel. */
	/* - TXSYSCLKSEL[1]. */
	RegVal |= (ChPtr->TxOutRefClkSel <<
		XVPHY_REF_CLK_SEL_TXSYSCLKSEL_OUT_SHIFT(GtType)) &
		XVPHY_REF_CLK_SEL_TXSYSCLKSEL_OUT_MASK(GtType);

	/* - RXSYSCLKSEL[1]. */
	RegVal |= (ChPtr->RxOutRefClkSel <<
		XVPHY_REF_CLK_SEL_RXSYSCLKSEL_OUT_SHIFT(GtType)) &
		XVPHY_REF_CLK_SEL_RXSYSCLKSEL_OUT_MASK(GtType);

	/* Write to hardware. */
	XVphy_WriteReg(InstancePtr->Config.BaseAddr, XVPHY_REF_CLK_SEL_REG,
			RegVal);

	return XST_SUCCESS;
}

/*****************************************************************************/
/**
* Configure the PLL reference clock selection for the specified channel(s).
* This is applied to both direction to the software configuration only.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
* @param	SysClkDataSel is the reference clock selection to configure.
*
* @return	None.
*
* @note		None.
*
******************************************************************************/
void XVphy_CfgPllRefClkSel(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId,
		XVphy_PllRefClkSelType RefClkSel)
{
	U8_T Id, Id0, Id1;

	XVphy_Ch2Ids(InstancePtr, ChId, &Id0, &Id1);
	for (Id = Id0; Id <= Id1; Id++) {
		InstancePtr->Quads[QuadId].Plls[XVPHY_CH2IDX(Id)].PllRefClkSel =
					RefClkSel;
	}
}

/*****************************************************************************/
/**
* Configure the SYSCLKDATA reference clock selection for the direction. Same
* configuration applies to all channels in the quad. This is applied to the
* software configuration only.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	Dir is an indicator for TX or RX.
* @param	SysClkDataSel is the reference clock selection to configure.
*
* @return	None.
*
* @note		None.
*
******************************************************************************/
void XVphy_CfgSysClkDataSel(XVphy *InstancePtr, U8_T QuadId,
		XVphy_DirectionType Dir, XVphy_SysClkDataSelType SysClkDataSel)
{
	XVphy_Channel *ChPtr;
	U8_T Id, Id0, Id1;

	XVphy_Ch2Ids(InstancePtr, XVPHY_CHANNEL_ID_CHA, &Id0, &Id1);
	/* Select in software - same for all channels. */
	for (Id = Id0; Id <= Id1; Id++) {
		ChPtr = &InstancePtr->Quads[QuadId].Plls[XVPHY_CH2IDX(Id)];
		ChPtr->DataRefClkSel[Dir] = SysClkDataSel;
	}
}

/*****************************************************************************/
/**
* Configure the SYSCLKOUT reference clock selection for the direction. Same
* configuration applies to all channels in the quad. This is applied to the
* software configuration only.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	Dir is an indicator for TX or RX.
* @param	SysClkOutSel is the reference clock selection to configure.
*
* @return	None.
*
* @note		None.
*
******************************************************************************/
void XVphy_CfgSysClkOutSel(XVphy *InstancePtr, U8_T QuadId,
		XVphy_DirectionType Dir, XVphy_SysClkOutSelType SysClkOutSel)
{
	XVphy_Channel *ChPtr;
	U8_T Id, Id0, Id1;

	XVphy_Ch2Ids(InstancePtr, XVPHY_CHANNEL_ID_CHA, &Id0, &Id1);
	/* Select in software - same for all channels. */
	for (Id = Id0; Id <= Id1; Id++) {
		ChPtr = &InstancePtr->Quads[QuadId].Plls[XVPHY_CH2IDX(Id)];
		ChPtr->OutRefClkSel[Dir] = SysClkOutSel;
	}
}

/*****************************************************************************/
/**
* Obtain the reconfiguration channel ID for given PLL type
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	Dir is an indicator for TX or RX.
* @param	PllType is the PLL type being used by the channel.
*
* @return	The Channel ID to be used for reconfiguration
*
* @note		None.
*
******************************************************************************/
XVphy_ChannelId XVphy_GetRcfgChId(XVphy *InstancePtr, U8_T QuadId,
		XVphy_DirectionType Dir, XVphy_PllType PllType)
{
	XVphy_ChannelId ChId;

	/* Suppress Warning Messages */
	InstancePtr = InstancePtr;
	QuadId = QuadId;
	Dir = Dir;

	/* Determine which channel(s) to operate on. */
	switch (PllType) {
	case XVPHY_PLL_TYPE_QPLL:
	case XVPHY_PLL_TYPE_QPLL0:
	case XVPHY_PLL_TYPE_PLL0:
		ChId = XVPHY_CHANNEL_ID_CMN0;
		break;
	case XVPHY_PLL_TYPE_QPLL1:
	case XVPHY_PLL_TYPE_PLL1:
		ChId = XVPHY_CHANNEL_ID_CMN1;
		break;
	default:
		ChId = XVPHY_CHANNEL_ID_CHA;
		break;
	}

	return ChId;
}

/*****************************************************************************/
/**
* Obtain the current reference clock frequency for the quad based on the
* reference clock type.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	RefClkType is the type to obtain the clock selection for.
*
* @return	The current reference clock frequency for the quad for the
*		specified type selection.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_GetQuadRefClkFreq(XVphy *InstancePtr, U8_T QuadId,
		XVphy_PllRefClkSelType RefClkType)
{
	U32_T FreqHz;

	U8_T RefClkIndex = RefClkType - XVPHY_PLL_REFCLKSEL_TYPE_GTREFCLK0;

	FreqHz = (RefClkType > XVPHY_PLL_REFCLKSEL_TYPE_GTGREFCLK) ? 0 :
		InstancePtr->Quads[QuadId].RefClkHz[RefClkIndex];

	return FreqHz;
}

/*****************************************************************************/
/**
* Obtain the current [RT]XSYSCLKSEL[0] configuration.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	Dir is an indicator for TX or RX.
* @param	ChId is the channel ID which to operate on.
*
* @return	The current [RT]XSYSCLKSEL[0] selection.
*
* @note		None.
*
******************************************************************************/
XVphy_SysClkDataSelType XVphy_GetSysClkDataSel(XVphy *InstancePtr, U8_T QuadId,
		XVphy_DirectionType Dir, XVphy_ChannelId ChId)
{
	U32_T Sel;
	U32_T RegVal;

	/* Suppress Warning Messages */
	QuadId = QuadId;
	ChId = ChId;

	RegVal = XVphy_ReadReg(InstancePtr->Config.BaseAddr,
							XVPHY_REF_CLK_SEL_REG);

	if (Dir == XVPHY_DIR_TX) {
		/* Synchronize software configuration to hardware. */
		Sel = RegVal & XVPHY_REF_CLK_SEL_TXSYSCLKSEL_DATA_MASK(
				InstancePtr->Config.XcvrType);
		Sel >>= XVPHY_REF_CLK_SEL_TXSYSCLKSEL_DATA_SHIFT(
				InstancePtr->Config.XcvrType);
	}
	else {
		/* Synchronize software configuration to hardware. */
		Sel = RegVal & XVPHY_REF_CLK_SEL_RXSYSCLKSEL_DATA_MASK(
				InstancePtr->Config.XcvrType);
		Sel >>= XVPHY_REF_CLK_SEL_RXSYSCLKSEL_DATA_SHIFT(
				InstancePtr->Config.XcvrType);
	}

	return (XVphy_SysClkDataSelType) Sel;
}

/*****************************************************************************/
/**
* Obtain the current [RT]XSYSCLKSEL[1] configuration.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	Dir is an indicator for TX or RX.
* @param	ChId is the channel ID which to operate on.
*
* @return	The current [RT]XSYSCLKSEL[1] selection.
*
* @note		None.
*
******************************************************************************/
XVphy_SysClkOutSelType XVphy_GetSysClkOutSel(XVphy *InstancePtr, U8_T QuadId,
		XVphy_DirectionType Dir, XVphy_ChannelId ChId)
{
	U32_T Sel;
	U32_T RegVal;

	/* Suppress Warning Messages */
	QuadId = QuadId;
	ChId = ChId;

	RegVal = XVphy_ReadReg(InstancePtr->Config.BaseAddr,
			XVPHY_REF_CLK_SEL_REG);

	if (Dir == XVPHY_DIR_TX) {
		/* Synchronize software configuration to hardware. */
		Sel = RegVal & XVPHY_REF_CLK_SEL_TXSYSCLKSEL_OUT_MASK(
				InstancePtr->Config.XcvrType);
		Sel >>= XVPHY_REF_CLK_SEL_TXSYSCLKSEL_OUT_SHIFT(
				InstancePtr->Config.XcvrType);
	}
	else {
		/* Synchronize software configuration to hardware. */
		Sel = RegVal & XVPHY_REF_CLK_SEL_RXSYSCLKSEL_OUT_MASK(
				InstancePtr->Config.XcvrType);
		Sel >>= XVPHY_REF_CLK_SEL_RXSYSCLKSEL_OUT_SHIFT(
				InstancePtr->Config.XcvrType);
	}

	return (XVphy_SysClkOutSelType)Sel;
}

/*****************************************************************************/
/**
* This function will check the status of a PLL lock on the specified channel.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID which to operate on.
*
* @return
*		- XST_SUCCESS if the specified PLL is locked.
*		- XST_FAILURE otherwise.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_IsPllLocked(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId)
{
	U32_T RegVal;
	U32_T MaskVal;
	XVphy_PllType TxPllType;
	XVphy_PllType RxPllType;

	/* Suppress Warning Messages */
	QuadId = QuadId;

	if (ChId == XVPHY_CHANNEL_ID_CMN0) {
		MaskVal = XVPHY_PLL_LOCK_STATUS_QPLL0_MASK;
	}
	else if (ChId == XVPHY_CHANNEL_ID_CMN1) {
		MaskVal = XVPHY_PLL_LOCK_STATUS_QPLL1_MASK;
	}
	else if (ChId == XVPHY_CHANNEL_ID_CMNA) {
		MaskVal = XVPHY_PLL_LOCK_STATUS_QPLL0_MASK |
			  XVPHY_PLL_LOCK_STATUS_QPLL1_MASK;
	}
	else if (ChId == XVPHY_CHANNEL_ID_CHA) {
		TxPllType = XVphy_GetPllType(InstancePtr, 0, XVPHY_DIR_TX,
				XVPHY_CHANNEL_ID_CH1);
		RxPllType = XVphy_GetPllType(InstancePtr, 0, XVPHY_DIR_RX,
				XVPHY_CHANNEL_ID_CH1);
		if (RxPllType == XVPHY_PLL_TYPE_CPLL &&
				InstancePtr->Config.RxProtocol == XVPHY_PROTOCOL_HDMI) {
			MaskVal = XVPHY_PLL_LOCK_STATUS_CPLL_HDMI_MASK;
		}
		else if (TxPllType == XVPHY_PLL_TYPE_CPLL &&
				InstancePtr->Config.TxProtocol == XVPHY_PROTOCOL_HDMI) {
			MaskVal = XVPHY_PLL_LOCK_STATUS_CPLL_HDMI_MASK;
		}
		else {
			MaskVal = XVPHY_PLL_LOCK_STATUS_CPLL_ALL_MASK;
		}
	}
	else {
		MaskVal = XVPHY_PLL_LOCK_STATUS_CPLL_MASK(ChId);
	}
	RegVal = XVphy_ReadReg(InstancePtr->Config.BaseAddr,
			XVPHY_PLL_LOCK_STATUS_REG);

	if ((RegVal & MaskVal) == MaskVal) {
		return XST_SUCCESS;
	}

	return XST_FAILURE;
}

/*****************************************************************************/
/**
* This function will reset and enable the Video PHY's user core logic.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID which to operate on.
* @param	Dir is an indicator for TX or RX.
* @param	Hold is an indicator whether to "hold" the reset if set to 1.
*		If set to 0: reset, then enable.
*
* @return
*		- XST_SUCCESS.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_GtUserRdyEnable(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId,
		XVphy_DirectionType Dir, U8_T Hold)
{
	U32_T RegVal;
	U32_T MaskVal;
	U32_T RegOffset;

	/* Suppress Warning Messages */
	QuadId = QuadId;

	if (Dir == XVPHY_DIR_TX) {
		RegOffset = XVPHY_TX_INIT_REG;

		if (ChId == XVPHY_CHANNEL_ID_CHA) {
			MaskVal = XVPHY_TX_INIT_USERRDY_ALL_MASK;
		}
		else {
			MaskVal = XVPHY_TX_INIT_USERRDY_MASK(ChId);
		}
	}
	else {
		RegOffset = XVPHY_RX_INIT_REG;
		if (ChId == XVPHY_CHANNEL_ID_CHA) {
			MaskVal = XVPHY_RX_INIT_USERRDY_ALL_MASK;
		}
		else {
			MaskVal = XVPHY_RX_INIT_USERRDY_MASK(ChId);
		}
	}

	RegVal = XVphy_ReadReg(InstancePtr->Config.BaseAddr, RegOffset);
	/* Assert reset. */
	RegVal |= MaskVal;
	XVphy_WriteReg(InstancePtr->Config.BaseAddr, RegOffset, RegVal);

	if (!Hold) {
		/* De-assert reset. */
		RegVal &= ~MaskVal;
		XVphy_WriteReg(InstancePtr->Config.BaseAddr, RegOffset, RegVal);
	}

	return XST_SUCCESS;
}

/*****************************************************************************/
/**
* This function will reset the mixed-mode clock manager (MMCM) core.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	Dir is an indicator for TX or RX.
* @param	Hold is an indicator whether to "hold" the reset if set to 1.
*		If set to 0: reset, then enable.
*
* @return
*		- XST_SUCCESS.
*
* @note		None.
*
******************************************************************************/
void XVphy_MmcmReset(XVphy *InstancePtr, U8_T QuadId, XVphy_DirectionType Dir,
		U8_T Hold)
{
	U32_T RegOffsetCtrl;
	U32_T RegVal;

	/* Suppress Warning Messages */
	QuadId = QuadId;

	if (Dir == XVPHY_DIR_TX) {
		RegOffsetCtrl = XVPHY_MMCM_TXUSRCLK_CTRL_REG;
	}
	else {
		RegOffsetCtrl = XVPHY_MMCM_RXUSRCLK_CTRL_REG;
	}

	/* Assert reset. */
	RegVal = XVphy_ReadReg(InstancePtr->Config.BaseAddr, RegOffsetCtrl);
	RegVal |= XVPHY_MMCM_USRCLK_CTRL_RST_MASK;
	XVphy_WriteReg(InstancePtr->Config.BaseAddr, RegOffsetCtrl, RegVal);

	if (!Hold) {
		/* De-assert reset. */
		RegVal &= ~XVPHY_MMCM_USRCLK_CTRL_RST_MASK;
		XVphy_WriteReg(InstancePtr->Config.BaseAddr, RegOffsetCtrl,
									RegVal);
	}
}

/*****************************************************************************/
/**
* This function will reset the mixed-mode clock manager (MMCM) core.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	Dir is an indicator for TX or RX.
* @param	Enable is an indicator whether to "Enable" the locked mask
*		if set to 1. If set to 0: reset, then disable.
*
* @return	None.
*
* @note		None.
*
******************************************************************************/
void XVphy_MmcmLockedMaskEnable(XVphy *InstancePtr, U8_T QuadId,
		XVphy_DirectionType Dir, U8_T Enable)
{
	U32_T RegOffsetCtrl;
	U32_T RegVal;

	/* Suppress Warning Messages */
	QuadId = QuadId;

	if (Dir == XVPHY_DIR_TX) {
		RegOffsetCtrl = XVPHY_MMCM_TXUSRCLK_CTRL_REG;
	}
	else {
		RegOffsetCtrl = XVPHY_MMCM_RXUSRCLK_CTRL_REG;
	}

	/* Assert reset. */
	RegVal = XVphy_ReadReg(InstancePtr->Config.BaseAddr, RegOffsetCtrl);
	RegVal |= XVPHY_MMCM_USRCLK_CTRL_LOCKED_MASK_MASK;
	XVphy_WriteReg(InstancePtr->Config.BaseAddr, RegOffsetCtrl, RegVal);

	if (!Enable) {
		/* De-assert reset. */
		RegVal &= ~XVPHY_MMCM_USRCLK_CTRL_LOCKED_MASK_MASK;
		XVphy_WriteReg(InstancePtr->Config.BaseAddr, RegOffsetCtrl,
									RegVal);
	}
}

/*****************************************************************************/
/**
* This function will get the lock status of the mixed-mode clock
* manager (MMCM) core.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	Dir is an indicator for TX or RX.

*
* @return	AVER_TRUE if Locked else AVER_FALSE.
*
* @note		None.
*
******************************************************************************/
U8_T XVphy_MmcmLocked(XVphy *InstancePtr, U8_T QuadId, XVphy_DirectionType Dir)
{
	U32_T RegOffsetCtrl;
	U32_T RegVal;

	/* Suppress Warning Messages */
	QuadId = QuadId;

	if (Dir == XVPHY_DIR_TX) {
		RegOffsetCtrl = XVPHY_MMCM_TXUSRCLK_CTRL_REG;
	}
	else {
		RegOffsetCtrl = XVPHY_MMCM_RXUSRCLK_CTRL_REG;
	}

	RegVal = XVphy_ReadReg(InstancePtr->Config.BaseAddr, RegOffsetCtrl) &
				XVPHY_MMCM_USRCLK_CTRL_LOCKED_MASK;

	return (RegVal ? AVER_TRUE : AVER_FALSE);

}

/*****************************************************************************/
/**
* This function obtains the divider value of the BUFG_GT peripheral.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	Dir is an indicator for TX or RX
* @param	Div 3-Bit divider value
*
* @return	None.
*
******************************************************************************/
void XVphy_SetBufgGtDiv(XVphy *InstancePtr, XVphy_DirectionType Dir, U8_T Div)
{
	U32_T RegVal;
	U32_T RegOffset;
	U8_T Divider = Div;

	if (Divider == 0) {
		Divider = 1;
	}
	else {
		Divider = Divider - 1;
	}


	if (Dir == XVPHY_DIR_TX) {
		RegOffset = XVPHY_BUFGGT_TXUSRCLK_REG;
	}
	else {
		RegOffset = XVPHY_BUFGGT_RXUSRCLK_REG;
	}

	/* Read BUFG_GT register. */
	RegVal = XVphy_ReadReg(InstancePtr->Config.BaseAddr, RegOffset);
	RegVal &= ~XVPHY_BUFGGT_XXUSRCLK_DIV_MASK;

	/* Shift divider value to correct position. */
	Divider <<= XVPHY_BUFGGT_XXUSRCLK_DIV_SHIFT;
	Divider &= XVPHY_BUFGGT_XXUSRCLK_DIV_MASK;
	RegVal |= Divider;

	/* Write new value to BUFG_GT ctrl register. */
	XVphy_WriteReg(InstancePtr->Config.BaseAddr, RegOffset, RegVal);
}

/*****************************************************************************/
/**
* This function will power down the specified GT PLL.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to power down the PLL for.
* @param	Dir is an indicator for TX or RX.
* @param	Hold is an indicator whether to "hold" the power down if set
*		to 1. If set to 0: power down, then power back up.
*
* @return
*		- XST_SUCCESS.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_PowerDownGtPll(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId,
		U8_T Hold)
{
	U32_T MaskVal = 0;
	U32_T RegVal;
	U8_T Id, Id0, Id1;

	/* Suppress Warning Messages */
	QuadId = QuadId;

	if (XVPHY_ISCH(ChId)) {
		XVphy_Ch2Ids(InstancePtr, ChId, &Id0, &Id1);
	}
	else {
		/* When powering down a QPLL, power down for all channels. */
		XVphy_Ch2Ids(InstancePtr, XVPHY_CHANNEL_ID_CHA, &Id0, &Id1);
	}
	for (Id = Id0; Id <= Id1; Id++) {
		if (ChId == XVPHY_CHANNEL_ID_CMN0) {
			MaskVal |= XVPHY_POWERDOWN_CONTROL_QPLL0PD_MASK(Id);
		}
		else if (ChId == XVPHY_CHANNEL_ID_CMN1) {
			MaskVal |= XVPHY_POWERDOWN_CONTROL_QPLL1PD_MASK(Id);
		}
		else if (ChId == XVPHY_CHANNEL_ID_CMNA) {
			MaskVal |= XVPHY_POWERDOWN_CONTROL_QPLL0PD_MASK(Id) |
				   XVPHY_POWERDOWN_CONTROL_QPLL1PD_MASK(Id);
		}
		else {
			MaskVal |= XVPHY_POWERDOWN_CONTROL_CPLLPD_MASK(Id);
		}
	}

	RegVal = XVphy_ReadReg(InstancePtr->Config.BaseAddr,
					XVPHY_POWERDOWN_CONTROL_REG);
	RegVal |= MaskVal;
	XVphy_WriteReg(InstancePtr->Config.BaseAddr,
					XVPHY_POWERDOWN_CONTROL_REG, RegVal);

	if (!Hold) {
		RegVal &= ~MaskVal;
		XVphy_WriteReg(InstancePtr->Config.BaseAddr,
					XVPHY_POWERDOWN_CONTROL_REG, RegVal);
	}

	return XST_SUCCESS;
}

/*****************************************************************************/
/**
* This function will try to find the necessary PLL divisor values to produce
* the configured line rate given the specified PLL input frequency. This will
* be done for all channels specified by ChId.
* This function is a wrapper for XVphy_PllCalculator.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to calculate the PLL values for.
* @param	ChId is the channel ID to calculate the PLL values for.
* @param	Dir is an indicator for TX or RX.
* @param	PllClkInFreqHz is the PLL input frequency on which to base the
*		calculations on. A value of 0 indicates to use the currently
*		configured quad PLL reference clock. A non-zero value indicates
*		to ignore what is currently configured in SW, and use a custom
*		frequency instead.
*
* @return
*		- XST_SUCCESS if valid PLL values were found to satisfy the
*		  constraints.
*		- XST_FAILURE otherwise.
*
* @note		If successful, the channel's PllParams structure will be
*		modified with the valid PLL parameters.
*
******************************************************************************/
U32_T XVphy_ClkCalcParams(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId,
		XVphy_DirectionType Dir, U32_T PllClkInFreqHz)
{
	U32_T Status = XST_SUCCESS;
	U8_T Id, Id0, Id1;

	XVphy_Ch2Ids(InstancePtr, ChId, &Id0, &Id1);
	for (Id = Id0; Id <= Id1; Id++) {
		Status = XVphy_PllCalculator(InstancePtr, QuadId,
					(XVphy_ChannelId)Id, Dir, PllClkInFreqHz);
		if (Status != XST_SUCCESS) {
			return Status;
		}
	}

	return Status;
}

/*****************************************************************************/
/**
* This function will set the current output divider configuration over DRP.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID for which to write the settings for.
* @param	Dir is an indicator for RX or TX.
*
* @return
*		- XST_SUCCESS if the configuration was successful.
*		- XST_FAILURE otherwise.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_OutDivReconfig(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId,
		XVphy_DirectionType Dir)
{
	U32_T Status;
	U8_T Id;
	U8_T Id0;
	U8_T Id1;

	if (!XVPHY_ISCH(ChId)) {
		ChId = XVPHY_CHANNEL_ID_CHA;
	}

	XVphy_LogWrite(InstancePtr, (Dir == XVPHY_DIR_TX) ?
		XVPHY_LOG_EVT_GT_TX_RECONFIG : XVPHY_LOG_EVT_GT_RX_RECONFIG, 0);

	XVphy_Ch2Ids(InstancePtr, ChId, &Id0, &Id1);
	for (Id = Id0; Id <= Id1; Id++) {
		Status = XVphy_OutDivChReconfig(InstancePtr, QuadId,
					(XVphy_ChannelId)Id, Dir);
		if (Status != XST_SUCCESS) {
			break;
		}
	}

	return Status;
}

/*****************************************************************************/
/**
* This function will set the current RX/TX configuration over DRP.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID for which to write the settings for.
* @param	Dir is an indicator for RX or TX.
*
* @return
*		- XST_SUCCESS if the configuration was successful.
*		- XST_FAILURE otherwise.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_DirReconfig(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId,
		XVphy_DirectionType Dir)
{
	U32_T Status = XST_SUCCESS;
	U8_T Id, Id0, Id1;

	if ((InstancePtr->Config.XcvrType == XVPHY_GT_TYPE_GTHE2) &&
			(Dir == XVPHY_DIR_TX)) {
		return XST_SUCCESS;
	}

    if ((InstancePtr->Config.XcvrType == XVPHY_GT_TYPE_GTPE2) &&
		((InstancePtr->Config.TxProtocol == XVPHY_PROTOCOL_DP) ||
		 (InstancePtr->Config.RxProtocol == XVPHY_PROTOCOL_DP))) {
               ChId = XVPHY_CHANNEL_ID_CHA;
    }

	XVphy_Ch2Ids(InstancePtr, ChId, &Id0, &Id1);
	for (Id = Id0; Id <= Id1; Id++) {
		if (Dir == XVPHY_DIR_TX) {
			Status |= XVphy_TxChReconfig(InstancePtr, QuadId,
											(XVphy_ChannelId)Id);
		}
		else {
			Status |= XVphy_RxChReconfig(InstancePtr, QuadId,
											(XVphy_ChannelId)Id);
		}
		if (Status != XST_SUCCESS) {
			Status = XST_FAILURE;
			break;
		}
	}

	XVphy_LogWrite(InstancePtr, (Dir == XVPHY_DIR_TX) ?
		XVPHY_LOG_EVT_GT_TX_RECONFIG : XVPHY_LOG_EVT_GT_RX_RECONFIG, 1);

	return Status;
}

/*****************************************************************************/
/**
* This function will set the current clocking settings for each channel to
* hardware based on the configuration stored in the driver's instance.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID for which to write the settings for.
*
* @return
*		- XST_SUCCESS if the configuration was successful.
*		- XST_FAILURE otherwise.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_ClkReconfig(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId)
{
	U32_T Status = XST_SUCCESS;
	U8_T Id;
	U8_T Id0;
	U8_T Id1;

	XVphy_Ch2Ids(InstancePtr, ChId, &Id0, &Id1);
	for (Id = Id0; Id <= Id1; Id++) {
		if (XVPHY_ISCH(Id)) {
			Status |= XVphy_ClkChReconfig(InstancePtr, QuadId,
											(XVphy_ChannelId)Id);
		}
		else if (XVPHY_ISCMN(ChId)) {
			if (((InstancePtr->Config.TxProtocol == XVPHY_PROTOCOL_HDMI) ||
				(InstancePtr->Config.RxProtocol == XVPHY_PROTOCOL_HDMI)) &&
			    (InstancePtr->HdmiIsQpllPresent == AVER_FALSE)) {
				XVphy_LogWrite(InstancePtr, XVPHY_LOG_EVT_NO_QPLL_ERR, 1);
				XVphy_CfgErrIntr(InstancePtr, XVPHY_ERR_NO_QPLL, 1);
				XVphy_ErrorHandler(InstancePtr);
				Status = XST_FAILURE;
				return Status;
			}
			Status |= XVphy_ClkCmnReconfig(InstancePtr, QuadId,
											(XVphy_ChannelId)Id);
		}
		if (Status != XST_SUCCESS) {
			Status = XST_FAILURE;
			return Status;
		}
	}

	if (XVPHY_ISCH(ChId)) {
		XVphy_LogWrite(InstancePtr, XVPHY_LOG_EVT_CPLL_RECONFIG, 1);
	}
	else if (XVPHY_ISCMN(ChId) &&
			(InstancePtr->Config.XcvrType != XVPHY_GT_TYPE_GTPE2)) {
		XVphy_LogWrite(InstancePtr, XVPHY_LOG_EVT_QPLL_RECONFIG, 1);
	}
	else if (XVPHY_ISCMN(ChId)) { /* GTPE2. */
		XVphy_LogWrite(InstancePtr, (ChId == XVPHY_CHANNEL_ID_CMN0) ?
			XVPHY_LOG_EVT_PLL0_RECONFIG :
			XVPHY_LOG_EVT_PLL1_RECONFIG, 1);
	}

	return Status;
}

/*****************************************************************************/
/**
* This function will set the channel IDs to correspond with the supplied
* channel ID based on the protocol. HDMI uses 3 channels; DP uses 4. This ID
* translation is done to allow other functions to operate iteratively over
* multiple channels.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	ChId is the channel ID used to determine the indices.
* @param	Id0 is a pointer to the start channel ID to set.
* @param	Id1 is a pointer to the end channel ID to set.
*
* @return	None.
*
* @note		The contents of Id0 and Id1 will be set according to ChId.
*
******************************************************************************/
void XVphy_Ch2Ids(XVphy *InstancePtr, XVphy_ChannelId ChId,
		U8_T *Id0, U8_T *Id1)
{
	U8_T Channels = 4;

	if (ChId == XVPHY_CHANNEL_ID_CHA) {
		*Id0 = XVPHY_CHANNEL_ID_CH1;
		if ((InstancePtr->Config.TxProtocol == XVPHY_PROTOCOL_HDMI) ||
			(InstancePtr->Config.RxProtocol == XVPHY_PROTOCOL_HDMI)) {
			if (InstancePtr->Config.UseGtAsTxTmdsClk == AVER_TRUE) {
				*Id1 = XVPHY_CHANNEL_ID_CH4;
			}
			else {
				*Id1 = XVPHY_CHANNEL_ID_CH3;
			}
		}
		else {
			Channels = ((InstancePtr->Config.TxChannels >=
							InstancePtr->Config.RxChannels) ?
									InstancePtr->Config.TxChannels :
									InstancePtr->Config.RxChannels);

			if (Channels == 1) {
				*Id1 = XVPHY_CHANNEL_ID_CH1;
			}
			else if (Channels == 2) {
				*Id1 = XVPHY_CHANNEL_ID_CH2;
			}
			else if (Channels == 3) {
				*Id1 = XVPHY_CHANNEL_ID_CH3;
			}
			else {
				*Id1 = XVPHY_CHANNEL_ID_CH4;
			}
		}
	}
	else if (ChId == XVPHY_CHANNEL_ID_CMNA) {
		*Id0 = XVPHY_CHANNEL_ID_CMN0;
		if ((InstancePtr->Config.XcvrType == XVPHY_GT_TYPE_GTHE3) ||
		    (InstancePtr->Config.XcvrType == XVPHY_GT_TYPE_GTHE4) ||
		    (InstancePtr->Config.XcvrType == XVPHY_GT_TYPE_GTYE4)) {
			*Id1 = XVPHY_CHANNEL_ID_CMN1;
		}
		else {
			*Id1 = XVPHY_CHANNEL_ID_CMN0;
		}
	}
	else {
		*Id0 = *Id1 = ChId;
	}
}

/*****************************************************************************/
/**
* This function will translate from XVphy_PllType to XVphy_SysClkDataSelType.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
*
* @return	The reference clock type based on the PLL selection.
*
* @note		None.
*
******************************************************************************/
XVphy_SysClkDataSelType Pll2SysClkData(XVphy_PllType PllSelect)
{
	return	(PllSelect == XVPHY_PLL_TYPE_CPLL) ?
			XVPHY_SYSCLKSELDATA_TYPE_CPLL_OUTCLK :
		(PllSelect == XVPHY_PLL_TYPE_QPLL) ?
			XVPHY_SYSCLKSELDATA_TYPE_QPLL_OUTCLK :
		(PllSelect == XVPHY_PLL_TYPE_QPLL0) ?
			XVPHY_SYSCLKSELDATA_TYPE_QPLL0_OUTCLK :
		(PllSelect == XVPHY_PLL_TYPE_QPLL1) ?
			XVPHY_SYSCLKSELDATA_TYPE_QPLL1_OUTCLK :
		(PllSelect == XVPHY_PLL_TYPE_PLL0) ?
			XVPHY_SYSCLKSELDATA_TYPE_PLL0_OUTCLK :
		XVPHY_SYSCLKSELDATA_TYPE_PLL1_OUTCLK;
}

/*****************************************************************************/
/**
* This function will translate from XVphy_PllType to XVphy_SysClkOutSelType.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
*
* @return	The reference clock type based on the PLL selection.
*
* @note		None.
*
******************************************************************************/
XVphy_SysClkOutSelType Pll2SysClkOut(XVphy_PllType PllSelect)
{
	return	(PllSelect == XVPHY_PLL_TYPE_CPLL) ?
			XVPHY_SYSCLKSELOUT_TYPE_CPLL_REFCLK :
		(PllSelect == XVPHY_PLL_TYPE_QPLL) ?
			XVPHY_SYSCLKSELOUT_TYPE_QPLL_REFCLK :
		(PllSelect == XVPHY_PLL_TYPE_QPLL0) ?
			XVPHY_SYSCLKSELOUT_TYPE_QPLL0_REFCLK :
		(PllSelect == XVPHY_PLL_TYPE_QPLL1) ?
			XVPHY_SYSCLKSELOUT_TYPE_QPLL1_REFCLK :
		(PllSelect == XVPHY_PLL_TYPE_PLL0) ?
			XVPHY_SYSCLKSELOUT_TYPE_PLL0_REFCLK :
		XVPHY_SYSCLKSELOUT_TYPE_PLL1_REFCLK;
}

/*****************************************************************************/
/**
* This function will try to find the necessary PLL divisor values to produce
* the configured line rate given the specified PLL input frequency.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to calculate the PLL values for.
* @param	ChId is the channel ID to calculate the PLL values for.
* @param	Dir is an indicator for TX or RX.
* @param	PllClkInFreqHz is the PLL input frequency on which to base the
*		calculations on. A value of 0 indicates to use the currently
*		configured quad PLL reference clock. A non-zero value indicates
*		to ignore what is currently configured in SW, and use a custom
*		frequency instead.
*
* @return
*		- XST_SUCCESS if valid PLL values were found to satisfy the
*		  constraints.
*		- XST_FAILURE otherwise.
*
* @note		If successful, the channel's PllParams structure will be
*		modified with the valid PLL parameters.
*
******************************************************************************/
U32_T XVphy_PllCalculator(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId, XVphy_DirectionType Dir,
		U32_T PllClkInFreqHz)
{
	U32_T Status;
	U64_T PllClkOutFreqHz;
	U64_T CalcLineRateFreqHz;
	U8_T Id, Id0, Id1;
	U64_T PllClkInFreqHzIn = PllClkInFreqHz;
	XVphy_Channel *PllPtr = &InstancePtr->Quads[QuadId].
		Plls[XVPHY_CH2IDX(ChId)];

	if (!PllClkInFreqHzIn) {
		PllClkInFreqHzIn = XVphy_GetQuadRefClkFreq(InstancePtr, QuadId,
					PllPtr->PllRefClkSel);
	}

	/* Select PLL value table offsets. */
	const XVphy_GtPllDivs *GtPllDivs;
	
#if 0
	if (XVPHY_ISCH(ChId)) {
		GtPllDivs = &InstancePtr->GtAdaptor->CpllDivs;
	}
	else {
		GtPllDivs = &InstancePtr->GtAdaptor->QpllDivs;
	}
#else
	const XVphy_GtPllDivs CpllDivs = {
		Gtpe2CpllDivsM,
		Gtpe2CpllDivsN1,
		Gtpe2CpllDivsN2,
		Gtpe2CpllDivsD,
	};
	const XVphy_GtPllDivs QpllDivs = {
		Gtpe2CpllDivsM,
		Gtpe2CpllDivsN1,
		Gtpe2CpllDivsN2,
		Gtpe2CpllDivsD,
	};
	if (XVPHY_ISCH(ChId)) {
		GtPllDivs = &CpllDivs;
	}
	else {
		GtPllDivs = &QpllDivs;
	}
#endif

	const U8_T *M, *N1, *N2, *D;
	for (N2 = GtPllDivs->N2; *N2 != 0; N2++) {
	for (N1 = GtPllDivs->N1; *N1 != 0; N1++) {
	for (M = GtPllDivs->M;   *M != 0;  M++) {

		PllClkOutFreqHz = (PllClkInFreqHzIn * *N1 * *N2) / *M;

		/* Test if the calculated PLL clock is in the VCO range. */
		Status = XVphy_CheckPllOpRange(InstancePtr, QuadId, ChId,
				PllClkOutFreqHz);

		if (Status != XST_SUCCESS) {
			continue;
		}

		if ((InstancePtr->Config.XcvrType == XVPHY_GT_TYPE_GTPE2) ||
				(XVPHY_ISCH(ChId))) {
			PllClkOutFreqHz *= 2;
		}
		/* Apply TX/RX divisor. */
		for (D = GtPllDivs->D; *D != 0; D++) {
			CalcLineRateFreqHz = PllClkOutFreqHz / *D;
			if (CalcLineRateFreqHz == PllPtr->LineRateHz) {
				goto calc_done;
			}
		}
	}
	}
	}
	/* Calculation failed, don't change divisor settings. */
	return XST_FAILURE;

calc_done:
	/* Found the multiplier and divisor values for requested line rate. */
	PllPtr->PllParams.MRefClkDiv = *M;
	PllPtr->PllParams.NFbDiv = *N1;
	PllPtr->PllParams.N2FbDiv = *N2; /* Won't be used for QPLL.*/
	PllPtr->PllParams.IsLowerBand = 1; /* Won't be used for CPLL. */

	if (XVPHY_ISCMN(ChId)) {
		/* Same divisor value for all channels if using a QPLL. */
		ChId = XVPHY_CHANNEL_ID_CHA;
	}

	XVphy_Ch2Ids(InstancePtr, ChId, &Id0, &Id1);
	for (Id = Id0; Id <= Id1; Id++) {
		InstancePtr->Quads[QuadId].Plls[XVPHY_CH2IDX(Id)].OutDiv[Dir] =
			*D;
		if (Dir == XVPHY_DIR_RX) {
			XVphy_CfgSetCdr(InstancePtr, QuadId, (XVphy_ChannelId)Id);
		}
	}

	return XST_SUCCESS;
}

/*****************************************************************************/
/**
* This function calculates the PLL VCO operating frequency.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	Dir is an indicator for TX or RX.
*
* @return	PLL VCO frequency in Hz
*
* @note		None.
*
******************************************************************************/
U64_T XVphy_GetPllVcoFreqHz(XVphy *InstancePtr, U8_T QuadId,
		XVphy_ChannelId ChId, XVphy_DirectionType Dir)
{
	U64_T PllxVcoRateHz;
	U64_T PllRefClkHz;
	XVphy_Channel *PllPtr = &InstancePtr->Quads[QuadId].
                    Plls[XVPHY_CH2IDX(ChId)];

	if (Dir == XVPHY_DIR_TX) {
		if (InstancePtr->Config.TxProtocol == XVPHY_PROTOCOL_HDMI) {
			PllRefClkHz = InstancePtr->HdmiTxRefClkHz;
		}
		else {
			PllRefClkHz = XVphy_GetQuadRefClkFreq(InstancePtr, QuadId,
									PllPtr->PllRefClkSel);
		}
	}
	else {
		if (InstancePtr->Config.RxProtocol == XVPHY_PROTOCOL_HDMI) {
#if defined (XPAR_XV_HDMITX_0_DEVICE_ID) || defined (XPAR_XV_HDMIRX_0_DEVICE_ID)
			if (InstancePtr->HdmiRxDruIsEnabled) {
				PllRefClkHz = XVphy_DruGetRefClkFreqHz(InstancePtr);
			}
			else {
				PllRefClkHz = InstancePtr->HdmiRxRefClkHz;
			}
#else
			PllRefClkHz = 0;
#endif
		}
		else {
			PllRefClkHz = XVphy_GetQuadRefClkFreq(InstancePtr, QuadId,
									PllPtr->PllRefClkSel);
		}
	}

	PllxVcoRateHz = (PllRefClkHz *
				InstancePtr->Quads[QuadId].Plls[XVPHY_CH2IDX(ChId)].
								PllParams.N1FbDiv *
				InstancePtr->Quads[QuadId].Plls[XVPHY_CH2IDX(ChId)].
								PllParams.N2FbDiv) /
				InstancePtr->Quads[QuadId].Plls[XVPHY_CH2IDX(ChId)].
								PllParams.MRefClkDiv;

	return PllxVcoRateHz;
}


/*****************************************************************************/
/**
* This function is a transceiver adaptor to set the clock and data recovery
* (CDR) values for a given channel.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
*
* @return
*		- XST_SUCCESS if the configuration was successful.
*		- XST_FAILURE otherwise.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_CfgSetCdr(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId)
{
#if 0 //Shawn_20190531 [Modify to GTPE2]
	return InstancePtr->GtAdaptor->CfgSetCdr(InstancePtr, QuadId, ChId);
#else
	return XVphy_Gtpe2CfgSetCdr(InstancePtr, QuadId, ChId);
#endif
}

/*****************************************************************************/
/**
* This function is a transceiver adaptor to check if a given PLL output
* frequency is within the operating range of the PLL for the GT type.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
* @param	PllClkOutFreqHz is the frequency to check.
*
* @return
*		- XST_SUCCESS if the frequency resides within the PLL's range.
*		- XST_FAILURE otherwise.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_CheckPllOpRange(XVphy *InstancePtr, U8_T QuadId,
							XVphy_ChannelId ChId, U64_T PllClkOutFreqHz)
{
#if 0 //Shawn_20190531 [Modify to GTPE2]
	return InstancePtr->GtAdaptor->CheckPllOpRange(InstancePtr, QuadId, ChId,
							PllClkOutFreqHz);
#else
	return XVphy_Gtpe2CheckPllOpRange(InstancePtr, QuadId, ChId,
		PllClkOutFreqHz);
#endif
}

/*****************************************************************************/
/**
* This function is a transceiver adaptor to set the output divider logic for
* a given channel.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
* @param	Dir is an indicator for RX or TX.
*
* @return
*		- XST_SUCCESS if the configuration was successful.
*		- XST_FAILURE otherwise.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_OutDivChReconfig(XVphy *InstancePtr, U8_T QuadId,
							XVphy_ChannelId ChId, XVphy_DirectionType Dir)
{
#if 0 //Shawn_20190531 [Modify to GTPE2]
	return InstancePtr->GtAdaptor->OutDivChReconfig(InstancePtr, QuadId,
		ChId, Dir);
#else
	return XVphy_Gtpe2OutDivChReconfig(InstancePtr, QuadId,
		ChId, Dir);
#endif
}

/*****************************************************************************/
/**
* This function is a transceiver adaptor to configure the channel
* clock settings.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
*
* @return
*		- XST_SUCCESS if the configuration was successful.
*		- XST_FAILURE otherwise.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_ClkChReconfig(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId)
{
#if 0 //Shawn_20190531 [Modify to GTPE2]
	return InstancePtr->GtAdaptor->ClkChReconfig(InstancePtr, QuadId, ChId);
#else
	return XVphy_Gtpe2ClkChReconfig(InstancePtr, QuadId, ChId);
#endif
}

/*****************************************************************************/
/**
* This function is a transceiver adaptor to configure the common channel
* clock settings.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	CmnId is the common channel ID to operate on.
*
* @return
*		- XST_SUCCESS if the configuration was successful.
*		- XST_FAILURE otherwise.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_ClkCmnReconfig(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId)
{
#if 0 //Shawn_20190531 [Modify to GTPE2]
	return InstancePtr->GtAdaptor->ClkCmnReconfig(InstancePtr, QuadId, ChId);
#else
	return XVphy_Gtpe2ClkCmnReconfig(InstancePtr, QuadId, ChId);
#endif
}

/*****************************************************************************/
/**
* This function is a transceiver adaptor to configure the channel's
* RX settings.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
*
* @return
*		- XST_SUCCESS if the configuration was successful.
*		- XST_FAILURE otherwise.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_RxChReconfig(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId)
{
#if 0 //Shawn_20190531 [Modify to GTPE2]
	return InstancePtr->GtAdaptor->RxChReconfig(InstancePtr, QuadId, ChId);
#else
	return XVphy_Gtpe2RxChReconfig(InstancePtr, QuadId, ChId);
#endif
}

/*****************************************************************************/
/**
* This function is a transceiver adaptor to configure the channel's
* TX settings.
*
* @param	InstancePtr is a pointer to the XVphy core instance.
* @param	QuadId is the GT quad ID to operate on.
* @param	ChId is the channel ID to operate on.
*
* @return
*		- XST_SUCCESS if the configuration was successful.
*		- XST_FAILURE otherwise.
*
* @note		None.
*
******************************************************************************/
U32_T XVphy_TxChReconfig(XVphy *InstancePtr, U8_T QuadId, XVphy_ChannelId ChId)
{
#if 0 //Shawn_20190531 [Modify to GTPE2]
	return InstancePtr->GtAdaptor->TxChReconfig(InstancePtr, QuadId, ChId);
#else
	return XVphy_Gtpe2TxChReconfig(InstancePtr, QuadId, ChId);
#endif
}


/*****************************************************************************/
/**
* This function is the error condition handler
*
* @param	InstancePtr is a pointer to the VPHY instance.
* @param    ErrIrqType is the error type
*
* @return	None.
*
* @note		None.
*
******************************************************************************/
void XVphy_ErrorHandler(XVphy *InstancePtr)
{
	if (InstancePtr->ErrorCallback != NULL) {
		InstancePtr->ErrorCallback(InstancePtr->ErrorRef);
	}
}

#if (XPAR_VPHY_0_TRANSCEIVER == XVPHY_GTXE2)
/*****************************************************************************/
/**
* This function is the error condition handler
*
* @param	InstancePtr is a pointer to the VPHY instance.
* @param    ErrIrqType is the error type
*
* @return	None.
*
* @note		None.
*
******************************************************************************/
void XVphy_PllLayoutErrorHandler(XVphy *InstancePtr)
{
	if (InstancePtr->PllLayoutErrorCallback != NULL) {
		InstancePtr->PllLayoutErrorCallback(InstancePtr->PllLayoutErrorRef);
	}
}
#endif
